{-# LANGUAGE FlexibleContexts #-}
--------------------------------------------------------------------------
-- |
-- Module      :  CGI app
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
--
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable
--
-- This is the muterm web main module.
--
-----------------------------------------------------------------------------

module Main (

-- * Exported functions

main

) where
import Data.Either
import Data.Maybe

-- cgi modules
import Network.CGI (output, getInput, CGIResult, CGIT, catchCGI, liftIO, readInput, handleErrors, runCGI)
import System.IO (hPutStrLn, stderr, stdout)

-- muterm base
import Interface.CLI (Opt (..), parseOptions, autoparse)
import Framework.Problem.Types(MProblem (..))
import MuTerm.Framework.Proof (Proof (..), PrettyInfo)
import MuTerm.Framework.Problem (Problem (..), MkProblem (..), IsProblem(..))
import Framework.Proofs (runProofSol, Solution (..))
import Framework.Strategies (rewStrat, crewStrat)
import Solver.SolverSMTExt (ModelGenerator(..))

-- parsers
import Text.ParserCombinators.Parsec.Error(ParseError)

-- pretty
import Text.PrettyPrint.HughesPJClass (Pretty (..), punctuate, char, hcat, text)

-- Auxiliar
import Utils.Auxiliar (replaceString)

-- errors
import Control.Exception as E (evaluate, Exception (..), catch, throwTo)
import Control.Concurrent
import Data.Typeable

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | The Timeout Exception
data TimeoutException = TimeoutException deriving (Show, Typeable)

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

instance Exception TimeoutException

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------


main :: IO ()
main = runCGI $ handleErrors cgiMain

removeJust :: Maybe String -> String
removeJust (Just a) = a
removeJust (Nothing) = ""

cgiMain :: CGIT IO CGIResult
cgiMain = do txtTrs <- getInput "txtTrs"
             txtTimeout <- readInput "txtTimeout"

             let trs' = removeJust txtTrs
             -- parametros para los tipos de interpretacion
             let timeout = let to = fromMaybe 60 txtTimeout
                           in if to < 30 || to > 300
                              then 60
                              else to

             -- obtenemos el trs desde el programa .maude
             let trs = autoparse "input.trs" trs'

             case trs of
               MRewriting problem -> do {
               let proof = rewStrat (Just timeout) problem
               ; sol <- liftIO $ withTimeout timeout $
               evaluate $ ((runProofSol proof) :: Solution (Proof PrettyInfo [] ()))
               ; case sol of
                   Just sol' -> output . replaceString ">" "&gt;" . replaceString "<" "&lt;" . show . pPrint $ sol'
                   Nothing -> output $ "MAYBE"
               }
               MCRewriting problem -> do {
               let proof = crewStrat (Just timeout) problem
               ; sol <- liftIO $ withTimeout timeout $
               evaluate $ ((runProofSol proof) :: Solution (Proof PrettyInfo [] ()))
               ; case sol of
                   Just sol' -> output . replaceString ">" "&gt;" . replaceString "<" "&lt;" . show . pPrint $ sol'
                   Nothing -> output $ "MAYBE"
               }
---               otherwise -> output $ "Problem not supported"

-- | taken from Pepe Iborra's Narradar. We launch two process in
-- parallel comunicated by a semaphore done. First to finish, first to
-- win.
withTimeout :: Int -> IO t -> IO (Maybe t)
withTimeout t m = do
  res  <- newEmptyMVar
  done <- newMVar ()

  worker_id <- forkIO $ (`E.catch` \TimeoutException -> return ()) $ do
             val <- m
             takeMVar done
             putMVar res (Just val)

  clocker_id <- forkIO $ do
             threadDelay (t * 1000000)
             echo "time out"
             takeMVar done
             echo "signaled"
             E.throwTo worker_id TimeoutException
             putMVar res Nothing
  takeMVar res
    where
      echo = hPutStrLn stderr
