{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}

module Framework.ExternalCalls (

-- * Exported functions

callToInfChecker, callToMuTerm, callToProver9, callToWeakNormFort

) where
import System.Process (runCommand, waitForProcess, getProcessExitCode, terminateProcess)
import System.Exit (ExitCode(ExitSuccess))
import System.Directory(findExecutable, removeFile, doesFileExist)
import System.IO (Handle, openTempFile, hClose)
import Control.Monad (MonadPlus, when)
import Control.Exception (catch, finally, throw)
import qualified Data.ByteString.Lazy.Char8 as B (hPut)
import qualified Data.Binary.Builder as C (Builder,toLazyByteString)
import Data.Set as S (Set, elems, null, member, fromList)


-- | Call external Solver
callToInfChecker :: Maybe Int -> C.Builder -> IO (String, Maybe String)
callToInfChecker timeout infProblem =
    do
      tempIn  <- openTempFile filepath fileInfCheckerIn
      B.hPut (snd tempIn) (C.toLazyByteString infProblem)
      hClose (snd tempIn)
      tempOut <- openTempFile filepath fileInfCheckerOut
      hClose (snd tempOut)
      finally (execInfChecker timeout tempIn tempOut)
                  (do
                    fIn <- doesFileExist . fst $ tempIn
                    fOut <- doesFileExist . fst $ tempOut
                    -- comment if you don't want to remove temporary files
                    when fIn $ (removeFile . fst $ tempIn)
                    when fOut $ (removeFile . fst $ tempOut)
                    return True
                  )

-- | Call to infChecker
execInfChecker :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> IO (String, Maybe String)
execInfChecker timeout tempIn tempOut
  = do { existCmd <- findExecutable "infChecker"
       ; idProcess <- runCommand infCheckerCall
       ; exitCode  <- waitForProcess idProcess
       ; output    <- readFile . fst $! tempOut
       ; let line = lines output
       ; let n = if (length line) > 0 then head line else "MAYBE"
       ; return (n, Just $ output)
       }
  where
    infCheckerCall = buildInfCheckerCall timeout tempIn tempOut

-- | Define command call to infChecker
buildInfCheckerCall :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> String
buildInfCheckerCall Nothing tempIn tempOut
  = "(./timelimit -q -t 30 ./infChecker -i " ++ (fst tempIn) ++ " -t 30) > " ++ (fst tempOut)
buildInfCheckerCall (Just timeout) tempIn tempOut
  = "(./timelimit -q -t " ++ (show timeout) ++ " ./infChecker -i " ++ (fst tempIn) ++ " -t " ++ (show timeout) ++ ") > " ++ (fst tempOut)


-- | Call external Solver
callToMuTerm :: Maybe Int -> C.Builder -> IO (String, Maybe String)
callToMuTerm timeout termProblem =
    do
      tempIn  <- openTempFile filepath fileMuTermIn
      B.hPut (snd tempIn) (C.toLazyByteString termProblem)
      hClose (snd tempIn)
      tempOut <- openTempFile filepath fileMuTermOut
      hClose (snd tempOut)
      finally (execMuTerm timeout tempIn tempOut)
                  (do
                    fIn <- doesFileExist . fst $ tempIn
                    fOut <- doesFileExist . fst $ tempOut
                    -- comment if you don't want to remove temporary files
                    when fIn $ (removeFile . fst $ tempIn)
                    when fOut $ (removeFile . fst $ tempOut)
                    return True
                  )

-- | Call to muterm
execMuTerm :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> IO (String, Maybe String)
execMuTerm timeout tempIn tempOut
  = do { existCmd  <- findExecutable "muterm"
       ; idProcess <- runCommand muTermCall
       ; exitCode  <- waitForProcess idProcess
       ; output    <- readFile . fst $! tempOut
       ; let line = lines output
       ; if (length line) > 0 then return (head line, Just output) else return ("MAYBE", Just "MAYBE")
       }
  where
    muTermCall = buildMuTermCall timeout tempIn tempOut

-- | Define command call to muterm
buildMuTermCall :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> String
buildMuTermCall Nothing tempIn tempOut
  = "(./timelimit -q -t 30 ./muterm -i " ++ (fst tempIn) ++ " -t 30) > " ++ (fst tempOut)
buildMuTermCall (Just timeout) tempIn tempOut
  = "(./timelimit -q -t " ++ (show timeout) ++ " ./muterm -i " ++ (fst tempIn) ++ " -t " ++ (show timeout) ++ ") > " ++ (fst tempOut)


-- | Call external Solver
callToProver9 :: Ord String => Maybe Int -> C.Builder -> IO (String, Maybe String)
callToProver9 timeout provProblem =
    do
      tempIn  <- openTempFile filepath fileProver9In
      B.hPut (snd tempIn) (C.toLazyByteString provProblem)
      hClose (snd tempIn)
      tempOut <- openTempFile filepath fileProver9Out
      hClose (snd tempOut)
      finally (execProver9 timeout tempIn tempOut)
                  (do
                    fIn <- doesFileExist . fst $ tempIn
                    fOut <- doesFileExist . fst $ tempOut
                    -- comment if you don't want to remove temporary files
                    when fIn $ (removeFile . fst $ tempIn)
                    when fOut $ (removeFile . fst $ tempOut)
                    return True
                  )

-- | Call to Prover9
execProver9 :: Ord String => Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> IO (String, Maybe String)
execProver9 timeout tempIn tempOut
  = do { existCmd <- findExecutable "prover9"
       ; idProcess <- runCommand prov9Call
       ; exitCode  <- waitForProcess idProcess
       ; output    <- readFile . fst $! tempOut
       ; let sOut = lines output
       ; let n = if Prelude.null sOut then "MAYBE" else (if elem "THEOREM PROVED" sOut then "YES" else "NO")
       ; return (n, Just $ output)
       }
  where
    prov9Call = buildProver9Call timeout tempIn tempOut

-- | Define command call to Prover9
buildProver9Call :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> String
buildProver9Call Nothing tempIn tempOut
  = "(./timelimit -q -t 30 ./prover9 -t 30 -f " ++ (fst tempIn) ++ " > " ++ (fst tempOut) ++ ")"
buildProver9Call (Just timeout) tempIn tempOut
  = "(./timelimit -q -t " ++ (show timeout) ++ " ./prover9 -t " ++ (show timeout) ++ " -f " ++ (fst tempIn) ++ " > " ++ (fst tempOut) ++ ")"


callToWeakNormFort :: Maybe Int -> C.Builder -> IO (String, Maybe String)
callToWeakNormFort timeout normProblem = callToFort timeout normProblem "WN"


-- | Call external Solver
callToFort :: Maybe Int -> C.Builder -> String -> IO (String, Maybe String)
callToFort timeout normProblem formula =
    do
      tempIn  <- openTempFile filepath fileFortIn
      B.hPut (snd tempIn) (C.toLazyByteString normProblem)
      hClose (snd tempIn)
      tempOut <- openTempFile filepath fileFortOut
      hClose (snd tempOut)
      finally (execFort timeout tempIn tempOut formula)
                  (do
                    fIn <- doesFileExist . fst $ tempIn
                    fOut <- doesFileExist . fst $ tempOut
                    -- comment if you don't want to remove temporary files
                    when fIn $ (removeFile . fst $ tempIn)
                    when fOut $ (removeFile . fst $ tempOut)
                    return True
                  )

-- | Call to fort
execFort :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> String -> IO (String, Maybe String)
execFort timeout tempIn tempOut formula
  = do { existCmd  <- findExecutable "fort"
       ; idProcess <- runCommand fortCall
       ; exitCode  <- waitForProcess idProcess
       ; output    <- readFile . fst $! tempOut
       ; let line = lines output
       ; if (length line) > 0 then return (head line, Just output) else return ("MAYBE", Just "MAYBE")
       }
  where
    fortCall = buildFortCall timeout tempIn tempOut formula

-- | Define command call to fort
buildFortCall :: Maybe Int -> (FilePath, Handle) -> (FilePath, Handle) -> String -> String
buildFortCall Nothing tempIn tempOut formula
  = "(./timelimit -q -t 30 ./fort -c " ++ (formula) ++ " " ++ (fst tempIn) ++ ") > " ++ (fst tempOut)
buildFortCall (Just timeout) tempIn tempOut formula
  = "(./timelimit -q -t " ++ (show timeout) ++ " ./fort -c " ++ (formula) ++ " " ++ (fst tempIn) ++ ") > " ++ (fst tempOut)


-----------------------------------------------------------------------------
-- Configuration
-----------------------------------------------------------------------------

fileInfCheckerIn  = "infChecker.trs"
fileInfCheckerOut = "infChecker.out"

fileMuTermIn  = "muterm.trs"
fileMuTermOut = "muterm.out"

fileProver9In  = "prover9.in"
fileProver9Out = "prover9.out"

fileFortIn  = "fort.trs"
fileFortOut = "fort.out"

#ifndef WINDOWS
filepath = "/tmp/"
#endif
#ifdef WINDOWS
filepath = "files/"
#endif