{-# LANGUAGE FlexibleContexts,FlexibleInstances #-}
{-# LANGUAGE RecordWildCards #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Output.Pretty
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
--
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable
--
-- This module contains the pretty print of the proof functor.
--
-----------------------------------------------------------------------------

module Framework.Output.Pretty where

import MuTerm.Framework.Proof (ProofF (..), PrettyInfo, SomeInfo, Proof (..))
import Framework.Proofs (Solution (..))

import Control.Monad.Free (Free (..), foldFree)
import Data.Foldable as F (Foldable, toList)
import Data.List (intersperse)
import Text.PrettyPrint.HughesPJClass as Doc hiding (Style)
import Data.Set as S (Set, null, elems)
import Data.Map as M (Map, toList)

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Pretty

instance (Pretty a) => Pretty (Set a) where
    pPrint p
        = if S.null p then
              text "Empty"
          else
              hsep . punctuate (char '\n') . map pPrint . elems $ p

instance (Pretty a, Pretty b) => Pretty (Map a b) where
  pPrint = pPrint . M.toList

instance Pretty a => Pretty (Solution a) where
    pPrint (YES x xs) = text "YES\n" Doc.<> pPrint x
    pPrint (NO a)  = text "NO\n" Doc.<> pPrint a
    pPrint MAYBE   = text "MAYBE\n"

instance (Pretty (SomeInfo info), Monad m) => Pretty (Free (ProofF info m) a) where
    pPrint x = (runpPrintProof x) [1]

{--
instance (Foldable mp, Pretty a) => Pretty (ProofF PrettyInfo mp a) where pPrint = pPrintProofF'
instance (Foldable mp, Pretty a, Pretty (SomeInfo info)) => Pretty (ProofF info mp a) where pPrint = pPrintProofF'

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Prints the proof information
pPrintProofF' = f where
      f Success{..} =
        {-- pPrint problem Doc.<> --} pPrint procInfo Doc.<>
        text ("\n\nThe problem is finite.\n")
      f Refuted{..} =
        {-- pPrint problem Doc.<> --} pPrint procInfo Doc.<>
        text ("\n\nThe problem is infinite.\n")
      f DontKnow{..} =
        {-- pPrint problem Doc.<> --} pPrint procInfo Doc.<>
        text ("\n\nDon't know.\n")
      f (And proc prob sub)
       | length sub > 1 =
        {-- pPrint prob Doc.<> --} pPrint proc Doc.<> text "\n" Doc.<>
        text ("\n\nThe problem is decomposed in " ++ show (length sub) ++ " subproblems.\n") Doc.<> (hcat . map pPrint $ sub)
       | otherwise =
        {-- pPrint prob Doc.<> --} pPrint proc Doc.<> (hcat . map pPrint $ sub)
      f (Single{..}) =
        {-- pPrint problem Doc.<> --} pPrint procInfo Doc.<> text "\n" Doc.<>
          nest 8 (pPrint subProblem)
      f (MAnd p1 p2) =
        text ("\n\nThe problem is decomposed in 2 subproblems.\n") Doc.<> nest 8 (vcat $ punctuate (text "\n") $ map pPrint [p1,p2])
      f MDone = text "Done"
      f (Search sub) =  vcat (intersperse (text "\n\nTrying something different\n") $ map pPrint $ F.toList sub)
--}

-- | Pretty print of the solution
pPrintProofF :: (Pretty (SomeInfo info)) => ProofF info t ([Int] -> Doc) -> [Int] -> Doc
pPrintProofF Success{..}    =
        (\i -> text "\nProblem " Doc.<> (hcat . punctuate (char '.') . map int $ i) Doc.<> text ": " Doc.<> pPrint procInfo Doc.<>
        text ("\n\nThe problem is joinable.\n"))
pPrintProofF Refuted{..}    =
        (\i -> text "\nProblem " Doc.<> (hcat . punctuate (char '.') . map int $ i) Doc.<> text ": " Doc.<> pPrint procInfo Doc.<>
        text ("\n\nThe problem is not joinable.\n"))
pPrintProofF DontKnow{..}     =
        (\i -> text "\nProblem " Doc.<> (hcat . punctuate (char '.') . map int $ i) Doc.<> text ": " Doc.<> pPrint procInfo Doc.<>
        text ("\n\nDon't know.\n"))
pPrintProofF (And proc prob sub)
       | length sub > 1 =
        (\i -> text "\nProblem " Doc.<> (hcat . punctuate (char '.') . map int $ i) Doc.<> text ": " Doc.<> pPrint proc Doc.<> text "\n" Doc.<>
        text ("\n\nThe problem is decomposed in " ++ show (length sub) ++ " subproblems.\n") Doc.<> (hcat [(sub!!(j - 1)) (i ++ [j]) | j <- [1..length sub]]))
       | otherwise =
        (\i -> text "\nProblem " Doc.<> (hcat . punctuate (char '.') . map int $ i) Doc.<> text ": " Doc.<> pPrint proc Doc.<> (hcat [(sub!!(j - 1)) (i ++ [j]) | j <- [1..length sub]]))
pPrintProofF (Single{..}) =
        (\i -> text "\nProblem " Doc.<> (hcat . punctuate (char '.') . map int $ i) Doc.<> text ": " Doc.<> pPrint procInfo
               Doc.<> text "\n" Doc.<> nest 8 (subProblem i))
pPrintProofF (MAnd  p1 p2)  =
        (\i -> text ("\n\nThe problem is decomposed in 2 subproblems.\n") Doc.<> nest 8 (vcat $ punctuate (text "\n") $ [p1 (i ++ [1]),p2 (i ++ [2])]))
pPrintProofF MDone          = (\_ -> text "Done")

-- | Print of the solution tree
runpPrintProof
  :: (Pretty (SomeInfo info), Monad m) =>
     Free (ProofF info m) a -> [Int] -> Doc
runpPrintProof = foldFree (const (\_ -> Doc.empty)) pPrintProofF
