{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE GADTs #-}
{-# LANGUAGE FunctionalDependencies #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}

-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Problem.Types
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
--
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable
--
-- This module contains the different problems.
--
-----------------------------------------------------------------------------

module Framework.Problem.Types (

-- * Exported data

  MProblem (..), InConfCSTRS (..), InConfCSCTRS (..), InCPCSTRS (..), InCPCSCTRS (..), InTermCSTRS (..), InTermCSCTRS (..),
  InCPTermCSTRS (..), InCPTermCSCTRS (..), InCPLinCSTRS (..), InCPLinCSCTRS (..), InLinCSTRS (..), InLinCSCTRS (..),
  InCPTermPCCSCTRS (..), InNormCSTRS (..), InNormCSCTRS (..), InCPNormCSTRS (..), InCPNormCSCTRS (..),
  InCPPCCSTRS (..), InCPPCCSCTRS (..)


) where

import TRS.TRSs (TRS (..), CTRS (..), OSTRS (..), PrettyCOPS (..))
import TRS.Terms (Term (..))
import TRS.Identifiers (IdFTRS (..), IdVTRS (..), IdFCS (..), IdFOS (..), IdVOS (..), IdFAvC (..), Sort (..))
import Framework.Problems (IsConfProblem(..), IsCPProblem (..), CriticalPairInfo, IsTProblem (..), IsCPTermProblem (..), IsCPLinProblem (..),
                           IsLProblem (..), IsCPTermPCProblem (..), IsNProblem (..), IsCPNormProblem (..), IsCPPCProblem (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import MuTerm.Framework.Problem (IsProblem (..))
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.Problem.TCRewriting (TCRewriting (..))
import Framework.Problem.NRewriting (NRewriting (..))
import Framework.Problem.NCRewriting (NCRewriting (..))
import Framework.Problem.LRewriting (LRewriting (..))
import Framework.Problem.LCRewriting (LCRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.Problem.CPTCRewriting (CPTCRewriting (..))
import Framework.Problem.CPNRewriting (CPNRewriting (..))
import Framework.Problem.CPNCRewriting (CPNCRewriting (..))
import Framework.Problem.CPLRewriting (CPLRewriting (..))
import Framework.Problem.CPLCRewriting (CPLCRewriting (..))
import Framework.Problem.CPPCRewriting (CPPCRewriting (..))
import Framework.Problem.CPPCCRewriting (CPPCCRewriting (..))
import Framework.Problem.CPTPCRewriting (CPTPCRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import TRS.Rules (Cond)
import Data.Maybe (isJust, fromJust)

import Text.PrettyPrint.HughesPJClass as HPJC (Pretty (..), (<+>), (<>), hcat, punctuate, text, char)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | 'InCSTRS' is the CS-TRS
type InCSTRS   = OSTRS ((TRS (IdFOS Sort (IdFCS IdFTRS)) (IdVOS Sort IdVTRS) (Term (IdFOS Sort (IdFCS IdFTRS)) (IdVOS Sort IdVTRS))))

-- | 'InCSCTRS' is the CS-CTRS
type InCSCTRS   = OSTRS ((CTRS (IdFOS Sort (IdFCS IdFTRS)) (IdVOS Sort IdVTRS) (Term (IdFOS Sort (IdFCS IdFTRS)) (IdVOS Sort IdVTRS))))

-- | 'InCritPairs' are the input critical pairs info type
type InCritPair = CriticalPairInfo (IdFOS Sort (IdFCS IdFTRS)) (IdVOS Sort IdVTRS)

-- | 'InTermination' is the input boolean type
type InTermination = Maybe(Bool)

-- | 'InNormalization' is the input boolean type
type InNormalization = Maybe(Bool)

-- | 'InLinear' is the input boolean type
type InLinear = Bool

-- | 'InPC' is the input problem conclusions CTRS type
type InPC = ProblemConclusionsCTRS

-- | Confluence Problem - CS-TRS
data InConfCSTRS trs = InConfCSTRS trs --InCSTRS

-- | Confluence Problem - CS-CTRS
data InConfCSCTRS ctrs = InConfCSCTRS ctrs --InCSCTRS

-- | Termination - CS-TRS
data InTermCSTRS trs term = InTermCSTRS trs term --InCSTRS term

-- | Termination - CS-CTRS
data InTermCSCTRS ctrs term = InTermCSCTRS ctrs term--InCSCTRS term

-- | Normalization - CS-TRS
data InNormCSTRS trs norm = InNormCSTRS trs norm --InCSTRS norm

-- | Normalization - CS-CTRS
data InNormCSCTRS ctrs norm = InNormCSCTRS ctrs norm--InCSCTRS norm

-- | Linearity - CS-TRS
data InLinCSTRS trs lin = InLinCSTRS trs lin --InCSTRS lin

-- | Linearity - CS-CTRS
data InLinCSCTRS ctrs lin = InLinCSCTRS ctrs lin--InCSCTRS lin

-- | Critical Pair Problem - CS-TRS
data InCPCSTRS trs cps = InCPCSTRS trs [cps] --InCSTRS InCritPair

-- | Critical Pair Problem - CS-CTRS
data InCPCSCTRS ctrs cps = InCPCSCTRS ctrs [cps] --InCSCTRS InCritPair

-- | Critical Pair - Termination Problem - CS-TRS
data InCPTermCSTRS trs cps term = InCPTermCSTRS trs [cps] term --InCSTRS InCritPair term

-- | Critical Pair - Termination Problem - CS-CTRS
data InCPTermCSCTRS ctrs cps term = InCPTermCSCTRS ctrs [cps] term --InCSCTRS InCritPair term

-- | Critical Pair - Normalization Problem - CS-TRS
data InCPNormCSTRS trs cps norm = InCPNormCSTRS trs [cps] norm --InCSTRS InCritPair norm

-- | Critical Pair - Normalization Problem - CS-CTRS
data InCPNormCSCTRS ctrs cps norm = InCPNormCSCTRS ctrs [cps] norm --InCSCTRS InCritPair norm

-- | Critical Pair - Linear Problem - CS-TRS
data InCPLinCSTRS trs cps lin = InCPLinCSTRS trs [cps] lin --InCSTRS InCritPair

-- | Critical Pair - Linear Problem - CS-CTRS
data InCPLinCSCTRS ctrs cps lin = InCPLinCSCTRS ctrs [cps] lin --InCSCTRS InCritPair

-- | Critical Pair - Problem Conclusions Problem - CS-TRS
data InCPPCCSTRS trs cps pc = InCPPCCSTRS trs [cps] pc --InCSCTRS InCritPair InPC

-- | Critical Pair - Problem Conclusions Problem - CS-CTRS
data InCPPCCSCTRS ctrs cps pc = InCPPCCSCTRS ctrs [cps] pc --InCSCTRS InCritPair InPC

-- | Critical Pair - Termination - Problem Conclusions Problem - CS-TRS
data InCPTermPCCSTRS trs cps term pc = InCPTermPCCSTRS trs [cps] term pc --InCSCTRS InCritPair term InPC

-- | Critical Pair - Termination - Problem Conclusions Problem - CS-CTRS
data InCPTermPCCSCTRS ctrs cps term pc = InCPTermPCCSCTRS ctrs [cps] term pc --InCSCTRS InCritPair term InPC

-- | The data 'MProblem' describes the different types of problems.
data MProblem info where
    MRewriting       :: Problem Rewriting       (InConfCSTRS InCSTRS)                                                    -> MProblem info
    MCRewriting      :: Problem CRewriting      (InConfCSCTRS InCSCTRS)                                                  -> MProblem info
    MTRewriting      :: Problem TRewriting      (InTermCSTRS      (InConfCSTRS InCSTRS) InTermination)                   -> MProblem info
    MNRewriting      :: Problem NRewriting      (InNormCSTRS      (InConfCSTRS InCSTRS) InNormalization)                 -> MProblem info
    MLRewriting      :: Problem LRewriting      (InLinCSTRS       (InConfCSTRS InCSTRS) InLinear)                        -> MProblem info
    MCPRewriting     :: Problem CPRewriting     (InCPCSTRS        (InConfCSTRS InCSTRS) InCritPair)                      -> MProblem info
    MCPTRewriting    :: Problem CPTRewriting    (InCPTermCSTRS    (InConfCSTRS InCSTRS) InCritPair InTermination)        -> MProblem info
    MCPNRewriting    :: Problem CPNRewriting    (InCPNormCSTRS    (InConfCSTRS InCSTRS) InCritPair InNormalization)      -> MProblem info
    MCPLRewriting    :: Problem CPLRewriting    (InCPLinCSTRS     (InConfCSTRS InCSTRS) InCritPair InLinear)             -> MProblem info
    MCPPCRewriting   :: Problem CPPCRewriting   (InCPPCCSTRS      (InConfCSTRS InCSTRS) InCritPair InPC)                 -> MProblem info
    MCPTPCRewriting  :: Problem CPTPCRewriting  (InCPTermPCCSTRS  (InConfCSTRS InCSTRS) InCritPair InTermination InPC)   -> MProblem info
    MTCRewriting     :: Problem TCRewriting     (InTermCSCTRS     (InConfCSCTRS InCSCTRS) InTermination)                 -> MProblem info
    MNCRewriting     :: Problem NCRewriting     (InNormCSCTRS     (InConfCSCTRS InCSCTRS) InNormalization)               -> MProblem info
    MLCRewriting     :: Problem LCRewriting     (InLinCSCTRS      (InConfCSCTRS InCSCTRS) InLinear)                      -> MProblem info
    MCPCRewriting    :: Problem CPCRewriting    (InCPCSCTRS       (InConfCSCTRS InCSCTRS) InCritPair)                    -> MProblem info
    MCPTCRewriting   :: Problem CPTCRewriting   (InCPTermCSCTRS   (InConfCSCTRS InCSCTRS) InCritPair InTermination)      -> MProblem info
    MCPNCRewriting   :: Problem CPNCRewriting   (InCPNormCSCTRS   (InConfCSCTRS InCSCTRS) InCritPair InNormalization)    -> MProblem info
    MCPLCRewriting   :: Problem CPLCRewriting   (InCPLinCSCTRS    (InConfCSCTRS InCSCTRS) InCritPair InLinear)           -> MProblem info
    MCPPCCRewriting  :: Problem CPPCCRewriting  (InCPPCCSCTRS     (InConfCSCTRS InCSCTRS) InCritPair InPC)               -> MProblem info
    MCPTPCCRewriting :: Problem CPTPCCRewriting (InCPTermPCCSCTRS (InConfCSCTRS InCSCTRS) InCritPair InTermination InPC) -> MProblem info


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- IsConfProblem

instance IsConfProblem (InConfCSTRS InCSTRS) InCSTRS where
  getConfProblem (InConfCSTRS trs) = trs
  setConfProblem trs (InConfCSTRS _) = InConfCSTRS trs

instance IsConfProblem (InConfCSCTRS InCSCTRS) InCSCTRS where
  getConfProblem (InConfCSCTRS ctrs) = ctrs
  setConfProblem ctrs (InConfCSCTRS _) = InConfCSCTRS ctrs

instance IsConfProblem (InCPCSTRS InCSTRS InCritPair) InCSTRS where
  getConfProblem (InCPCSTRS trs _) = trs
  setConfProblem trs (InCPCSTRS _ cps) = InCPCSTRS trs cps

instance IsConfProblem (InCPCSCTRS InCSCTRS InCritPair) InCSCTRS where
  getConfProblem (InCPCSCTRS ctrs _) = ctrs
  setConfProblem ctrs (InCPCSCTRS _ cps) = InCPCSCTRS ctrs cps

instance IsConfProblem (InTermCSTRS InCSTRS InTermination) InCSTRS where
  getConfProblem (InTermCSTRS trs _) = trs
  setConfProblem trs (InTermCSTRS _ term) = InTermCSTRS trs term

instance IsConfProblem (InTermCSCTRS InCSCTRS InTermination) InCSCTRS where
  getConfProblem (InTermCSCTRS ctrs _) = ctrs
  setConfProblem ctrs (InTermCSCTRS _ term) = InTermCSCTRS ctrs term

instance IsConfProblem (InNormCSTRS InCSTRS InNormalization) InCSTRS where
  getConfProblem (InNormCSTRS trs _) = trs
  setConfProblem trs (InNormCSTRS _ norm) = InNormCSTRS trs norm

instance IsConfProblem (InNormCSCTRS InCSCTRS InNormalization) InCSCTRS where
  getConfProblem (InNormCSCTRS ctrs _) = ctrs
  setConfProblem ctrs (InNormCSCTRS _ norm) = InNormCSCTRS ctrs norm

instance IsConfProblem (InLinCSTRS InCSTRS InLinear) InCSTRS where
  getConfProblem (InLinCSTRS trs _) = trs
  setConfProblem trs (InLinCSTRS _ lin) = InLinCSTRS trs lin

instance IsConfProblem (InLinCSCTRS InCSCTRS InLinear) InCSCTRS where
  getConfProblem (InLinCSCTRS ctrs _) = ctrs
  setConfProblem ctrs (InLinCSCTRS _ lin) = InLinCSCTRS ctrs lin

instance IsConfProblem (InCPTermCSTRS InCSTRS InCritPair InTermination) InCSTRS where
  getConfProblem (InCPTermCSTRS trs _ _) = trs
  setConfProblem trs (InCPTermCSTRS _ cps term) = InCPTermCSTRS trs cps term

instance IsConfProblem (InCPTermCSCTRS InCSCTRS InCritPair InTermination) InCSCTRS where
   getConfProblem (InCPTermCSCTRS ctrs _ _) = ctrs
   setConfProblem ctrs (InCPTermCSCTRS _ cps term) = InCPTermCSCTRS ctrs cps term

instance IsConfProblem (InCPNormCSTRS InCSTRS InCritPair InNormalization) InCSTRS where
  getConfProblem (InCPNormCSTRS trs _ _) = trs
  setConfProblem trs (InCPNormCSTRS _ cps norm) = InCPNormCSTRS trs cps norm

instance IsConfProblem (InCPNormCSCTRS InCSCTRS InCritPair InNormalization) InCSCTRS where
   getConfProblem (InCPNormCSCTRS ctrs _ _) = ctrs
   setConfProblem ctrs (InCPNormCSCTRS _ cps norm) = InCPNormCSCTRS ctrs cps norm

instance IsConfProblem (InCPLinCSTRS InCSTRS InCritPair InLinear) InCSTRS where
  getConfProblem (InCPLinCSTRS trs _ _) = trs
  setConfProblem trs (InCPLinCSTRS _ cps lin) = InCPLinCSTRS trs cps lin

instance IsConfProblem (InCPLinCSCTRS InCSCTRS InCritPair InLinear) InCSCTRS where
   getConfProblem (InCPLinCSCTRS ctrs _ _) = ctrs
   setConfProblem ctrs (InCPLinCSCTRS _ cps lin) = InCPLinCSCTRS ctrs cps lin

instance IsConfProblem (InCPPCCSTRS InCSTRS InCritPair InPC) InCSTRS where
   getConfProblem (InCPPCCSTRS trs _ _) = trs
   setConfProblem trs (InCPPCCSTRS _ cps pc) = InCPPCCSTRS trs cps pc

instance IsConfProblem (InCPPCCSCTRS InCSCTRS InCritPair InPC) InCSCTRS where
   getConfProblem (InCPPCCSCTRS ctrs _ _) = ctrs
   setConfProblem ctrs (InCPPCCSCTRS _ cps pc) = InCPPCCSCTRS ctrs cps pc

instance IsConfProblem (InCPTermPCCSTRS InCSTRS InCritPair InTermination InPC) InCSTRS where
   getConfProblem (InCPTermPCCSTRS trs _ _ _) = trs
   setConfProblem trs (InCPTermPCCSTRS _ cps term pc) = InCPTermPCCSTRS trs cps term pc

instance IsConfProblem (InCPTermPCCSCTRS InCSCTRS InCritPair InTermination InPC) InCSCTRS where
   getConfProblem (InCPTermPCCSCTRS ctrs _ _ _) = ctrs
   setConfProblem ctrs (InCPTermPCCSCTRS _ cps term pc) = InCPTermPCCSCTRS ctrs cps term pc

-- IsCPProblem

instance IsCPProblem (InCPCSTRS InCSTRS InCritPair) InCritPair where
  getCPairs (InCPCSTRS _ cps) = cps
  setCPairs cps (InCPCSTRS trs _) = InCPCSTRS trs cps

instance IsCPProblem (InCPCSCTRS InCSCTRS InCritPair) InCritPair where
  getCPairs (InCPCSCTRS _ cps) = cps
  setCPairs cps (InCPCSCTRS ctrs _) = InCPCSCTRS ctrs cps

-- IsTProblem

instance IsTProblem (InTermCSTRS InCSTRS InTermination) InTermination where
  getTermination (InTermCSTRS _ termination) = termination
  setTermination termination (InTermCSTRS trs _) = InTermCSTRS trs termination

instance IsTProblem (InTermCSCTRS InCSCTRS InTermination) InTermination where
  getTermination (InTermCSCTRS _ termination) = termination
  setTermination termination (InTermCSCTRS ctrs _) = InTermCSCTRS ctrs termination

-- IsNProblem

instance IsNProblem (InNormCSTRS InCSTRS InNormalization) InNormalization where
  getNormalization (InNormCSTRS _ normalization) = normalization
  setNormalization normalization (InNormCSTRS trs _) = InNormCSTRS trs normalization

instance IsNProblem (InNormCSCTRS InCSCTRS InNormalization) InNormalization where
  getNormalization (InNormCSCTRS _ normalization) = normalization
  setNormalization normalization (InNormCSCTRS ctrs _) = InNormCSCTRS ctrs normalization

-- IsLProblem

instance IsLProblem (InLinCSTRS InCSTRS InLinear) InLinear where
  getLinearity (InLinCSTRS _ lin) = lin
  setLinearity lin (InLinCSTRS trs _) = InLinCSTRS trs lin

instance IsLProblem (InLinCSCTRS InCSCTRS InLinear) InLinear where
  getLinearity (InLinCSCTRS _ lin) = lin
  setLinearity lin (InLinCSCTRS ctrs _) = InLinCSCTRS ctrs lin

-- IsCPTermProblem

instance IsCPTermProblem (InCPTermCSTRS InCSTRS InCritPair InTermination) InCritPair InTermination where
  getTermCPT (InCPTermCSTRS _ _ termination) = termination
  setTermCPT termination (InCPTermCSTRS trs cps _) = InCPTermCSTRS trs cps termination
  getCPairsCPT (InCPTermCSTRS _ cps _) = cps
  setCPairsCPT cps (InCPTermCSTRS trs _ termination) = InCPTermCSTRS trs cps termination

instance IsCPTermProblem (InCPTermCSCTRS InCSCTRS InCritPair InTermination) InCritPair InTermination where
  getTermCPT (InCPTermCSCTRS _ _ termination) = termination
  setTermCPT termination (InCPTermCSCTRS ctrs cps _) = InCPTermCSCTRS ctrs cps termination
  getCPairsCPT (InCPTermCSCTRS _ cps _) = cps
  setCPairsCPT cps (InCPTermCSCTRS ctrs _ termination) = InCPTermCSCTRS ctrs cps termination

-- IsCPNormProblem

instance IsCPNormProblem (InCPNormCSTRS InCSTRS InCritPair InNormalization) InCritPair InNormalization where
  getNormCPN (InCPNormCSTRS _ _ normalization) = normalization
  setNormCPN normalization (InCPNormCSTRS trs cps _) = InCPNormCSTRS trs cps normalization
  getCPairsCPN (InCPNormCSTRS _ cps _) = cps
  setCPairsCPN cps (InCPNormCSTRS trs _ normalization) = InCPNormCSTRS trs cps normalization

instance IsCPNormProblem (InCPNormCSCTRS InCSCTRS InCritPair InNormalization) InCritPair InNormalization where
  getNormCPN (InCPNormCSCTRS _ _ normalization) = normalization
  setNormCPN normalization (InCPNormCSCTRS ctrs cps _) = InCPNormCSCTRS ctrs cps normalization
  getCPairsCPN (InCPNormCSCTRS _ cps _) = cps
  setCPairsCPN cps (InCPNormCSCTRS ctrs _ normalization) = InCPNormCSCTRS ctrs cps normalization

-- IsCPLinProblem

instance IsCPLinProblem (InCPLinCSTRS InCSTRS InCritPair InLinear) InCritPair InLinear where
  getLinCPL (InCPLinCSTRS _ _ linear) = linear
  setLinCPL linear (InCPLinCSTRS trs cps _) = InCPLinCSTRS trs cps linear
  getCPairsCPL (InCPLinCSTRS _ cps _) = cps
  setCPairsCPL cps (InCPLinCSTRS trs _ linear) = InCPLinCSTRS trs cps linear

instance IsCPLinProblem (InCPLinCSCTRS InCSCTRS InCritPair InLinear) InCritPair InLinear where
  getLinCPL (InCPLinCSCTRS _ _ linear) = linear
  setLinCPL linear (InCPLinCSCTRS ctrs cps _) = InCPLinCSCTRS ctrs cps linear
  getCPairsCPL (InCPLinCSCTRS _ cps _) = cps
  setCPairsCPL cps (InCPLinCSCTRS ctrs _ linear) = InCPLinCSCTRS ctrs cps linear

-- IsCPPCProblem

instance IsCPPCProblem (InCPPCCSTRS InCSTRS InCritPair InPC) InCritPair InPC where
  getCPairsCPPC (InCPPCCSTRS _ cps _) = cps
  setCPairsCPPC cps (InCPPCCSTRS trs _ pc) = InCPPCCSTRS trs cps pc
  getProbConcCPPC (InCPPCCSTRS _ _ pc) = pc
  setProbConcCPPC pc (InCPPCCSTRS trs cps _) = InCPPCCSTRS trs cps pc

instance IsCPPCProblem (InCPPCCSCTRS InCSCTRS InCritPair InPC) InCritPair InPC where
  getCPairsCPPC (InCPPCCSCTRS _ cps _) = cps
  setCPairsCPPC cps (InCPPCCSCTRS ctrs _ pc) = InCPPCCSCTRS ctrs cps pc
  getProbConcCPPC (InCPPCCSCTRS _ _ pc) = pc
  setProbConcCPPC pc (InCPPCCSCTRS ctrs cps _) = InCPPCCSCTRS ctrs cps pc

-- IsCPTermPCProblem

instance IsCPTermPCProblem (InCPTermPCCSTRS InCSTRS InCritPair InTermination InPC) InCritPair InTermination InPC where
  getTermCPTPC (InCPTermPCCSTRS _ _ termination _) = termination
  setTermCPTPC termination (InCPTermPCCSTRS trs cps _ pc) = InCPTermPCCSTRS trs cps termination pc
  getCPairsCPTPC (InCPTermPCCSTRS _ cps _ _) = cps
  setCPairsCPTPC cps (InCPTermPCCSTRS trs _ termination pc) = InCPTermPCCSTRS trs cps termination pc
  getProbConcCPTPC (InCPTermPCCSTRS _ _ _ pc) = pc
  setProbConcCPTPC pc (InCPTermPCCSTRS trs cps termination _) = InCPTermPCCSTRS trs cps termination pc

instance IsCPTermPCProblem (InCPTermPCCSCTRS InCSCTRS InCritPair InTermination InPC) InCritPair InTermination InPC where
  getTermCPTPC (InCPTermPCCSCTRS _ _ termination _) = termination
  setTermCPTPC termination (InCPTermPCCSCTRS ctrs cps _ pc) = InCPTermPCCSCTRS ctrs cps termination pc
  getCPairsCPTPC (InCPTermPCCSCTRS _ cps _ _) = cps
  setCPairsCPTPC cps (InCPTermPCCSCTRS ctrs _ termination pc) = InCPTermPCCSCTRS ctrs cps termination pc
  getProbConcCPTPC (InCPTermPCCSCTRS _ _ _ pc) = pc
  setProbConcCPTPC pc (InCPTermPCCSCTRS ctrs cps termination _) = InCPTermPCCSCTRS ctrs cps termination pc


-- Pretty

instance Pretty (InConfCSTRS InCSTRS) where
  pPrint (InConfCSTRS trs) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance Pretty (InConfCSCTRS InCSCTRS) where
  pPrint (InConfCSCTRS ctrs) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps) => Pretty (InCPCSTRS InCSTRS cps) where
  pPrint (InCPCSTRS trs cps) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps) => Pretty (InCPCSCTRS InCSCTRS cps) where
  pPrint (InCPCSCTRS ctrs cps) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty InTermination) => Pretty (InTermCSTRS InCSTRS InTermination) where
  pPrint (InTermCSTRS trs term) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nTerminating:" HPJC.<> pPrint (if isJust term then if fromJust term then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty InTermination) => Pretty (InTermCSCTRS InCSCTRS InTermination) where
  pPrint (InTermCSCTRS ctrs term) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nTerminating:" HPJC.<> pPrint (if isJust term then if fromJust term then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty InNormalization) => Pretty (InNormCSTRS InCSTRS InNormalization) where
  pPrint (InNormCSTRS trs norm) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nNormalizating:" HPJC.<> pPrint (if isJust norm then if fromJust norm then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty InNormalization) => Pretty (InNormCSCTRS InCSCTRS InNormalization) where
  pPrint (InNormCSCTRS ctrs norm) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nNormalizating:" HPJC.<> pPrint (if isJust norm then if fromJust norm then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty InLinear) => Pretty (InLinCSTRS InCSTRS InLinear) where
  pPrint (InLinCSTRS trs lin) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nLinear:" HPJC.<> text (if lin then "YES" else "NO")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty InLinear) => Pretty (InLinCSCTRS InCSCTRS InLinear) where
  pPrint (InLinCSCTRS ctrs lin) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nLinear:" HPJC.<> text (if lin then "YES" else "NO")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InTermination) => Pretty (InCPTermCSTRS InCSTRS cps InTermination) where
  pPrint (InCPTermCSTRS trs cps term) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nTerminating:" HPJC.<> pPrint (if isJust term then if fromJust term then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InTermination) => Pretty (InCPTermCSCTRS InCSCTRS cps InTermination) where
  pPrint (InCPTermCSCTRS ctrs cps term) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nTerminating:" HPJC.<> pPrint (if isJust term then if fromJust term then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InNormalization) => Pretty (InCPNormCSTRS InCSTRS cps InNormalization) where
  pPrint (InCPNormCSTRS trs cps norm) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nNormalizating:" HPJC.<> pPrint (if isJust norm then if fromJust norm then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InNormalization) => Pretty (InCPNormCSCTRS InCSCTRS cps InNormalization) where
  pPrint (InCPNormCSCTRS ctrs cps norm) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nNormalizating:" HPJC.<> pPrint (if isJust norm then if fromJust norm then "YES" else "NO" else "MAYBE")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InLinear) => Pretty (InCPLinCSTRS InCSTRS cps InLinear) where
  pPrint (InCPLinCSTRS trs cps lin) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nLinear:" HPJC.<> text (if lin then "YES" else "NO")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InLinear) => Pretty (InCPLinCSCTRS InCSCTRS cps InLinear) where
  pPrint (InCPLinCSCTRS ctrs cps lin) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nLinear:" HPJC.<> text (if lin then "YES" else "NO")
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InPC) => Pretty (InCPPCCSTRS InCSTRS cps InPC) where
  pPrint (InCPPCCSTRS trs cps pc) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> pPrint pc
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InPC) => Pretty (InCPPCCSCTRS InCSCTRS cps InPC) where
  pPrint (InCPPCCSCTRS ctrs cps pc) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> pPrint pc
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InTermination, Pretty InPC) => Pretty (InCPTermPCCSTRS InCSTRS cps InTermination InPC) where
  pPrint (InCPTermPCCSTRS trs cps term pc) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS trs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nTerminating:" HPJC.<> text (if isJust term then if fromJust term then "YES" else "NO" else "MAYBE")
    HPJC.<> pPrint pc
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"

instance (Pretty cps, Pretty InTermination, Pretty InPC) => Pretty (InCPTermPCCSCTRS InCSCTRS cps InTermination InPC) where
  pPrint (InCPTermPCCSCTRS ctrs cps term pc) = text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
    HPJC.<> text "Confluence Problem:\n" HPJC.<> pPrintCOPS ctrs 
    HPJC.<> text "\nCritical Pairs:\n" HPJC.<> (hcat . punctuate (char '\n') . map (\c -> pPrint c) $ cps)
    HPJC.<> text "\nTerminating:" HPJC.<> text (if isJust term then if fromJust term then "YES" else "NO" else "MAYBE")
    HPJC.<> pPrint pc
    HPJC.<> text "\n:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::\n"
