{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleContexts #-}

-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.ProblemConclusions
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module has data and functions for problem conclusions
--
-----------------------------------------------------------------------------
module Framework.ProblemConclusions (

-- * Exported data

ProblemConclusionsTRS (..), ProblemConclusionsCTRS (..),

-- * Exported functions

problemConclusionTRS, problemConclusionCTRS, isOriented, isJoin, isSemiequational

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..), isLinear, vars, isVTerm)
import TRS.Rules as R (Rule(..), CRule(..), Cond (..), lhs, rhs, rule, conds, lcond, rcond, condVars, deterministic, 
                       orientedConds)
import TRS.Identifiers (HasName (..), HasMu (..), HasArity (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..))
import Framework.Problems (CriticalPairInfo (..)) 
import Framework.TRSAux (isLHRV, isCSTRSSymb)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, braces, brackets, parens, hsep)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty, (\\), member, map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

data ProblemConclusionsTRS =
  ProblemConclusionsTRS { leftLinearTRS       :: Bool
                        , rightLinearTRS      :: Bool
                        , linearTRS           :: Bool
                        , weaklyOrthogonalTRS :: Bool
                        , almostOrthogonalTRS :: Bool
                        , orthogonalTRS       :: Bool
                        , huetLevyConfluent   :: Bool
                        , newman              :: Bool
                        , isCSTRS             :: Bool
                        , leftHomogMuRepVars  :: Maybe Bool
                        }


data ProblemConclusionsCTRS =
  ProblemConclusionsCTRS { leftLinearCTRS       :: Bool
                         , rightLinearCTRS      :: Bool
                         , linearCTRS           :: Bool
                         , weaklyOrthogonalCTRS :: Bool
                         , almostOrthogonalCTRS :: Bool
                         , orthogonalCTRS       :: Bool
                         , ctrsType             :: Int
                         , determ               :: Bool
                         , stronglyDet          :: Maybe Bool
                         , oriented             :: Bool
                         , properlyOriented     :: Bool
                         , join                 :: Bool
                         , semiequational       :: Bool
                         , overlayCTRS          :: Bool
                         , rightStable          :: Maybe Bool
                         , normalCTRS           :: Maybe Bool
                         , almostNormalCTRS     :: Maybe Bool
                         , terminatingCTRS      :: Maybe Bool
                         , opTerminatingCTRS    :: Maybe Bool
                         , joinableCTRS         :: Maybe Bool
                         , levelConfluent       :: Bool
                         , confluentCTRS        :: Bool
                         }

-- Pretty

instance Pretty ProblemConclusionsTRS where
  pPrint (ProblemConclusionsTRS
    { leftLinearTRS = ll
    , rightLinearTRS = rl
    , linearTRS = l
    , weaklyOrthogonalTRS = wo
    , almostOrthogonalTRS = ao
    , orthogonalTRS = o
    , huetLevyConfluent = hlc
    , newman = n
    , isCSTRS = cstrs
    , leftHomogMuRepVars = lhrv
    }) = text "\n\n-> Problem conclusions:\n"
          Doc.<+> text (if ll then "Left linear" else "Not left linear") Doc.<> comma
          Doc.<+> text (if rl then "Right linear" else "Not right linear") Doc.<> comma
          Doc.<+> text (if l then "Linear" else "Not linear") Doc.<> text "\n"
          Doc.<+> text (if wo then "Weakly orthogonal" else "Not weakly orthogonal") Doc.<> comma
          Doc.<+> text (if ao then "Almost orthogonal" else "Not almost orthogonal") Doc.<> comma
          Doc.<+> text (if o then "Orthogonal" else "Not orthogonal") Doc.<> text "\n"
          Doc.<+> text (if hlc then "Huet-Levy confluent" else "Not Huet-Levy confluent") Doc.<> comma
          Doc.<+> text (if n then "Newman confluent" else "Not Newman confluent") Doc.<> text "\n"
          Doc.<+> text (if cstrs then "R is a CS-TRS," else "R is a TRS")
          Doc.<+> text (if isNothing lhrv then "" else if isTrue lhrv then "Left-homogeneous u-replacing variables" else "Not left-homogeneous u-replacing variables")



instance Pretty ProblemConclusionsCTRS where
  pPrint (ProblemConclusionsCTRS
    { leftLinearCTRS = ll
    , rightLinearCTRS = rl
    , linearCTRS = l
    , weaklyOrthogonalCTRS = wo
    , almostOrthogonalCTRS = ao
    , orthogonalCTRS = o
    , ctrsType = ct
    , determ = d
    , stronglyDet = sd
    , oriented = or
    , properlyOriented = por
    , join = j
    , semiequational = seq
    , overlayCTRS = ov
    , rightStable = rs
    , normalCTRS = nc
    , almostNormalCTRS = anc
    , terminatingCTRS = t
    , opTerminatingCTRS = opt
    , joinableCTRS = jnb
    , levelConfluent = lc
    , confluentCTRS = cf
    }) = text "\n\n-> Problem conclusions:\n"
    Doc.<+> text (if ll then "Left linear" else "Not left linear") Doc.<> comma
    Doc.<+> text (if rl then "Right linear" else "Not right linear") Doc.<> comma
    Doc.<+> text (if l then "Linear" else "Not linear") Doc.<> text "\n"
    Doc.<+> text (if wo then "Weakly orthogonal" else "Not weakly orthogonal") Doc.<> comma
    Doc.<+> text (if ao then "Almost orthogonal" else "Not almost orthogonal") Doc.<> comma
    Doc.<+> text (if o then "Orthogonal" else "Not orthogonal") Doc.<> text "\n"
    Doc.<+> text "CTRS Type:" Doc.<+> pPrint ct Doc.<> text "\n"
    Doc.<+> text (if d then "Deterministic" else "Not deterministic") Doc.<> comma
    Doc.<+> text (if isNothing sd then "Maybe strongly deterministic" else if (isTrue sd) then "Strongly deterministic" else "Not strongly deterministic") Doc.<> text "\n"
    Doc.<+> text (if or then "Oriented CTRS" else "Not oriented CTRS") Doc.<> comma
    Doc.<+> text (if por then "Properly oriented CTRS" else "Not properly oriented CTRS") Doc.<> comma
    Doc.<+> text (if j then "Join CTRS" else "Not join CTRS") Doc.<> comma
    Doc.<+> text (if seq then "Semiequational CTRS" else "Not semiequational CTRS") Doc.<> text "\n"
    Doc.<+> text (if isNothing rs then "Maybe right-stable CTRS" else if (isTrue rs) then "Right-stable CTRS" else "Not right-stable CTRS") Doc.<> comma
    Doc.<+> text (if ov then "Overlay CTRS" else "Not overlay CTRS") Doc.<> text "\n"
    Doc.<+> text (if isNothing nc then "Maybe normal CTRS" else if (isTrue nc) then "Normal CTRS" else "Not normal CTRS") Doc.<> comma
    Doc.<+> text (if isNothing anc then "Maybe almost normal CTRS" else if (isTrue anc) then "Almost normal CTRS" else "Not almost normal CTRS") Doc.<> text "\n"
    Doc.<+> text (if isNothing t then "Maybe terminating CTRS" else if (isTrue t) then "Terminating CTRS" else "Not terminating CTRS") Doc.<> comma
    Doc.<+> text (if isNothing opt then "Maybe operational terminating CTRS" else if (isTrue opt) then "Operational terminating CTRS" else "Not operational terminating CTRS") Doc.<> comma
    Doc.<+> text (if isNothing jnb then "Maybe joinable CCPs" else if (isTrue jnb) then "Joinable CCPs" else "Not joinable CCPs") Doc.<> text "\n"
    Doc.<+> text (if lc then "Level confluent" else "Maybe level confluent") Doc.<> text "\n"
    Doc.<+> text (if cf then "Confluent" else "Maybe confluent")




-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

problemConclusionCTRS :: (HasSignature ctrs idF idV, HasCRules ctrs (Term idF idV), Ord idV, Ord idF) => 
  ctrs -> [CriticalPairInfo idF idV] -> Maybe Bool -> Maybe Bool -> Maybe Bool -> Maybe Bool -> Maybe Bool -> ProblemConclusionsCTRS
problemConclusionCTRS ctrs [] normtjs rightSt t opt joinable
  = ProblemConclusionsCTRS ll rl l weakOrt almOrt ort ty det stDet orient propOrient j seq ov rightSt nor alNor term opt joinable levelcf cf
  where ll = isLeftLinearCTRS ctrs
        rl = isRightLinearCTRS ctrs
        l = ll && rl
        ort = ll
        almOrt = ort
        weakOrt = almOrt
        ov = True
        ty = getTypeCTRS ctrs
        crs = elems $ getCTRSRules ctrs
        constructors = getTRSCSymbols ctrs
        orient = isOriented crs
        propOrient = isProperlyOriented crs
        --propOrient
        --  | orient = isProperlyOriented crs
        --  | otherwise = False
        j = isJoin crs
        seq = isSemiequational crs
        nor = isNormalCTRS crs normtjs
        alNor = isAlmostNormal orient rightSt nor
        det = and $ Prelude.map deterministic crs
        stDet = isStronglyDeterministic crs constructors ty det nor
        term
          | isTrue opt = Just True
          | otherwise = t
        levelcf = (((ty <= 2) && almOrt && (isTrue alNor)) || ((ty <= 3) && ort && propOrient && (isTrue rightSt)))
        cf = (levelcf 
          || ((isTrue opt) && (isTrue stDet)) -- ((isTrue term) && (isTrue nor)) || (ort && (isTrue opt) && ((isTrue stDet) || (isTrue alNor)))
          || ((isTrue term) && (isTrue nor))
          || ((isTrue term) && j && (ty <= 1)))
problemConclusionCTRS ctrs cpis normtjs rightSt t opt joinable
  = ProblemConclusionsCTRS ll rl l weakOrt almOrt False ty det stDet orient propOrient j seq ov rightSt nor alNor term opt joinable levelcf cf
  where weakOrt = ll && (and (Prelude.map trivial cpis))
        almOrt = weakOrt && (and (Prelude.map overlay cpis))
        ll = isLeftLinearCTRS ctrs
        rl = isRightLinearCTRS ctrs
        l = ll && rl
        ov = and $ Prelude.map overlay cpis
        ty = getTypeCTRS ctrs
        crs = elems $ getCTRSRules ctrs
        constructors = getTRSCSymbols ctrs
        orient = isOriented crs
        propOrient = isProperlyOriented crs
        --propOrient
        --  | orient = isProperlyOriented crs
        --  | otherwise = False
        j = isJoin crs
        seq = isSemiequational crs
        nor = isNormalCTRS crs normtjs
        alNor = isAlmostNormal orient rightSt nor
        det = and $ Prelude.map deterministic crs
        stDet = isStronglyDeterministic crs constructors ty det nor
        term
          | isTrue opt = Just True
          | otherwise = t
        levelcf = (ty <= 2) && almOrt && (isTrue alNor)
        cf = (levelcf 
          || ((isTrue term) && ov && (isTrue nor) && (isTrue joinable))
          || ((isTrue opt) && (isTrue stDet) && (isTrue joinable))
          || ((isTrue term) && (isTrue nor) && (ty <= 1) && (isTrue joinable))
          || ((isTrue term) && j && (ty <= 1) && (isTrue joinable)))


problemConclusionTRS :: (HasRules trs (Term idF idV), HasSignature trs idF idV, HasMu idF, HasArity idF, Ord idV) => 
  trs -> [CriticalPairInfo idF idV] -> Maybe Bool -> ProblemConclusionsTRS
problemConclusionTRS trs [] termination = ProblemConclusionsTRS ll rl l ll ll ll hlc nm cstrs lhrv
  where ll = isLeftLinearTRS trs
        rl = isRightLinearTRS trs
        l = ll && rl
        weakOrt = ll
        almOrt = weakOrt
        hlc = weakOrt
        nm = isTrue termination
        cstrs = or $ S.map isCSTRSSymb (getTRSSymbols trs)
        rs = elems $ getTRSRules trs
        lhrv = if cstrs then Just (and $ Prelude.map isLHRV rs) else Nothing
problemConclusionTRS trs cpis termination = ProblemConclusionsTRS ll rl l weakOrt almOrt False hlc nm cstrs lhrv
  where weakOrt = ll && (and (Prelude.map trivial cpis))
        almOrt = weakOrt && (and (Prelude.map overlay cpis))
        ll = isLeftLinearTRS trs
        rl = isRightLinearTRS trs
        l = ll && rl
        hlc = weakOrt
        nm = False
        cstrs = or $ S.map isCSTRSSymb (getTRSSymbols trs)
        rs = elems $ getTRSRules trs
        lhrv = if cstrs then Just (and $ Prelude.map isLHRV rs) else Nothing


getTypeCTRS :: (HasCRules trs (Term idF idV), Ord idV, Ord idF) => trs -> Int
getTypeCTRS trs
  | and $ Prelude.map (<=1) rtypes = 1
  | and $ Prelude.map (<=2) rtypes = 2
  | and $ Prelude.map (<=3) rtypes = 3
  | otherwise = 4
  where crs = elems $ getCTRSRules trs
        rtypes = Prelude.map getTypeCRule crs


getTypeCRule :: (Ord idV, Ord idF) => CRule (Term idF idV) -> Int
getTypeCRule crl
  | (union rvs cvs) `isSubsetOf` lvs = 1
  | rvs `isSubsetOf` lvs = 2
  | rvs `isSubsetOf` (union lvs cvs) = 3
  | otherwise = 4
  where rl = R.rule crl
        rvs = vars $ rhs rl
        lvs = vars $ lhs rl
        cvs = foldl S.union S.empty . Prelude.map condVars . conds $ crl


isStronglyDeterministic :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> Set idF -> Int -> Bool -> Maybe Bool -> Maybe Bool
isStronglyDeterministic crs constructors ty det Nothing
  | ty <= 3 && det && (va || constr) = Just True
  | otherwise = Nothing
  where cs = concatMap conds crs
        ts = Prelude.map rcond cs
        va = and $ Prelude.map isVTerm ts
        constr = areConstructors constructors ts
isStronglyDeterministic crs constructors ty det (Just grNorm)
  | ty <= 3 && det && (grNorm || va || constr) = Just True
  | otherwise = Just False
  where cs = concatMap conds crs
        ts = Prelude.map rcond cs
        va = and $ Prelude.map isVTerm ts
        constr = areConstructors constructors ts


areConstructors :: (Ord idV, Ord idF) => Set idF -> [Term idF idV] -> Bool
areConstructors _ [] = True
areConstructors constr (t:ts)
  | isConstructor constr t = areConstructors constr ts
  | otherwise = False


isConstructor :: (Ord idV, Ord idF) => Set idF -> Term idF idV -> Bool
isConstructor constr t@(F f tt) = f `member` constr
isConstructor constr _ = False


isNormalCTRS :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> Maybe Bool -> Maybe Bool
isNormalCTRS [] _ = Just True
isNormalCTRS crs Nothing
  | maybeNorm = Nothing
  | otherwise = Just True
  where cs = concatMap conds crs
        ts = Prelude.map rcond cs
        gr = and $ Prelude.map isGroundTerm ts
        maybeNorm = not (gr && (Prelude.null cs))
isNormalCTRS crs (Just normtjs) = Just (gr && normtjs)
  where cs = concatMap conds crs
        ts = Prelude.map rcond cs
        gr = and $ Prelude.map isGroundTerm ts


isAlmostNormal :: Bool -> Maybe Bool -> Maybe Bool -> Maybe Bool
isAlmostNormal _ Nothing Nothing = Nothing
isAlmostNormal o (Just rs) Nothing
  | o && rs = Just True
  | otherwise = Nothing
isAlmostNormal _ Nothing (Just nor) = Just nor
isAlmostNormal o (Just rs) (Just nor) = Just (nor || (o && rs))


isOriented :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> Bool
isOriented [] = True
isOriented (c:crs)
  | orientedConds c = isOriented crs
  | otherwise = False


isJoin :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> Bool
isJoin [] = True
isJoin (c:crs)
  | joinConds c = isJoin crs
  | otherwise = False
  where
    joinConds x = and . Prelude.map joinCond $ conds x
    joinCond (_ :->*<- _) = True
    joinCond _ = False


isSemiequational :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> Bool
isSemiequational [] = True
isSemiequational (c:crs)
  | semieqConds c = isSemiequational crs
  | otherwise = False
  where
    semieqConds x = and . Prelude.map semieqCond $ conds x
    semieqCond (_ :<-->* _) = True
    semieqCond _ = False


isProperlyOriented :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> Bool
isProperlyOriented [] = True
isProperlyOriented (c:crs)
  | subsCond = isProperlyOriented crs
  | n == 0 = False
  | poCond = isProperlyOriented crs
  | otherwise = False
  where cds = conds c
        n = length cds
        (l :-> r) = R.rule c
        lvars = vars l
        rvars = vars r
        subsCond = rvars `isSubsetOf` lvars
        poCond = and [(vars $ lcond (cds!!i)) `isSubsetOf` (union lvars (condsTVarsSetTo cds i)) | i <- [0..(n-1)]]


condsTVarsSetTo :: (Ord idV, Ord idF) => [Cond (Term idF idV)] -> Int -> Set idV
condsTVarsSetTo [] _ = S.empty
condsTVarsSetTo cs i
  | i > 0 = union (vars $ rcond (cs!!j)) (condsTVarsSetTo cs j)
  | otherwise = S.empty
  where j = i - 1


isRightLinearTRS :: (HasRules ctrs (Term idF idV), Ord idV) => ctrs -> Bool
isRightLinearTRS ctrs = and (S.map isRuleRightLinear $ getTRSRules ctrs)

isRightLinearCTRS :: (HasCRules ctrs (Term idF idV), Ord idV, Ord idF) => ctrs -> Bool
isRightLinearCTRS ctrs = and (S.map isRuleRightLinear (S.map R.rule $ getCTRSRules ctrs))

isRuleRightLinear :: Ord idV => Rule (Term idF idV) -> Bool
isRuleRightLinear (l :-> r) = isLinear r


isLeftLinearTRS :: (HasRules trs (Term idF idV), Ord idV) => trs -> Bool
isLeftLinearTRS trs = and (S.map isRuleLeftLinear $ getTRSRules trs)

isLeftLinearCTRS :: (HasCRules ctrs (Term idF idV), Ord idV, Ord idF) => ctrs -> Bool
isLeftLinearCTRS ctrs = and (S.map isRuleLeftLinear (S.map R.rule $ getCTRSRules ctrs))

isRuleLeftLinear :: Ord idV => Rule (Term idF idV) -> Bool
isRuleLeftLinear (l :-> r) = isLinear l


isTRSLinear :: (HasRules trs (Term idF idV), Ord idV) => trs -> Bool
isTRSLinear trs = (isLeftLinearTRS trs) && (isRightLinearTRS trs)

isCTRSLinear :: (HasCRules ctrs (Term idF idV), Ord idV, Ord idF) => ctrs -> Bool
isCTRSLinear ctrs = (isLeftLinearCTRS ctrs) && (isRightLinearCTRS ctrs)


isGroundTerm :: Ord idV => Term idF idV -> Bool
isGroundTerm t = (length $ vars t) == 0