{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE FunctionalDependencies #-}
--{-# LANGUAGE AllowAmbiguousTypes #-} -- COMMENT
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Problems
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
--
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable
--
-- This module contains the different muterm problems with its type.
--
-----------------------------------------------------------------------------

module Framework.Problems (

-- * Exported classes

Minimality (..), PScheme, HasPScheme (..), IsConfProblem (..), IsCPProblem (..),
CriticalPair (..), IsTProblem (..), IsCPTermProblem (..), IsCPLinProblem (..),
IsLProblem (..), CriticalPairInfo (..), IsCPTermPCProblem (..), IsNProblem (..),
IsCPNormProblem (..), IsCPPCProblem (..)

-- * Exported functions

, getPSchemeFromProblem

) where

import MuTerm.Framework.Problem (IsProblem (..), MkProblem (..))
import TRS.Terms (Term)
import TRS.Rules (Cond (..))

import Data.Maybe (isJust, fromJust, isNothing)
import Text.PrettyPrint.HughesPJClass as HPJC (Pretty (..), (<>), (<+>), comma, char, text, hcat, punctuate, hsep, braces)

-----------------------------------------------------------------------------
-- Classes
-----------------------------------------------------------------------------

-- | Minimality flag
data Minimality  = M | A | Black | Undefined deriving (Eq, Ord, Show)

-- | Critical Pair
data CriticalPair idF idV =
  CriticalPair { s          :: Term idF idV
               , t          :: Term idF idV
               , cpPos      :: [Int]
               , ccpConds   :: [Cond (Term idF idV)]
               , normWeight :: Int
               }
  deriving (Eq, Ord)

data CriticalPairInfo idF idV =
  CriticalPairInfo { trivial    :: Bool
                   , overlay    :: Bool
                   , critPair   :: CriticalPair idF idV
                   , cpNum      :: Int
                   , hasFeasCds :: Maybe Bool
                   }



-- | Termination Scheme
type PScheme  = (Minimality,Minimality)

-- | Add Problem Schemes to Problems
class IsProblem typ => HasPScheme typ where
    getPScheme :: typ -> PScheme
    setPScheme :: PScheme -> Problem typ trs -> Problem typ trs

-- | Obtains the plugging scheme
getPSchemeFromProblem :: HasPScheme typ => Problem typ trs -> PScheme
getPSchemeFromProblem = getPScheme . getProblemType

-- | Confluence Problems are the systems
class IsConfProblem typ trs | typ -> trs  where
    getConfProblem :: typ -> trs
    setConfProblem :: trs -> typ -> typ

-- | Conditional Critical Pair Problems have a list of critical pairs
class IsCPProblem typ cps | typ -> cps  where
    getCPairs :: typ -> [cps]
    setCPairs :: [cps] -> typ -> typ

-- | Termination Problems have info about termination
class IsTProblem typ termination | typ -> termination  where
    getTermination :: typ -> termination
    setTermination :: termination -> typ -> typ

-- | Normalization Problems have info about normalization
class IsNProblem typ normalization | typ -> normalization  where
    getNormalization :: typ -> normalization
    setNormalization :: normalization -> typ -> typ

-- | Linearity Problems have info about linearity
class IsLProblem typ linear | typ -> linear  where
    getLinearity :: typ -> linear
    setLinearity :: linear -> typ -> typ

-- | Critical Pair _ Termination Problems have a list of critical pairs and termination
class IsCPTermProblem typ cps termination | typ -> cps termination where
    getTermCPT :: typ -> termination
    setTermCPT :: termination -> typ -> typ
    getCPairsCPT :: typ -> [cps]
    setCPairsCPT :: [cps] -> typ -> typ

-- | Critical Pair _ Normalization Problems have a list of critical pairs and normalization
class IsCPNormProblem typ cps normalization | typ -> cps normalization where
    getNormCPN :: typ -> normalization
    setNormCPN :: normalization -> typ -> typ
    getCPairsCPN :: typ -> [cps]
    setCPairsCPN :: [cps] -> typ -> typ

-- | Critical Pair _ Linear Problems have a list of critical pairs and if it's linear
class IsCPLinProblem typ cps linear | typ -> cps linear where
    getLinCPL :: typ -> linear
    setLinCPL :: linear -> typ -> typ
    getCPairsCPL :: typ -> [cps]
    setCPairsCPL :: [cps] -> typ -> typ

-- | Critical Pair  _ Problem Conclusions Problems have a list of critical pairs and problem conclusions
class IsCPPCProblem typ cps pc | typ -> cps pc where
    getCPairsCPPC :: typ -> [cps]
    setCPairsCPPC :: [cps] -> typ -> typ
    getProbConcCPPC :: typ -> pc
    setProbConcCPPC :: pc -> typ -> typ

-- | Critical Pair _ Termination _ Problem Conclusions Problems have a list of critical pairs, termination and problem conclusions
class IsCPTermPCProblem typ cps termination pc | typ -> cps termination pc where
    getTermCPTPC :: typ -> termination
    setTermCPTPC :: termination -> typ -> typ
    getCPairsCPTPC :: typ -> [cps]
    setCPairsCPTPC :: [cps] -> typ -> typ
    getProbConcCPTPC :: typ -> pc
    setProbConcCPTPC :: pc -> typ -> typ

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Pretty

instance (Pretty idF, Pretty idV) => Pretty (CriticalPair idF idV) where
  pPrint (CriticalPair { s = left
                       , t = right
                       , ccpConds = cs
                       }) = char '<' HPJC.<> pPrint left HPJC.<> char ',' HPJC.<> pPrint right HPJC.<> char '>'
                            HPJC.<+> (if Prelude.null cs then text ""
                                     else char '|' HPJC.<+> (hsep . punctuate comma . map(\c -> pPrint c) $ cs))


instance (Pretty idF, Pretty idV) => Pretty (CriticalPairInfo idF idV) where
  pPrint (CriticalPairInfo { trivial = t
                           , overlay = o
                           , critPair = cp
                           , cpNum = n
                           , hasFeasCds = (Just feasCds)
                           }) = pPrint cp HPJC.<+> text " => " HPJC.<+> text (if t then "Trivial" else "Not trivial") HPJC.<> comma
                              HPJC.<+> text (if o then "Overlay" else "Not overlay") HPJC.<> comma 
                              HPJC.<+> text (
                                if feasCds then "Feasible conditions"
                                else "Infeasible conditions") HPJC.<> comma
                              HPJC.<+> text "NW" HPJC.<> pPrint (normWeight cp) HPJC.<> comma
                              HPJC.<+> text "N" HPJC.<> pPrint n
  pPrint (CriticalPairInfo { trivial = t
                           , overlay = o
                           , critPair = cp
                           , cpNum = n
                           , hasFeasCds = Nothing
                           }) = pPrint cp HPJC.<+> text " => " HPJC.<+> text (if t then "Trivial" else "Not trivial") HPJC.<> comma
                              HPJC.<+> text (if o then "Overlay" else "Not overlay") HPJC.<> comma
                              HPJC.<+> text "NW" HPJC.<> pPrint (normWeight cp) HPJC.<> comma
                              HPJC.<+> text "N" HPJC.<> pPrint n

