{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.CleanCTRSProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the cleanCTRSProc processor
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.CleanCTRSProcessor (

-- * Exported functions

cleanCTRSProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasOSSignature(..))
import TRS.Terms(Term)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity (..))
import TRS.Rules(Rule(..), CRule(..), lhs, rhs, cRuleVars, condVars, rule, lcond, rcond)
import TRS.Terms(Term(..))
import Framework.Problem.Types(InConfCSCTRS (..),)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem)
import Framework.ProblemConclusions (isOriented, isJoin, isSemiequational)
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (singleP, Proof, Info (..), success)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), hcat, hsep, comma, punctuate)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Set as S (Set, elems, fromList, map, unions, delete, union)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------


data CleanCTRSProcessor  = CleanCTRSProcessor {timeout :: Maybe Int}

data CleanCTRSProcInfo problem idF idV = 
  CleanCTRSProcInfo { inProblem :: problem 
                    , inCRuleInfCheckerInfo :: [CRuleInfCheckerInfo idF idV]
                    }

data CRuleInfCheckerInfo idF idV =
  CRuleInfCheckerInfo { cRuleI          :: CRule (Term idF idV)
                      , cRuleIOut       :: String
                      , cRuleDeleted    :: Bool
                      }

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InConfCSCTRS trs
  , IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , HasArity idF
  , HasName idF, HasName idV
  , Show idF, Show idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CRewriting conftrs)
  , Info info (CleanCTRSProcInfo (Problem CRewriting conftrs) idF idV)
  ) => Processor info CleanCTRSProcessor (Problem CRewriting conftrs) (Problem CRewriting conftrs) where
  apply CleanCTRSProcessor {timeout = to} inP
    = singleP (CleanCTRSProcInfo newP criics) inP newP 
    where trs = getConfProblem . getR $ inP
          crules = getCTRSRules trs
          ctype
            | isOriented $ elems crules = 1
            | isJoin $ elems crules = 2
            | isSemiequational $ elems crules = 3
            | otherwise = 0
          criics = cleanCRules ctype to (elems crules) crules (getTRSVariables trs)
          newRules = S.fromList $ getRemainingRules criics
          newVars = unions . S.elems $ S.map cRuleVars newRules
          rs = S.map rule newRules
          newTRS = (setTRSVariables newVars . setCTRSRules newRules) trs
          newP = mkProblem (CRewriting (getPSchemeFromProblem inP)) (InConfCSCTRS newTRS)

-- Pretty

instance (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CRewriting conftrs)
  ) => Pretty (CleanCTRSProcInfo (Problem CRewriting conftrs) idF idV) where
  pPrint CleanCTRSProcInfo { inProblem = problem 
                           , inCRuleInfCheckerInfo = criics
                           }
    = text "\n\nClean CTRS Processor:\n"
      Doc.<> pPrint problem
      Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\c -> pPrint c) $ criics)

instance (
    Pretty idF, Pretty idV
  ) => Pretty (CRuleInfCheckerInfo idF idV) where
  pPrint (CRuleInfCheckerInfo
    { cRuleI = cr
    , cRuleIOut = out
    , cRuleDeleted = del
    }) = text "\nCRule InfChecker Info:\n"
    Doc.<> pPrint cr Doc.<> text "\n"
    Doc.<> text (if del then "Rule deleted" else "Rule remains") Doc.<> text "\n"
    Doc.<> text "Proof:\n" Doc.<> text (translateFromRiskyChars out)

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Clean CTRS Processor
cleanCTRSProc :: (Processor info CleanCTRSProcessor o d
          , MonadPlus mp
          , Info info o
          , Info info d) => Maybe (Int) -> o -> Proof info mp d
cleanCTRSProc to = apply CleanCTRSProcessor {timeout = to}


cleanCRules :: (Pretty idF, Pretty idV, HasName idF, HasName idV, HasArity idF, Ord idF, Ord idV, Show idF, Show idV, Eq idF, Eq idV) => 
  Int -> Maybe Int -> [CRule (Term idF idV)] -> Set (CRule (Term idF idV)) -> Set (idV) -> [CRuleInfCheckerInfo idF idV]
cleanCRules _ _ [] _ _ = []
cleanCRules ctype timeout (cr:crs) crSet trsVsSet
  | (lhs r) == (rhs r) = (CRuleInfCheckerInfo cr "SAME_LHS_AND_RHS" True):(cleanCRules ctype timeout crs newCrSet newTrsVsSet)
  | Prelude.null cds = (CRuleInfCheckerInfo cr "NO_CONDS" False):(cleanCRules ctype timeout crs crSet trsVsSet)
  | isTrue infRes = (CRuleInfCheckerInfo cr iout True):(cleanCRules ctype timeout crs newCrSet newTrsVsSet)
  | otherwise = (CRuleInfCheckerInfo cr iout False):(cleanCRules ctype timeout crs crSet trsVsSet)
  where
    (infRes, iout) = unsafePerformIO $!
                    do { 
                       solution <- callToInfChecker timeout infProblem
                       ; case solution of
                          ("NO", Just sol)  -> return ((Just False, sol))
                          ("YES", Just sol) -> return ((Just True, sol))
                          (_, Just sol)     -> return ((Nothing, "MAYBE"))
                    }
    (r :=> cds) = cr
    cdsvs = unions . S.elems . S.map condVars $ S.fromList cds
    newCrSet = delete cr crSet
    newTrsVsSet = unions . S.elems $ S.map cRuleVars newCrSet
    conditionType
      | ctype == 1 = "ORIENTED"
      | ctype == 2 = "JOIN"
      | ctype == 3 = "SEMI-EQUATIONAL"
      | otherwise = error "Relation not supported. CONDITIONTYPE not specified"
    infProblem = 
      bconcat $ 
          [ frmStr "(CONDITIONTYPE "
          , frmStr  conditionType
          , frmStr ")\n"
          , frmStr "(VAR "
          , frmStr . noRiskyChars $ show . hsep . Prelude.map (\v -> pPrint v) $ (elems trsVsSet)
          , frmStr ")\n"
          , frmStr "(RULES\n"
          , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . Prelude.map (\r -> pPrint r) $ (elems crSet)
          , frmStr ")\n"
          , frmStr "(VAR "
          , frmStr . noRiskyChars $ show . hsep . Prelude.map(\v -> pPrint v) $ (elems cdsvs)
          , frmStr ")\n"
          , frmStr "(CONDITION "
          , frmStr . noRiskyChars $ show . hsep . punctuate comma . Prelude.map(\c -> pPrint c) $ (cds)
          , frmStr ")"]
  
getRemainingRules :: [CRuleInfCheckerInfo idF idV] -> [CRule (Term idF idV)]
getRemainingRules [] = []
getRemainingRules (cri:cris)
  | cRuleDeleted cri = getRemainingRules cris
  | otherwise = (cRuleI cri):(getRemainingRules cris)
