{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.CleanInfCCPsCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that delete CCPs with infeasible conditions
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.CleanInfCCPsCProcessor (

-- * Exported functions

cleanICCPs

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(CRule (..), Cond (..), lcond, rcond)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), IsCPProblem (..), getPSchemeFromProblem)
import Framework.ProblemConclusions (isOriented, isJoin, isSemiequational)
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted, singleP)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS, CritPairInfCheckerInfo (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, int)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty, (\\), singleton, insert,
                      lookupIndex, elemAt, showTree)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data CleanInfCCPsCProcessor = CleanInfCCPsCProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data CleanICCPsInfo problem idF idV = 
  CleanICCPsInfo { inProblem            :: problem
                 , inCriticalPInfCInfo  :: [CritPairInfCheckerInfo idF idV]
                 , inNewCPIs            :: [CriticalPairInfo idF idV]
                 , inProblemConclusions :: ProblemConclusionsCTRS
                 , inTimeout            :: Maybe Int
                 , inTimeoutEach        :: Maybe Int
                 , inNumCPIs            :: Int
                 }
  | CleanICCPsDontKnow { inProblem            :: problem
                       , inCriticalPInfCInfo  :: [CritPairInfCheckerInfo idF idV]
                       , inNewCPIs            :: [CriticalPairInfo idF idV]
                       , inProblemConclusions :: ProblemConclusionsCTRS
                       , inTimeout            :: Maybe Int
                       , inTimeoutEach        :: Maybe Int
                       , inNumCPIs            :: Int
                       }




-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermPCCSCTRS trs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Show idF, Show idV
  , Info info (CleanICCPsInfo (Problem CPTPCCRewriting conftrs) idF idV)
  ) => Processor info CleanInfCCPsCProcessor (Problem CPTPCCRewriting conftrs) (Problem CPTPCCRewriting conftrs) where
  apply CleanInfCCPsCProcessor {timeout = to} inP
    | conf = success (CleanICCPsInfo inP cpiics newCpis newPconc to toEach (length cpis)) inP -- True = success (CleanICCPsInfo inP cpiics newCpis pconc toEach) inP -- 
    | otherwise = singleP (CleanICCPsInfo inP cpiics newCpis newPconc to toEach (length cpis)) inP 
                    (mkProblem (CPTPCCRewriting (getPSchemeFromProblem inP)) (InCPTermPCCSCTRS trs newCpis (terminatingCTRS pconc) newPconc))
    where trs = getConfProblem . getR $ inP
          crules = getCTRSRules trs
          cpis = getCPairsCPTPC . getR $ inP
          cps = map critPair cpis
          iccs = [buildInfCheckerCriteriaCTRS trs cp | cp <- cps]
          --cpisWithCds = numOfCPIsWithConds cpis
          --newTimeout = if cpisWithCds == 0 then 0 else div (fromJust to) cpisWithCds
          --minTo = div (fromJust to) 10
          --toEach = if isNothing to then Nothing else if newTimeout < minTo then Just minTo else Just newTimeout
          toEach = to
          ctype
            | isOriented $ elems crules = 1
            | isJoin $ elems crules = 2
            | isSemiequational $ elems crules = 3
            | otherwise = 0
          cpiics = cleanCCPsByCrits ctype toEach iccs cpis S.empty
          newCpis = getCPIsByCPIICs cpiics
          pconc = getProbConcCPTPC . getR $ inP
          joinable
            | Prelude.null newCpis = Just True
            | otherwise = Nothing
          newPconc = problemConclusionCTRS trs newCpis Nothing Nothing (terminatingCTRS pconc) (opTerminatingCTRS pconc) joinable
          --sd = stronglyDet newPconc
          conf = confluentCTRS pconc --((Prelude.null newCpis) && (isTrue opterm) && (isTrue sd)) || (levelConfluent newPconc)


instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Show idF, Show idV
  , Info info (CleanICCPsInfo (Problem CPCRewriting conftrs) idF idV)
  ) => Processor info CleanInfCCPsCProcessor (Problem CPCRewriting conftrs) (Problem CPCRewriting conftrs) where
  apply CleanInfCCPsCProcessor {timeout = to} inP
    | conf = success (CleanICCPsInfo inP cpiics newCpis pconc to toEach (length cpis)) inP -- True = success (CleanICCPsInfo inP cpiics newCpis pconc to toEach (length cpis)) inP --
    | otherwise = singleP (CleanICCPsInfo inP cpiics newCpis pconc to toEach (length cpis)) inP 
                    (mkProblem (CPCRewriting (getPSchemeFromProblem inP)) (InCPCSCTRS trs newCpis))
    where trs = getConfProblem . getR $ inP
          crules = getCTRSRules trs
          cpis = getCPairs . getR $ inP
          cps = map critPair cpis
          iccs = [buildInfCheckerCriteriaCTRS trs cp | cp <- cps]
          --cpisWithCds = numOfCPIsWithConds cpis
          --newTimeout = if cpisWithCds == 0 then 0 else div (fromJust to) cpisWithCds
          --minTo = div (fromJust to) 10
          --toEach = if isNothing to then Nothing else if newTimeout < minTo then Just minTo else Just newTimeout
          toEach = to
          ctype
            | isOriented $ elems crules = 1
            | isJoin $ elems crules = 2
            | isSemiequational $ elems crules = 3
            | otherwise = 0
          cpiics = cleanCCPsByCrits ctype toEach iccs cpis S.empty
          newCpis = getCPIsByCPIICs cpiics
          joinable
            | Prelude.null newCpis = Just True
            | otherwise = Nothing
          pconc = problemConclusionCTRS trs newCpis Nothing Nothing Nothing Nothing joinable
          conf = confluentCTRS pconc

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , HasName idF, HasName idV
  , HasArity idF
  , Pretty idF, Pretty idV
  , Pretty (Problem CPTPCCRewriting conftrs)
  ) => Pretty (CleanICCPsInfo (Problem CPTPCCRewriting conftrs) idF idV) where
  pPrint CleanICCPsInfo { inProblem = problem
                        , inCriticalPInfCInfo = cpiics
                        , inNewCPIs = newCpis
                        , inProblemConclusions = pconc
                        , inTimeout = to
                        , inTimeoutEach = toEach
                        , inNumCPIs = ncpis
                        }
    = text "\nClean Infeasible CCPs Processor:\n"
      Doc.<> (
      if ncpis > 0 then
      text "Num of CPIs:" Doc.<+> int ncpis Doc.<> text "\n"
      Doc.<> (
          if isJust to then text "Timeout:" Doc.<+> int (fromJust to) Doc.<> text "\n"
          else text ""
        )
        Doc.<> (
            if isJust toEach then
            text "Timeout for each infeasibility problem:"
            Doc.<+> int (fromJust toEach) Doc.<+> text "s\n"
            else text ""
          )
      else text ""
      )
      Doc.<> (hcat . punctuate (text "\n\n") . map (\info ->
        pPrint (cpairI info)
        Doc.<> (if Prelude.null . freshCCPConds $ cpIcccrit info
                then text "\n\n"
                else (infCIn info))
        Doc.<> text "Proof:\n" Doc.<> text (translateFromRiskyChars $ cpIOut info) Doc.<> text "\n\n"
        ) $ cpiics)
      Doc.<> pPrint pconc
      Doc.<> text "\n\nResulting CCPs:\n"
      Doc.<> (
        if Prelude.null newCpis then text "No CCPs left"
        else (hcat . punctuate (text "\n") . map (\c -> pPrint c) $ newCpis)
        )
    where trs = getConfProblem . getR $ problem
          conditionType
            | oriented pconc = "ORIENTED"
            | join pconc = "JOIN"
            | semiequational pconc = "SEMI-EQUATIONAL"
            | otherwise = error "Relation not supported. CONDITIONTYPE not specified"
          infCIn info =  
            text "\n(CONDITIONTYPE "
            Doc.<> text conditionType
            Doc.<> text ")\n"
            Doc.<> text "(VAR "
            Doc.<> (hsep . map (\v -> pPrint v) $ (trsVars $ cpIcccrit info))
            Doc.<> text ")\n"
            Doc.<> text "(RULES\n"
            Doc.<> (hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC $ cpIcccrit info))
            Doc.<> text ")\n"
            Doc.<> (if Prelude.null . freshCCPCondsVs $ cpIcccrit info then text ""
                    else text "(VAR "
                         Doc.<> (hsep . map(\v -> pPrint v) $ (freshCCPCondsVs $ cpIcccrit info)) 
                         Doc.<> text ")\n")
            Doc.<> text "(CONDITION "
            Doc.<> (hsep . punctuate comma . map(\c -> pPrint c) $ (freshCCPConds $ cpIcccrit info))
            Doc.<> text ")\n\n"


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , HasName idF, HasName idV
  , HasArity idF
  , Pretty idF, Pretty idV
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (CleanICCPsInfo (Problem CPCRewriting conftrs) idF idV) where
  pPrint CleanICCPsInfo { inProblem = problem
                        , inCriticalPInfCInfo = cpiics
                        , inNewCPIs = newCpis
                        , inProblemConclusions = pconc
                        , inTimeout = to
                        , inTimeoutEach = toEach
                        , inNumCPIs = ncpis
                        }
    = text "\nClean Infeasible CCPs Processor:\n"
      Doc.<> (
      if ncpis > 0 then
      text "Num of CPIs:" Doc.<+> int ncpis Doc.<> text "\n"
      Doc.<> (
          if isJust to then text "Timeout:" Doc.<+> int (fromJust to) Doc.<> text "\n"
          else text ""
        )
        Doc.<> (
            if isJust toEach then
            text "Timeout for each infeasibility problem:"
            Doc.<+> int (fromJust toEach) Doc.<+> text "s\n"
            else text ""
          )
      else text ""
      )
      Doc.<> (hcat . punctuate (text "\n\n") . map (\info ->
        pPrint (cpairI info)
        Doc.<> (if Prelude.null . freshCCPConds $ cpIcccrit info
                then text "\n\n"
                else (infCIn info))
        Doc.<> text "Proof:\n" Doc.<> text (translateFromRiskyChars $ cpIOut info) Doc.<> text "\n\n"
        ) $ cpiics)
      Doc.<> pPrint pconc
      Doc.<> text "\n\nResulting CCPs:\n"
      Doc.<> (
        if Prelude.null newCpis then text "No CCPs left"
        else (hcat . punctuate (text "\n") . map (\c -> pPrint c) $ newCpis)
        )
    where trs = getConfProblem . getR $ problem
          conditionType
            | oriented pconc = "ORIENTED"
            | join pconc = "JOIN"
            | semiequational pconc = "SEMI-EQUATIONAL"
            | otherwise = error "Relation not supported. CONDITIONTYPE not specified"
          infCIn info =  
            text "\n(CONDITIONTYPE "
            Doc.<> text conditionType
            Doc.<> text ")\n"
            Doc.<> text "(VAR "
            Doc.<> (hsep . map (\v -> pPrint v) $ (trsVars $ cpIcccrit info))
            Doc.<> text ")\n"
            Doc.<> text "(RULES\n"
            Doc.<> (hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC $ cpIcccrit info))
            Doc.<> text ")\n"
            Doc.<> (if Prelude.null . freshCCPCondsVs $ cpIcccrit info then text ""
                    else text "(VAR "
                         Doc.<> (hsep . map(\v -> pPrint v) $ (freshCCPCondsVs $ cpIcccrit info)) 
                         Doc.<> text ")\n")
            Doc.<> text "(CONDITION "
            Doc.<> (hsep . punctuate comma . map(\c -> pPrint c) $ (freshCCPConds $ cpIcccrit info))
            Doc.<> text ")\n\n"

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Clean Infeasible CCPs Processor
cleanICCPs :: ( Processor info CleanInfCCPsCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
cleanICCPs to = apply CleanInfCCPsCProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
cleanCCPsByCrits :: (Pretty idF, Pretty idV, HasName idF, HasName idV, HasArity idF, Ord idF, Ord idV, Show idF, Show idV, Eq idF, Eq idV) => 
  Int -> Maybe Int -> [InfCheckerCriteria] -> [CriticalPairInfo idF idV] -> Set ([Cond (Term IdFTRS IdVTRS)], Bool) -> [CritPairInfCheckerInfo idF idV]
cleanCCPsByCrits _ _ _ [] _ = []
cleanCCPsByCrits _ _ [] _ _ = []
cleanCCPsByCrits ctype timeout (cr:crits) (cpi:cpis) cdsSet
  | Prelude.null cds = (CritPairInfCheckerInfo cpi cr "NO_CONDS" False):(cleanCCPsByCrits ctype timeout crits cpis cdsSet)
  | isRepCond && wasInf = (CritPairInfCheckerInfo cpi cr "ALREADY_CHECKED_INFEASIBLE_CONDS" True):(cleanCCPsByCrits ctype timeout crits cpis cdsSet)
  | isRepCond && (not wasInf) = (CritPairInfCheckerInfo cpi cr "ALREADY_CHECKED_FEASIBLE_CONDS" False):(cleanCCPsByCrits ctype timeout crits cpis cdsSet)
  | isTrue infRes = (CritPairInfCheckerInfo cpi cr iout True):(cleanCCPsByCrits ctype timeout crits cpis newCdsSet)
  | otherwise = (CritPairInfCheckerInfo feasibleCPI cr iout False):(cleanCCPsByCrits ctype timeout crits cpis newCdsSet)
  where
    (infRes, iout) = unsafePerformIO $!
                    do { 
                       solution <- callToInfChecker timeout infProblem
                       ; case solution of
                          ("NO", Just sol)  -> return ((Just False, sol))
                          ("YES", Just sol) -> return ((Just True, sol))
                          (_, Just sol)     -> return ((Nothing, "MAYBE"))
                    }
    feasRes = if isNothing infRes then Nothing else Just (not $ isTrue infRes)
    feasibleCPI =  CriticalPairInfo (trivial cpi) (overlay cpi) (critPair cpi) (cpNum cpi) feasRes
    cds = freshCCPConds cr
    cdsvs = freshCCPCondsVs cr
    cdsIndexT = lookupIndex (cds, True) cdsSet
    cdsIndexF = lookupIndex (cds, False) cdsSet
    cdsIndex = if isJust cdsIndexT then cdsIndexT else if isJust cdsIndexF then cdsIndexF else Nothing
    isRepCond = isJust cdsIndex
    wasInf = snd $ elemAt (fromJust cdsIndex) cdsSet
    newCdsSet = insert (cds, (isTrue infRes)) cdsSet
    conditionType
      | ctype == 1 = "ORIENTED"
      | ctype == 2 = "JOIN"
      | ctype == 3 = "SEMI-EQUATIONAL"
      | otherwise = error "Relation not supported. CONDITIONTYPE not specified"
    infProblem = 
          bconcat $ 
          [ frmStr "(CONDITIONTYPE "
          , frmStr  conditionType
          , frmStr ")\n"
          , frmStr "(VAR "
          , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (trsVars cr)
          , frmStr ")\n"
          , frmStr "(RULES\n"
          , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC cr)
          , frmStr ")\n"
          , frmStr "(VAR "
          , frmStr . noRiskyChars $ show . hsep . map(\v -> pPrint v) $ (cdsvs)
          , frmStr ")\n"
          , frmStr "(CONDITION "
          , frmStr . noRiskyChars $ show . hsep . punctuate comma . map(\c -> pPrint c) $ (cds)
          , frmStr ")"]


getCPIsByCPIICs :: [CritPairInfCheckerInfo idF idV] -> [CriticalPairInfo idF idV]
getCPIsByCPIICs [] = []
getCPIsByCPIICs (c:cs)
  | cpInfConds c = getCPIsByCPIICs cs
  | otherwise = (cpairI c):(getCPIsByCPIICs cs)


numOfCPIsWithConds :: [CriticalPairInfo idF idV] -> Int
numOfCPIsWithConds [] = 0
numOfCPIsWithConds (cpi:cpis)
  | trivial cpi = numOfCPIsWithConds cpis
  | hasConds = 1 + numOfCPIsWithConds cpis
  | otherwise = numOfCPIsWithConds cpis
  where
    hasConds = not . Prelude.null . ccpConds $ critPair cpi

{-
filterICCCPIsSameConds :: (Ord idF, Ord idV) => [InfCheckerCriteria] -> [CriticalPairInfo idF idV] -> [Cond (Term idF idV)] -> [(InfCheckerCriteria, CriticalPairInfo idF idV)]
filterICCCPIsSameConds [] _ _ = []
filterICCCPIsSameConds _ [] _ = []
filterICCCPIsSameConds (icc:iccs) (cpi:cpis) cds
  | (isNothing $ freshCCPConds icc) = (icc, cpi):(filterICCCPIsSameConds iccs cpis cds)
  | S.null diff = filterICCCPIsSameConds iccs cpis cds
  | otherwise = (icc, cpi):(filterICCCPIsSameConds iccs cpis cds)
  where diff = (fromList . fromJust $ freshCCPConds icc) \\ (fromList cds)
-}