{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.ConvOverlayCCPProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that checks if all CCPs are joinable overlay
-- [DerOkaSiv88-T4]
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.ConvOverlayCCPProcessor (

-- * Exported functions

convOvProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(CRule (..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), 
                           getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPTermPCCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted, andP)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS, CritPairInfCheckerInfo (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Convergent Overlay CCP Processor
data ConvOverlayCCPProcessor = ConvOverlayCCPProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data ConvOvCCPProcInfo problem idF idV = 
  ConvOvCCPProcInfo { inProblem            :: problem
                    , inAllOvCCPsJoin      :: Bool
                    , inCriticalPInfCInfo  :: [CritPairInfCheckerInfo idF idV]
                    , inProblemConclusions :: ProblemConclusionsCTRS
                    }
  | ConvOvCCPDontKnow { inProblem            :: problem
                      , inAllOvCCPsJoin      :: Bool
                      , inCriticalPInfCInfo  :: [CritPairInfCheckerInfo idF idV]
                      , inProblemConclusions :: ProblemConclusionsCTRS
                      }



-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermPCCSCTRS trs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV
  , HasArity idF
  , HasName idF , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (ConvOvCCPProcInfo (Problem CPTPCCRewriting conftrs) idF idV)
  ) => Processor info ConvOverlayCCPProcessor (Problem CPTPCCRewriting conftrs) (Problem CPTPCCRewriting conftrs) where
  apply ConvOverlayCCPProcessor {timeout = to} inP
    -- | True = success (ConvOvCCPProcInfo inP allCCPsJoin cpiics newPconc) inP
    | not ov = dontKnow (ConvOvCCPDontKnow inP allCCPsJoin cpiics newPconc) inP
    | conf = success (ConvOvCCPProcInfo inP allCCPsJoin cpiics newPconc) inP
    | otherwise = dontKnow (ConvOvCCPDontKnow inP allCCPsJoin cpiics newPconc) inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairsCPTPC . getR $ inP
          pconc = getProbConcCPTPC . getR $ inP
          cps = map critPair cpis
          iccs = [buildInfCheckerCriteriaCTRS trs cp | cp <- cps]
          term = getTermCPTPC . getR $ inP
          ov = overlayCTRS pconc
          cpiics = solveOvCCPsJoinability to iccs cpis
          allCCPsJoin = (and $ map (\c -> not $ cpInfConds c) cpiics) || (Prelude.null cpis)
          newPconc = problemConclusionCTRS trs cpis (normalCTRS pconc) (rightStable pconc) term Nothing (Just allCCPsJoin)
          conf = confluentCTRS newPconc

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , HasArity idF
  , HasName idF, HasName idV
  , Pretty idF, Pretty idV
  , Pretty (Problem CPTPCCRewriting conftrs)
  ) => Pretty (ConvOvCCPProcInfo (Problem CPTPCCRewriting conftrs) idF idV) where
  pPrint ConvOvCCPProcInfo { inProblem = problem
                           , inAllOvCCPsJoin = join
                           , inCriticalPInfCInfo = cpiics
                           , inProblemConclusions = pc
                           }
    = pPrint problem
      Doc.<> text "\nConvergent Overlay CCP Processor:\n"
      Doc.<> text (if join then "All CCPs are overlay joinable\n" else "Not all CCPs are overlay joinable\n")
      Doc.<> (hcat . punctuate (text "\n\n") . map (\info ->
        pPrint (cpairI info)
        Doc.<> (infCIn info)
        Doc.<> text "Proof:\n" Doc.<> text (translateFromRiskyChars $ cpIOut info) Doc.<> text "\n\n"
        ) $ cpiics)
      Doc.<> pPrint pc
    where trs = getConfProblem . getR $ problem
          infCIn info =  
            text "\n(PROBLEM INFEASIBILITY)\n"
            Doc.<> text "(VAR "
            Doc.<> (hsep . map (\v -> pPrint v) $ (trsVars $ cpIcccrit info))
            Doc.<> text ")\n"
            Doc.<> text "(RULES\n"
            Doc.<> (hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC $ cpIcccrit info))
            Doc.<> text ")\n"
            Doc.<> text "(VAR" Doc.<+> text (infCheckerJVN $ cpIcccrit info) Doc.<> text ")\n"  
            Doc.<> text "(CONDITION "
            Doc.<> pPrint (fst . cpsConst $ cpIcccrit info) Doc.<+> text "==" Doc.<+> text (infCheckerJVN $ cpIcccrit info) Doc.<> comma 
            Doc.<+> pPrint (snd . cpsConst $ cpIcccrit info) Doc.<+> text "==" Doc.<+> text (infCheckerJVN $ cpIcccrit info)
            Doc.<> text ")\n\n"

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Convergent Overlay CCP Processor
convOvProc :: ( Processor info ConvOverlayCCPProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
convOvProc to = apply ConvOverlayCCPProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveOvCCPsJoinability :: (Pretty idV, HasName idF, HasName idV, HasArity idF, Ord idF, Ord idV, Eq idF, Eq idV) => 
  Maybe Int -> [InfCheckerCriteria] -> [CriticalPairInfo idF idV] -> [CritPairInfCheckerInfo idF idV]
solveOvCCPsJoinability _ _ [] = []
solveOvCCPsJoinability _ [] _ = []
solveOvCCPsJoinability timeout (cr:crits) (cpi:cpis)
  | (hasConds) && (not . isTrue $ hasFeasCds cpi) = [CritPairInfCheckerInfo cpi cr "MAYBE_INFEASIBLE_CCP_CONDITIONS" True]
  | feasible = (CritPairInfCheckerInfo cpi cr iout False):(solveOvCCPsJoinability timeout crits cpis)
  | otherwise = [CritPairInfCheckerInfo cpi cr iout True]
  where
    hasConds = not . Prelude.null . ccpConds $ critPair cpi
    (feasible, iout) = unsafePerformIO $!
                    do { 
                       solution <- callToInfChecker timeout infProblem
                       ; case solution of
                          ("NO", Just sol)  -> return ((True, sol))
                          ("YES", Just sol) -> return ((False, sol))
                          (_, Just sol)     -> return ((False, "MAYBE"))
                    }
    infProblem =
      bconcat $ [ frmStr "(PROBLEM INFEASIBILITY)\n"
                , frmStr "(VAR "
                , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (trsVars cr)
                , frmStr ")\n"
                , frmStr "(RULES\n"
                , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC cr)
                , frmStr ")\n"
                , frmStr "(VAR " 
                , frmStr . noRiskyChars $ show . text $ infCheckerJVN cr
                , frmStr ")\n"
                , frmStr "(CONDITION "
                , frmStr . noRiskyChars $ show . pPrint . fst $ cpsConst cr 
                , frmStr " == "
                , frmStr . noRiskyChars $ show . text $ infCheckerJVN cr
                , frmStr ", "
                , frmStr . noRiskyChars $ show . pPrint . snd $ cpsConst cr 
                , frmStr " == "
                , frmStr . noRiskyChars $ show . text $ infCheckerJVN cr
                , frmStr ")"]


{-
solveOvCCPsJoinability :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), Pretty idF, Pretty idV, Pretty trs, HasName idF, HasName idV, HasArity idF, Ord idF, Ord idV, Eq idF, Eq idV) => 
  Maybe Int -> trs -> [InfCheckerCriteria] -> [CriticalPairInfo idF idV] -> [CritPairInfCheckerInfo idF idV]
solveOvCCPsJoinability _ _ _ [] = []
solveOvCCPsJoinability _ _ [] _ = []
solveOvCCPsJoinability timeout trs (cr:crits) (cpi:cpis)
  | isNothing $ hasFeasCds cpi = [CritPairInfCheckerInfo cpi cr "MAYBE_INFEASIBLE_CCP_CONDITIONS" True]
  | feasible = (CritPairInfCheckerInfo cpi cr iout False):(solveOvCCPsJoinability timeout trs crits cpis)
  | otherwise = [CritPairInfCheckerInfo cpi cr iout True]
  where
    (feasible, iout) = unsafePerformIO $!
                    do { 
                       solution <- callToInfChecker timeout infProblem
                       ; case solution of
                          ("NO", Just sol)  -> return ((True, sol))
                          ("YES", Just sol) -> return ((False, sol))
                          (_, Just sol)     -> return ((False, "MAYBE"))
                    }
    rls =  elems $ getCTRSRules trs
    infProblem =
      bconcat $ [ frmStr "(PROBLEM INFEASIBILITY)\n"
                , frmStr "(CONDITIONTYPE ORIENTED)\n"
                , frmStr "(VAR "
                , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (trsVars cr)
                , frmStr ")\n"
                , frmStr "(RULES\n"
                , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ rls
                , frmStr ")\n"
                , frmStr "(VAR "
                , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (freshCPVars cr)
                , frmStr " "
                , frmStr $ infCheckerJVN cr
                , frmStr ")\n"
                , frmStr "(CONDITION "
                , frmStr . noRiskyChars $ show . pPrint . fst . fromJust $ freshCP cr
                , frmStr " == "
                , frmStr $ infCheckerJVN cr
                , frmStr ", "
                , frmStr . noRiskyChars $ show . pPrint . snd . fromJust $ freshCP cr
                , frmStr " == "
                , frmStr $ infCheckerJVN cr
                , frmStr ")"]
-}