{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.CriticalPairsCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that calculate TRS CP Pairs and applies
-- Huet Levy theorem
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.CriticalPairsCProcessor (

-- * Exported data


-- * Exported functions

critPairsCProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasArity (..), HasMu (..))
import Framework.Problem.Types(InConfCSCTRS (..), InCPCSCTRS (..), InTermCSCTRS (..), InCPTermPCCSCTRS (..))
import Framework.Problems(IsConfProblem (..), IsCPProblem (..), CriticalPair (..), CriticalPairInfo (..), getPSchemeFromProblem, IsTProblem (..), 
      IsCPTermPCProblem (..), IsCPLinProblem (..), IsLProblem (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, andP, Proof, Info (..), isSuccess, refuted, runProof, singleP)
import TRS.Rules (freshCRule)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.Problem.TCRewriting (TCRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Framework.TRSAux (RuleId (..), UnificationsCPBuilder (..), RuleLhsPosSubterms (..), PosSubt (..), getRulesIds, maybeConvergentCPs, cpsInfo,
                         criticalPairUnificationNW, createCPs, cleanRepeatedCCPs, quicksortCPByNumOfNormalTerms, checkUnify, cleanUnifications,
                         posMuSubtermsRule, pairVarsCreation, getVarsCRules, getCRulesIds, criticalPairsInfoVars)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, braces, brackets, parens, hsep)
import Control.Monad (MonadPlus)
import Data.Set as S (Set, elems, null, findMax, union, fromList)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Critical Pairs Processor
data CriticalPairsCProcessor = CriticalPairsCProcessor

-- | The information obtained are the rules
data CriticalPairsCProcInfo problem idF idV = 
  CriticalPairsCProcInfo { inProblem            :: problem
                         , inRLhsPosSbts        :: [RuleLhsPosSubterms idF idV]
                         , inUnifications       :: [UnificationsCPBuilder idF idV]
                         , inCriticalPairsInfo  :: [CriticalPairInfo idF idV]
                         , inProblemConclusions :: ProblemConclusionsCTRS
                         , inVars               :: [idV]
                         }
  | CriticalPairsDontKnow { inProblem            :: problem
                          , inRLhsPosSbts        :: [RuleLhsPosSubterms idF idV]
                          , inUnifications       :: [UnificationsCPBuilder idF idV]
                          , inCriticalPairsInfo  :: [CriticalPairInfo idF idV]
                          , inProblemConclusions :: ProblemConclusionsCTRS
                          , inVars               :: [idV]
                          } 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor


instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InConfCSCTRS trs
  , conftrs2 ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsCPProblem conftrs2 (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (CriticalPairsCProcInfo (Problem CRewriting conftrs1) idF idV)
  ) => Processor info CriticalPairsCProcessor (Problem CRewriting conftrs1) (Problem CPCRewriting conftrs2) where
  apply CriticalPairsCProcessor inP
    -- | True = success (CriticalPairsCProcInfo inP rLhsPosSbts unifications cpis pconc vs) inP
    | conf = success (CriticalPairsCProcInfo inP rLhsPosSbts unifications cpis pconc vs) inP
    | otherwise = singleP (CriticalPairsCProcInfo inP rLhsPosSbts unifications cpis pconc vs) inP 
                    (mkProblem (CPCRewriting (getPSchemeFromProblem inP)) (InCPCSCTRS newTRS potConvCps))
    where trs = getConfProblem . getR $ inP
          crules = elems . getCTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshCRules,newMaxId) = runState (sequence . map freshCRule $ crules) maxid
          rIds = getCRulesIds crules
          vs = elems . S.fromList $ getVarsCRules crules
          rLhsPosSbts = map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = quicksortCPByNumOfNormalTerms . cleanRepeatedCCPs $ createCPs [criticalPairUnificationNW u rIds | u <- unifications]
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          joinable
            | Prelude.null potConvCps = Just True
            | otherwise = Nothing
          pconc = problemConclusionCTRS newTRS cpis Nothing Nothing Nothing Nothing joinable
          conf = confluentCTRS pconc


-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CRewriting conftrs)
  ) => Pretty (CriticalPairsCProcInfo (Problem CRewriting conftrs) idF idV) where
  pPrint CriticalPairsCProcInfo { inProblem = problem
                                , inRLhsPosSbts = rlps
                                , inUnifications = unfs
                                , inCriticalPairsInfo = cpis
                                , inProblemConclusions = pconc
                                , inVars = vs}
    = pPrint problem
      Doc.<> text "\nCritical Pairs Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<> (hcat . punctuate (text "\n") . map pPrint . elems . getCTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<> (hsep . punctuate comma . map(\v-> pPrint (getName v)) $ vs)
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<> (hcat . punctuate (text "\n") . map (\r ->
               text "crule: " Doc.<> pPrint (crule (rid r)) Doc.<> comma
        Doc.<+> text "id: " Doc.<> pPrint (cidr (rid r)) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . map(\ps-> pPrint (subterm ps) Doc.<> text "-> "
          Doc.<> (brackets . hcat . punctuate comma . map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))) $ rlps) 
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . map (\u -> 
               text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<+> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<+> text "p: " Doc.<> (brackets . hcat . punctuate comma . map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l: " Doc.<> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp: " Doc.<> pPrint (lp u) Doc.<> comma
        Doc.<+> text "conds: " Doc.<> (braces . hsep . punctuate comma . map (\c -> pPrint c) $ (css u)) Doc.<> comma
        Doc.<+> text "sig: " Doc.<> (braces . hcat . punctuate comma . map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l': " Doc.<> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r: " Doc.<> pPrint (r u) Doc.<> comma
        Doc.<+> text "r': " Doc.<> pPrint (r1 u)) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . map (\cpi -> pPrint cpi) $ cpis)
        Doc.<> pPrint pconc
    where trs = getConfProblem . getR $ problem


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Critical Pairs Processor
critPairsCProc :: ( Processor info CriticalPairsCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
critPairsCProc  = apply CriticalPairsCProcessor