{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.DiffGroundAndNormalNoCdsCCPTermsProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- if ccp terms are distinct, ground and normal with InfChecker.
-- The ccp must be unconditional
-- Analysed CCPs are not trivial before entering this processor
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.DiffGroundAndNormalNoCdsCCPTermsProcessor (

-- * Exported functions

dganncdsCCPProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..), vars)
import TRS.Rules(Rule (..), CRule (..), lcond, rcond, lhs, rhs, conds, rule)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), IsCPProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted, singleP)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Framework.TRSAux (isGroundTerm)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, lparen, rparen, char)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Different Normal CP Terms Processor
data DiffGANNCdsCCPTsProcessor = DiffGANNCdsCCPTsProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data DiffGANNCdsCCPTsProcInfo problem = 
  DiffGANNCdsCCPTsProcInfo { inProblem :: problem
                           , inICCrit  :: InfCheckerCriteria
                           , inSol     :: String
                           }
  | DiffGANNCdsCCPTsDontKnow { inProblem :: problem
                             , inICCrit  :: InfCheckerCriteria
                             , inSol     :: String
                             }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor


instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idV
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (DiffGANNCdsCCPTsProcInfo (Problem CPCRewriting conftrs))
  ) => Processor info DiffGANNCdsCCPTsProcessor (Problem CPCRewriting conftrs) (Problem CPCRewriting conftrs) where
  apply DiffGANNCdsCCPTsProcessor {timeout = to} inP
    | hasConds = dontKnow (DiffGANNCdsCCPTsDontKnow inP icc sol) inP
    | notGround = dontKnow (DiffGANNCdsCCPTsDontKnow inP icc sol) inP
    | fstNormal && sndNormal = refuted (DiffGANNCdsCCPTsProcInfo inP icc sol) inP
    | otherwise = dontKnow (DiffGANNCdsCCPTsDontKnow inP icc sol) inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          icc = buildInfCheckerCriteriaCTRS trs cp
          notGround = (not . isGroundTerm $ s cp) || (not . isGroundTerm $ t cp)
          (fstNormal, fstSol) = unsafePerformIO $!
                      do { 
                          solution <- solveInfCheckerFst to icc
                          ; case solution of
                              ("YES", Just sol) -> return (True, sol)
                              (_, Just sol)     -> return (False, sol)
                      }
          (sndNormal, sndSol) = unsafePerformIO $!
                      do { 
                          solution <- solveInfCheckerSnd to icc
                          ; case solution of
                              ("YES", Just sol) -> return (True, sol)
                              (_, Just sol)     -> return (False, sol)
                      }
          sol = fstSol ++ sndSol
          hasConds = not . Prelude.null $ ccpConds cp

          
-- Pretty


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty idV
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (DiffGANNCdsCCPTsProcInfo (Problem CPCRewriting conftrs)) where
  pPrint DiffGANNCdsCCPTsProcInfo { inProblem = problem
                              , inICCrit = icc
                              , inSol = sol
                              }
    = pPrint problem
    Doc.<> text "\nInfChecker Different Ground and Normal No Conditions CCP Terms Processor:\n"
    Doc.<> text "Proof:\n"
      Doc.<> text (translateFromRiskyChars sol)
    where trs = getConfProblem . getR $ problem



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the InfChecker Different Normal No Conditions CCP Terms Processor
dganncdsCCPProc :: ( Processor info DiffGANNCdsCCPTsProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
dganncdsCCPProc to = apply DiffGANNCdsCCPTsProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveInfCheckerFst :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)--IO (Maybe String)
solveInfCheckerFst timeout crit
  | isNothing timeout = callToInfChecker timeout infProblem
  | otherwise = callToInfChecker (Just div2to) infProblem
  where
    div2to = div (fromJust timeout) 2
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text " ") . map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(VAR y)\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show $ pPrint (fst $ cpsConst crit) Doc.<> text " -> y"
                 , frmStr ")"]

-- | Generate infChecker input and try to get a solution
solveInfCheckerSnd :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)--IO (Maybe String)
solveInfCheckerSnd timeout crit
  | isNothing timeout = callToInfChecker timeout infProblem
  | otherwise = callToInfChecker (Just div2to) infProblem
  where
    div2to = div (fromJust timeout) 2
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text " ") . map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(VAR x)\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show $ pPrint (snd $ cpsConst crit) Doc.<> text " -> x"
                 , frmStr ")"]