{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.InfConvCCPProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- with infChecker if ccp has conditions and the disjointed vars condition
-- is achieved [LucFSTTCS21-Prop22]
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.InfConvCCPProcessor (

-- * Exported functions

infConvCCPProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..), vars)
import TRS.Rules(Rule (..), CRule (..), lcond, rcond, lhs, rhs, conds, rule, fromCondToRule, condVars)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), IsCPProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, lparen, rparen, char)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty, map)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Infeasible Convergence CCP Processor
data InfConvCCPProcessor = InfConvCCPProcessor {timeout :: Maybe (Int)}

data InfConvCCPProcInfo problem = 
  InfConvCCPProcInfo { inProblem :: problem
                     , inICCrit  :: InfCheckerCriteria
                     , inSol     :: String
                     }
  | InfConvCCPDontKnow { inProblem :: problem
                       , inICCrit  :: InfCheckerCriteria
                       , inSol     :: String
                       }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idV
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (InfConvCCPProcInfo (Problem CPCRewriting conftrs))
  ) => Processor info InfConvCCPProcessor (Problem CPCRewriting conftrs) (Problem CPCRewriting conftrs) where
  apply InfConvCCPProcessor {timeout = to} inP
    | not hasConds = dontKnow (InfConvCCPDontKnow inP icc sol) inP
    | not feasConds = dontKnow (InfConvCCPDontKnow inP icc sol) inP
    | not disjointedVarsCond = dontKnow (InfConvCCPDontKnow inP icc sol) inP
    | notJoin = refuted (InfConvCCPProcInfo inP icc sol) inP
    | otherwise = dontKnow (InfConvCCPDontKnow inP icc sol) inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          icc = buildInfCheckerCriteriaCTRS trs cp
          hasConds = not . Prelude.null $ ccpConds cp
          feasConds = isTrue $ hasFeasCds cpi
          disjointedVarsCond = 
            S.null (
              ((vars $ s cp) `union` (vars $ t cp)) 
              `intersection`
              (unions . S.elems . S.map condVars . S.fromList $ ccpConds cp))
          (notJoin, sol) = unsafePerformIO $!
                           do { 
                                solution <- solveInfChecker to icc
                                ; case solution of
                                    ("YES", Just sol) -> return (True, sol)
                                    (_, Just sol)     -> return (False, sol)
                            }

          
-- Pretty

instance (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV, Ord String
  , Pretty idV
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (InfConvCCPProcInfo (Problem CPCRewriting conftrs)) where
  pPrint InfConvCCPProcInfo { inProblem = problem
                            , inICCrit = icc
                            , inSol = sol
                            }
    = pPrint problem
    Doc.<> text "\nInfeasible Convergence CCP Processor:\n"
    Doc.<> text "Infeasible convergence?\n"
      Doc.<> text (translateFromRiskyChars sol)
    where trs = getConfProblem . getR $ problem



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Infeasible Convergence CCP Processor
infConvCCPProc :: ( Processor info InfConvCCPProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
infConvCCPProc to = apply InfConvCCPProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveInfChecker :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
solveInfChecker timeout crit = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . Prelude.map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . Prelude.map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(VAR z)\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show $ pPrint (fst $ cpsConst crit) Doc.<+> text "== z" Doc.<> comma 
                                   Doc.<+> pPrint (snd $ cpsConst crit) Doc.<+> text "== z"
                 , frmStr ")"]


{-
-- | Generate infChecker input and try to get a solution
solveInfChecker :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), HasName idF, HasName idV, HasArity idF, Pretty idF, Pretty idV, Pretty trs) => Maybe Int -> trs -> InfCheckerCriteria -> IO (String, Maybe String)--IO (Maybe String)
solveInfChecker timeout trs crit = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show. hsep . Prelude.map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show. hcat . punctuate (text "\n") . Prelude.map (\r -> pPrint r) $ (elems $ getCTRSRules trs)
                 , frmStr ")\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show. hsep . Prelude.map (\v -> pPrint v) $ (freshCPVars crit)
                 , frmStr " "
                 , frmStr $ infCheckerJVN crit
                 , frmStr ")\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show. pPrint . fst . fromJust $ freshCP crit
                 , frmStr " == "
                 , frmStr $ infCheckerJVN crit
                 , frmStr ", "
                 , frmStr . noRiskyChars $ show. pPrint . snd . fromJust $ freshCP crit
                 , frmStr " == "
                 , frmStr $ infCheckerJVN crit
                 , frmStr ")"]
-}
