{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.InfConvConstCCPProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- with infChecker if ccp has conditions and "constantizing" the ccp
-- [LucFSTTCS21-Prop20]
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.InfConvConstCCPProcessor (

-- * Exported functions

infConvConstCCPProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..), vars)
import TRS.Rules(Rule (..), CRule (..), lcond, rcond, lhs, rhs, conds, rule, fromCondToRule, condVars)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), IsCPProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, lparen, rparen, char)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty, map)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Infeasible Convergence Constantized CCP Processor
data InfConvConstCCPProcessor = InfConvConstCCPProcessor {timeout :: Maybe (Int)}

data InfConvConstCCPProcInfo problem = 
  InfConvConstCCPProcInfo { inProblem  :: problem
                          , inICCrit   :: InfCheckerCriteria
                          , inCondsSol :: String
                          , inJoinSol  :: String
                          }
  | InfConvConstCCPDontKnow { inProblem  :: problem
                            , inICCrit   :: InfCheckerCriteria
                            , inCondsSol :: String
                            , inJoinSol  :: String
                            }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idV
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (InfConvConstCCPProcInfo (Problem CPCRewriting conftrs))
  ) => Processor info InfConvConstCCPProcessor (Problem CPCRewriting conftrs) (Problem CPCRewriting conftrs) where
  apply InfConvConstCCPProcessor {timeout = to} inP
    | not hasConds = dontKnow (InfConvConstCCPDontKnow inP icc "NO_CONDS" "NO_CONDS") inP
    | not feasConds = dontKnow (InfConvConstCCPDontKnow inP icc condsSol joinSol) inP
    | notJoin = refuted (InfConvConstCCPProcInfo inP icc condsSol joinSol) inP
    | otherwise = dontKnow (InfConvConstCCPDontKnow inP icc condsSol joinSol) inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          icc = buildInfCheckerCriteriaCTRS trs cp
          (feasConds, condsSol) =
            unsafePerformIO $!
              do { 
                  solution <- infeasibleConds to icc
                  ; case solution of
                      ("NO", Just sol) -> return (True, sol)
                      (_, Just sol)    -> return (False, sol)
              }
          (notJoin, joinSol) = 
            unsafePerformIO $!
              do { 
                  solution <- infeasibleConv to icc
                  ; case solution of
                      ("YES", Just sol) -> return (True, sol)
                      (_, Just sol)     -> return (False, sol)
              }
          hasConds = not . Prelude.null $ ccpConds cp

          
-- Pretty

instance (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV, Ord String
  , Pretty idV
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (InfConvConstCCPProcInfo (Problem CPCRewriting conftrs)) where
  pPrint InfConvConstCCPProcInfo { inProblem = problem
                            , inICCrit = icc
                            , inCondsSol = csol
                            , inJoinSol = jsol
                            }
    = pPrint problem
    Doc.<> text "\nInfeasible Convergence Constantized CCP Processor:\n"
    Doc.<> text "\nInfeasible constantized conditions?\n"
      Doc.<> text (translateFromRiskyChars csol)
    Doc.<> text "\nInfeasible constantized CCP convergence?\n"
      Doc.<> text (translateFromRiskyChars jsol)



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Infeasible Convergence Constantized CCP Processor
infConvConstCCPProc :: ( Processor info InfConvConstCCPProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
infConvConstCCPProc to = apply InfConvConstCCPProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
infeasibleConds :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
infeasibleConds timeout crit = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . Prelude.map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . Prelude.map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show . hsep . punctuate comma . Prelude.map (\c -> pPrint c) $ (condsConst crit)
                 , frmStr ")"]


-- | Generate infChecker input and try to get a solution
infeasibleConv :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
infeasibleConv timeout crit = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . Prelude.map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . Prelude.map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . text $ infCheckerJVN crit
                 , frmStr ")\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show . pPrint . fst $ cpsConst crit
                 , frmStr " == "
                 , frmStr . noRiskyChars $ show . text $ infCheckerJVN crit
                 , frmStr ", "
                 , frmStr . noRiskyChars $ show . pPrint . snd $ cpsConst crit
                 , frmStr " == "
                 , frmStr . noRiskyChars $ show . text $ infCheckerJVN crit
                 , frmStr ")"]
