{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.InfConvNoCdsCCPProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- with infChecker if ccp has no conditions
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.InfConvNoCdsCCPProcessor (

-- * Exported functions

infConvNCdsCCPProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(Rule (..), CRule (..), lcond, rcond, lhs, rhs, conds, rule)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), IsCPProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaCTRS)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, lparen, rparen, char)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Infeasible Convergence No Conditions CCP Processor
data InfConvNCdsCCPProcessor = InfConvNCdsCCPProcessor {timeout :: Maybe (Int)}

data InfConvNCdsCCPProcInfo problem = 
  InfConvNCdsCCPProcInfo { inProblem :: problem
                         , inICCrit  :: InfCheckerCriteria
                         , inSol     :: String
                         }
  | InfConvNCdsCCPDontKnow { inProblem :: problem
                           , inICCrit  :: InfCheckerCriteria
                           , inSol     :: String
                           }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor


instance (
    conftrs ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (InfConvNCdsCCPProcInfo (Problem CPCRewriting conftrs))
  ) => Processor info InfConvNCdsCCPProcessor (Problem CPCRewriting conftrs) (Problem CPCRewriting conftrs) where
  apply InfConvNCdsCCPProcessor {timeout = to} inP
    | hasConds = dontKnow (InfConvNCdsCCPDontKnow inP icc sol) inP
    | notJoin = refuted (InfConvNCdsCCPProcInfo inP icc sol) inP
    | otherwise = dontKnow (InfConvNCdsCCPDontKnow inP icc sol) inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          icc = buildInfCheckerCriteriaCTRS trs cp
          hasConds = not . Prelude.null $ ccpConds cp
          (notJoin, sol) = unsafePerformIO $!
                           do { 
                                solution <- solveInfChecker to icc
                                ; case solution of
                                    ("YES", Just sol) -> return (True, sol)
                                    (_, Just sol)     -> return (False, sol)
                            }

          
-- Pretty

instance (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV, Ord String
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (InfConvNCdsCCPProcInfo (Problem CPCRewriting conftrs)) where
  pPrint InfConvNCdsCCPProcInfo { inProblem = problem
                                , inICCrit = icc
                                , inSol = sol
                                }
    = pPrint problem
    Doc.<> text "\nInfeasible Convergence No Conditions CCP Processor:\n"
    Doc.<> text "Infeasible convergence?\n"
      Doc.<> text (translateFromRiskyChars sol)
    where trs = getConfProblem . getR $ problem



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Infeasible Convergence No Conditions CCP Processor
infConvNCdsCCPProc :: ( Processor info InfConvNCdsCCPProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
infConvNCdsCCPProc to = apply InfConvNCdsCCPProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveInfChecker :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
solveInfChecker timeout crit = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(VAR z)\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show $ pPrint (fst $ cpsConst crit) Doc.<+> text "== z" Doc.<> comma 
                                   Doc.<+> pPrint (snd $ cpsConst crit) Doc.<+> text "== z"
                 , frmStr ")"]

{-
-- | Generate infChecker input and try to get a solution
solveInfChecker :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), HasName idF, HasName idV, HasArity idF, Pretty idF, Pretty idV, Pretty trs) => Maybe Int -> trs -> InfCheckerCriteria -> IO (String, Maybe String)--IO (Maybe String)
solveInfChecker timeout trs crit = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (elems $ getCTRSRules trs)
                 , frmStr ")\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (freshCPVars crit)
                 , frmStr " "
                 , frmStr $ infCheckerJVN crit
                 , frmStr ")\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show . pPrint . fst . fromJust $ freshCP crit
                 , frmStr " == "
                 , frmStr $ infCheckerJVN crit
                 , frmStr ", "
                 , frmStr . noRiskyChars $ show . pPrint . snd . fromJust $ freshCP crit
                 , frmStr " == "
                 , frmStr $ infCheckerJVN crit
                 , frmStr ")"]
-}