{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.InliningOfCondsProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the InliningOfConds processor that, given a conditional rule,
-- inlines its conditions if possible and if R is oriented
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.InliningOfCondsProcessor (

-- * Exported functions

inlCondsProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasOSSignature(..))
import TRS.Terms(Term, applySigma, isVTerm, vars)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..))
import TRS.Rules(Rule(..), CRule(..), Cond(..), lhs, rhs, lcond, rcond)
import TRS.Terms(Term(..), isVTerm)
import Framework.Problem.Types(InConfCSCTRS (..),)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem)
import Framework.TRSAux (applySigmaLCond)
import Framework.ProblemConclusions (isOriented)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (singleP, Proof, Info (..), success)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), hcat, punctuate, hsep, comma)
import Control.Monad (MonadPlus)
import Data.Set as S (Set, elems, fromList, map, toList, delete, elemAt, size, union, unions, member)
import Data.Map as M (Map, insert, empty, fromList)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------


data InliningOfCondsProcessor  = InliningOfCondsProcessor

data InlCondsProcInfo problem = InlCondsProcInfo { inProblem :: problem }

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InConfCSCTRS trs
  , IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CRewriting conftrs)
  , Info info (InlCondsProcInfo (Problem CRewriting conftrs))
  ) => Processor info InliningOfCondsProcessor (Problem CRewriting conftrs) (Problem CRewriting conftrs) where
  apply InliningOfCondsProcessor inP
    | isOriented $ elems rules = singleP (InlCondsProcInfo newP) inP newP 
    | otherwise = singleP (InlCondsProcInfo inP) inP inP 
    where trs = getConfProblem . getR $ inP
          rules = getCTRSRules $ trs
          newRules = S.map inlineRuleConds rules
          newTRS = setCTRSRules newRules trs
          newP = mkProblem (CRewriting (getPSchemeFromProblem inP)) (InConfCSCTRS newTRS)

-- Pretty

instance (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CRewriting conftrs)
  , Ord idF, Ord idV
  ) => Pretty (InlCondsProcInfo (Problem CRewriting conftrs)) where
  pPrint InlCondsProcInfo { inProblem = problem}
    = text "\n\nInlining of Conditions Processor [STERN17]:\n"
      Doc.<> pPrint problem


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Inlining of Conditions Processor
inlCondsProc :: (Processor info InliningOfCondsProcessor o d
          , MonadPlus mp
          , Info info o
          , Info info d) => o -> Proof info mp d
inlCondsProc  = apply InliningOfCondsProcessor


inlineRuleConds :: (Ord idF, Ord idV, Eq idF, Eq idV) => CRule (Term idF idV) -> CRule (Term idF idV)
inlineRuleConds (rl :=> cds)
  | Prelude.null cds = (rl :=> cds)
  | otherwise = (nrl :=> (S.toList newCds))
  where (newCds, nrl) = inlineConds 0 (S.fromList cds) rl


inlineConds :: (Ord idF, Ord idV) => Int -> Set (Cond (Term idF idV)) -> Rule (Term idF idV) -> (Set (Cond (Term idF idV)), Rule (Term idF idV))
inlineConds i scds rl
  | i >= size scds = (scds, rl)
  | not $ isVTerm rc = inlineConds (i+1) scds rl
  | isInliningNotPossible v (lhs rl) lc rcs = inlineConds (i+1) scds rl
  | otherwise = inlineConds 0 nScds nrl
  where c = elemAt i scds
        rc = rcond c
        rcs = S.map rcond (delete c scds)
        (V v) = rc
        lc = lcond c
        subs = M.fromList [(v, lc)]
        aslc = applySigmaLCond subs
        nScds = S.map aslc (delete c scds)
        nr = applySigma subs $ rhs rl
        nrl = ((lhs rl) :-> nr)


isInliningNotPossible :: (Ord idV) => idV -> Term idF idV -> Term idF idV -> Set (Term idF idV) -> Bool
isInliningNotPossible v l lc rcs = inlNotPossible
  where lvars = vars l
        lcvars = vars lc
        vs = unions . S.elems $ S.map vars rcs
        inlNotPossible = (member v lvars) || (member v lcvars) || (member v vs)
