{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.NoCondsTransformCTRS
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that transforms CTRSs to TRSs if
-- it has no conditions
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.NoCondsTransformCTRS (

-- * Exported functions

noCondsTrCTRSProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasCRules (..), HasOSSignature(..), TRS (..), CTRS (..), Signature (..), OSTRS(..))
import TRS.Terms(Term (..), vars, matches)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..), HasMu(..))
import TRS.Rules(Rule(..), CRule(..), Cond(..), lhs, rhs, rcond, lcond, conds, rule)
import Framework.Problem.Types(InCPTermPCCSCTRS (..), InTermCSTRS (..), InConfCSTRS (..), InConfCSCTRS (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem, CriticalPairInfo (..), IsCPTermPCProblem (..), IsTProblem (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Framework.TRSAux (getDefinedSymbs, getConstructorSymbs)
import Utils.Auxiliar (isTrue)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (singleP, Proof, Info (..), success, dontKnow)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), hcat, hsep, comma, punctuate)
import Control.Monad (MonadPlus)
import Data.Set as S (Set, elems, fromList, toList, map, unions, elemAt, union, null, empty, insert, singleton, findMax, size, 
                      intersection)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------


data NoCondsTransformCTRS  = NoCondsTransformCTRS

data NoCondsTransformCTRSProcInfo problem = 
    NoCondsTransformCTRSProcInfo { inProblem    :: problem
                                 , inTRSString :: Maybe String
                                 }
    | NoCondsTransformCTRSDontKnow { inProblem    :: problem
                                   , inTRSString :: Maybe String
                                   }

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-}  (
    ctrs ~ CTRS idF idV (Term idF idV)
  , trs ~ TRS idF idV (Term idF idV)
  , ostrs1 ~ OSTRS ctrs
  , ostrs2 ~ OSTRS trs
  , conftrs1 ~ InConfCSCTRS ostrs1
  , conftrs2 ~ InConfCSTRS ostrs2
  , IsConfProblem conftrs1 ostrs1
  , IsConfProblem conftrs2 ostrs2
  , HasSignature ostrs1 idF idV
  , HasCRules ostrs1 (Term idF idV)
  , HasRules ostrs2 (Term idF idV)
  , HasMu idF
  , HasArity idF
  , HasName idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty ostrs1, Pretty ostrs2
  , Pretty (Problem CRewriting conftrs1)
  , Pretty (Problem Rewriting conftrs2)
  , Info info (NoCondsTransformCTRSProcInfo (Problem CRewriting conftrs1))
  ) => Processor info NoCondsTransformCTRS (Problem CRewriting conftrs1) (Problem Rewriting conftrs2) where
  apply NoCondsTransformCTRS inP
    | not crlsNoConds = dontKnow (NoCondsTransformCTRSDontKnow inP Nothing) inP
    | otherwise = singleP (NoCondsTransformCTRSProcInfo inP (Just showTRS)) inP newP
    where ctrs = getConfProblem . getR $ inP
          crls = getCTRSRules ctrs
          crlsNoConds = and . Prelude.map (\c -> Prelude.null $ conds c) $ (elems crls)
          trs = noCondsTransform ctrs
          newP = mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS trs)
          showTRS = show $ pPrint trs

-- Pretty

instance {-# OVERLAPPABLE #-} (
    ctrs ~ CTRS idF idV (Term idF idV)
  , ostrs ~ OSTRS ctrs
  , conftrs ~ InConfCSCTRS ostrs
  , IsConfProblem conftrs ostrs
  , HasSignature ostrs idF idV
  , HasCRules ostrs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty conftrs
  , Pretty (Problem CRewriting conftrs)
  ) => Pretty (NoCondsTransformCTRSProcInfo (Problem CRewriting conftrs)) where
  pPrint NoCondsTransformCTRSProcInfo { inProblem = problem
                                , inTRSString = Nothing
                                }
    = text "\n\nTransform No Conds CTRS Processor:\n"
      Doc.<> pPrint problem
  pPrint NoCondsTransformCTRSProcInfo { inProblem = problem
                                , inTRSString = (Just trsStr)
                                }
    = text "\n\nTransform No Conds CTRS Processor:\n"
      Doc.<> pPrint problem
      Doc.<> text "Resulting R:\n"
      Doc.<> text trsStr

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Transform No Conds CTRS Processor
noCondsTrCTRSProc :: (Processor info NoCondsTransformCTRS o d
          , MonadPlus mp
          , Info info o
          , Info info d) => o -> Proof info mp d
noCondsTrCTRSProc  = apply NoCondsTransformCTRS


noCondsTransform :: (
    ctrs ~ CTRS idF idV (Term idF idV)
  , trs ~ TRS idF idV (Term idF idV)
  , ostrs1 ~ OSTRS ctrs
  , ostrs2 ~ OSTRS trs
  , HasSignature ctrs idF idV
  , HasCRules ctrs (Term idF idV)
  , Eq idF, Eq idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , Ord idF, Ord idV) => ostrs1 -> ostrs2
noCondsTransform cProblem = ostrs2
  where 
    ostrs1@(OSTRS { ostrsprev = prev
                  , ossignature = ossig
                  }) = cProblem
    ctrs@(CTRS { 
            ctrsSignature
              = Signature { defined      = cdefined
                          , constructors = cconstructors
                          , relations    = relations
                          }
          , ctrsVariables = variables
          , ctrsRules     = crules
          , ctrsLabel     = label
          }) = prev
    rs = S.map (\cr -> rule cr) $ crules
    trs = TRS { trsSignature
                  = Signature { defined      = cdefined
                              , constructors = cconstructors
                              , relations    = relations
                              }
              , trsVariables = variables
              , trsRules     = rs
              , trsLabel     = label
              }
    ostrs2 = OSTRS { ostrsprev = trs
                   , ossignature = ossig
                   }
