{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.NormalRCondsCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that checks if rconds of rules are normal
-- with a level confluence verification cause
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.NormalRCondsCProcessor (

-- * Exported functions

normRCondsCProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(CRule (..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), 
                           getPSchemeFromProblem, IsCPPCProblem (..))
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPTermPCCSCTRS (..), InCPPCCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted, singleP)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPPCCRewriting (CPPCCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteriaRCondsCRule)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Normal RConds Processor
data NormalRCondsCProcessor = NormalRCondsCProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data NormalRCondsInfo problem = 
  NormalRCondsInfo { inProblem            :: problem
                   , inSol                :: String
                   , inProblemConclusions :: ProblemConclusionsCTRS
                   }
  | NormalRCondsDontKnow { inProblem            :: problem
                         , inSol                :: String
                         , inProblemConclusions :: ProblemConclusionsCTRS
                         }



-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermPCCSCTRS trs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idV
  , HasArity idF
  , HasName idF , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (NormalRCondsInfo (Problem CPTPCCRewriting conftrs))
  ) => Processor info NormalRCondsCProcessor (Problem CPTPCCRewriting conftrs) (Problem CPTPCCRewriting conftrs) where
  apply NormalRCondsCProcessor {timeout = to} inP
    -- | True = success (NormalRCondsInfo inP sol newPconc) inP
    | conf = success (NormalRCondsInfo inP sol newPconc) inP
    | Prelude.null cpis = dontKnow (NormalRCondsDontKnow inP sol newPconc) inP
    | otherwise = singleP (NormalRCondsInfo inP sol newPconc) inP 
                    (mkProblem (CPTPCCRewriting (getPSchemeFromProblem inP)) (InCPTermPCCSCTRS trs cpis (terminatingCTRS newPconc) newPconc))
    where trs = getConfProblem . getR $ inP
          cpis = getCPairsCPTPC . getR $ inP
          icc = buildInfCheckerCriteriaRCondsCRule trs
          (normtjs, sol) = unsafePerformIO $!
                      do { 
                          solution <- normalTJsInfChecker to icc
                          ; case solution of
                              ("YES", Just sol) -> return (Just True, sol)
                              ("NO", Just sol)  -> return (Just False, sol)
                              (_, Just sol)     -> return (Nothing, sol)
                      }
          pconc = getProbConcCPTPC . getR $ inP
          newPconc = problemConclusionCTRS trs cpis normtjs (rightStable pconc) (terminatingCTRS pconc) (opTerminatingCTRS pconc) (joinableCTRS pconc)
          conf = confluentCTRS newPconc

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPPCCSCTRS trs (CriticalPairInfo idF idV) ProblemConclusionsCTRS
  , IsConfProblem conftrs trs
  , IsCPPCProblem conftrs (CriticalPairInfo idF idV) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idV
  , HasArity idF
  , HasName idF, HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (NormalRCondsInfo (Problem CPPCCRewriting conftrs))
  ) => Processor info NormalRCondsCProcessor (Problem CPPCCRewriting conftrs) (Problem CPPCCRewriting conftrs) where
  apply NormalRCondsCProcessor {timeout = to} inP
    | conf = success (NormalRCondsInfo inP sol newPconc) inP -- True = success (NormalRCondsInfo inP sol newPconc) inP -- 
    | otherwise = dontKnow (NormalRCondsDontKnow inP sol newPconc) inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairsCPPC . getR $ inP
          icc = buildInfCheckerCriteriaRCondsCRule trs
          (normtjs, sol) = unsafePerformIO $!
                      do { 
                          solution <- normalTJsInfChecker to icc
                          ; case solution of
                              ("YES", Just sol) -> return (Just True, sol)
                              ("NO", Just sol)  -> return (Just False, sol)
                              (_, Just sol)     -> return (Nothing, sol)
                      }
          pconc = getProbConcCPPC . getR $ inP
          newPconc = problemConclusionCTRS trs cpis normtjs (rightStable pconc) (terminatingCTRS pconc) (opTerminatingCTRS pconc) (joinableCTRS pconc)
          conf = confluentCTRS newPconc

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty (Problem CPTPCCRewriting conftrs)
  ) => Pretty (NormalRCondsInfo (Problem CPTPCCRewriting conftrs)) where
  pPrint NormalRCondsInfo { inProblem = problem
                          , inSol = sol
                          , inProblemConclusions = pconc
                          }
    = pPrint problem
      Doc.<> text "\nNormal RConds Processor:\n"
      Doc.<> text (translateFromRiskyChars sol)
      Doc.<> pPrint pconc
    where trs = getConfProblem . getR $ problem

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPPCProblem conftrs (CriticalPairInfo idF idV) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty (Problem CPPCCRewriting conftrs)
  ) => Pretty (NormalRCondsInfo (Problem CPPCCRewriting conftrs)) where
  pPrint NormalRCondsInfo { inProblem = problem
                          , inSol = sol
                          , inProblemConclusions = pconc
                          }
    = pPrint problem
      Doc.<> text "\nNormal RConds Processor:\n"
      Doc.<> text (translateFromRiskyChars sol)
      Doc.<> pPrint pconc
    where trs = getConfProblem . getR $ problem

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Normal RConds Processor
normRCondsCProc :: ( Processor info NormalRCondsCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
normRCondsCProc to = apply NormalRCondsCProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
normalTJsInfChecker :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
normalTJsInfChecker timeout crit
  | Prelude.null ts = 
    do
      return ("YES", Just "No conditional rules")
  | otherwise = callToInfChecker timeout infProblem
  where
    ts = rCondTs crit
    js = rCondJVarNames crit
    n = length ts
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text " ") . map (\v -> pPrint v) $ (trsVars crit)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (crulesICC crit)
                 , frmStr ")\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . map (\j -> text j) $ js
                 , frmStr ")\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show . hsep . punctuate comma . map (\i -> pPrint (ts!!i) Doc.<> text " -> " Doc.<> text (js!!i)) 
                            $ [0..(n-1)]
                 , frmStr ")"]