{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.Prov9ConvCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check convergence of a ccp
-- calling to Prover9 [LucFSTTCS21-Cor17]
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.Prov9ConvCProcessor (

-- * Exported functions

prov9CvCProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(Rule (..), CRule (..), lcond, rcond, lhs, rhs, conds, rule, cRuleVars)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..), HasMu(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermPCCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.ExternalCalls (callToProver9)
import Framework.ExtCallsCriteria (Prover9Criteria (..), buildProver9Criteria)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, lparen, rparen, char, space)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data Prov9ConvCProcessor = Prov9ConvCProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data Prov9CvProcInfo problem idF idV = 
  Prov9CvProcInfo { inProblem            :: problem 
                  , inP9Crit             :: Prover9Criteria idF idV
                  , inCType              :: Int
                  , inSol                :: String
                  }
  | Prov9CvDontKnow { inProblem            :: problem 
                    , inP9Crit             :: Prover9Criteria idF idV
                    , inCType              :: Int
                    , inSol                :: String 
                    }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermPCCSCTRS trs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV, Ord String
  , Info info (Prov9CvProcInfo (Problem CPTPCCRewriting conftrs) idF idV)
  ) => Processor info Prov9ConvCProcessor (Problem CPTPCCRewriting conftrs) (Problem CPTPCCRewriting conftrs) where
  apply Prov9ConvCProcessor {timeout = to} inP
    | hasConds && not feasConds = dontKnow (Prov9CvDontKnow inP p9c ctype "INFEASIBLE_CONDS") inP
    | semiequational pconc = dontKnow (Prov9CvDontKnow inP p9c ctype "SEMIEQUATIONAL") inP
    | confIfJoinable = 
        unsafePerformIO $!
          do { 
              solution <- solveProver9 to trs p9c ctype
              ; case solution of
                  ("YES", Just sol) -> return (success (Prov9CvProcInfo inP p9c ctype sol) inP)
                  ("NO", Just sol)  -> return (dontKnow (Prov9CvDontKnow inP p9c ctype sol) inP) -- return (success (Prov9CvProcInfo inP p9c ctype sol) inP) -- 
                  (_, Just sol)     -> return (dontKnow (Prov9CvDontKnow inP p9c ctype sol) inP) -- return (success (Prov9CvProcInfo inP p9c ctype sol) inP) -- 
            }
    | otherwise = dontKnow (Prov9CvDontKnow inP p9c ctype "MAYBE") inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairsCPTPC . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          pconc = getProbConcCPTPC . getR $ inP
          term = terminatingCTRS pconc
          stDet = stronglyDet pconc
          nor = normalCTRS pconc
          opt = opTerminatingCTRS pconc
          ty = ctrsType pconc
          j = join pconc
          ctype
            | oriented pconc = 1
            | j = 2
            | otherwise = 0
          p9c = buildProver9Criteria trs cp
          hasConds = not . Prelude.null $ ccpConds cp
          feasConds = isTrue $ hasFeasCds cpi
          confIfJoinable = (
               ((isTrue opt) && (isTrue stDet))
            || ((isTrue term) && (isTrue nor) && (ty <= 1))
            || ((isTrue term) && j && (ty <= 1))
            )
          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV, Ord String
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPTPCCRewriting conftrs)
  ) => Pretty (Prov9CvProcInfo (Problem CPTPCCRewriting conftrs) idF idV) where
  pPrint Prov9CvProcInfo { inProblem = problem
                         , inP9Crit = crit
                         , inCType = ctype
                         , inSol = sol
                         }
    = pPrint problem
      Doc.<> text "\nProver9 CCP Convergence Checker Processor:\n"
      Doc.<> text "Proof:\n"
        Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Prover9 CCP Convergence Checker Processor
prov9CvCProc :: ( Processor info Prov9ConvCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
prov9CvCProc to = apply Prov9ConvCProcessor {timeout = to}


-- | Generate Prover9 input and try to get a solution
solveProver9 :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), Pretty idF, Pretty idV, Pretty trs, Ord String) => 
  Maybe Int -> trs -> Prover9Criteria idF idV -> Int -> IO (String, Maybe String)
solveProver9 timeout trs crit ctype
  = callToProver9 timeout prov9Problem
  where
    replacs
      | ctype == 1 =
        (hcat . map (\cr ->
        if Prelude.null (conds cr) then
            (hsep . map (\v -> text "all" Doc.<+> pPrint v) $ (elems $ cRuleVars cr))
            Doc.<+> lparen
              Doc.<+> text "red" Doc.<> lparen Doc.<> pPrint (lhs $ rule cr) Doc.<> comma 
              Doc.<> pPrint (rhs $ rule cr) Doc.<> rparen
            Doc.<+> rparen
          Doc.<> text ".\n"
        else (hsep . map (\v -> text "all" Doc.<+> pPrint v) $ (elems $ cRuleVars cr))
            Doc.<+> lparen
              Doc.<+> (hsep . punctuate (char '&') . map(\c -> 
                text "reds" Doc.<> lparen Doc.<> pPrint (lcond c) Doc.<> comma Doc.<> pPrint (rcond c) Doc.<> rparen
              ) $ (conds cr))
              Doc.<+> text "->"
              Doc.<+> text "red" Doc.<> lparen Doc.<> pPrint (lhs $ rule cr) Doc.<> comma
                Doc.<> pPrint (rhs $ rule cr) Doc.<> rparen
            Doc.<+> rparen
            Doc.<> text ".\n"
        ) $ (fromJust $ crulesNotSVars crit))
      | ctype == 2 =
        (hcat . map (\cr ->
        if Prelude.null (conds cr) then
            (hsep . map (\v -> text "all" Doc.<+> pPrint v) $ (elems $ cRuleVars cr))
            Doc.<+> lparen
              Doc.<+> text "red" Doc.<> lparen Doc.<> pPrint (lhs $ rule cr) Doc.<> comma 
              Doc.<> pPrint (rhs $ rule cr) Doc.<> rparen
            Doc.<+> rparen
          Doc.<> text ".\n"
        else (hsep . map (\v -> text "all" Doc.<+> pPrint v) $ (elems $ cRuleVars cr))
            Doc.<+> (hsep . map(\j -> text "exists" Doc.<+> text j) $ (cruleCondsJvs crit))
            Doc.<+> lparen
              Doc.<+> (hsep . punctuate (char '&') . map(\i -> 
                  text "reds" Doc.<> lparen 
                  Doc.<> pPrint (lcond ((conds cr)!!(i-1))) Doc.<> comma Doc.<> text ((cruleCondsJvs crit)!!(i - 1)) Doc.<> rparen
                Doc.<+> char '&'
                Doc.<+> text "reds" Doc.<> lparen 
                  Doc.<> pPrint (rcond ((conds cr)!!(i-1))) Doc.<> comma Doc.<> text ((cruleCondsJvs crit)!!(i - 1)) Doc.<> rparen
              ) $ [1..(length $ conds cr)])
              Doc.<+> text "->"
              Doc.<+> text "red" Doc.<> lparen Doc.<> pPrint (lhs $ rule cr) Doc.<> comma
                Doc.<> pPrint (rhs $ rule cr) Doc.<> rparen
            Doc.<+> rparen
            Doc.<> text ".\n"
        ) $ (fromJust $ crulesNotSVars crit))
      | otherwise = error "Relation not supported. CONDITIONTYPE not specified"
    cp = fromJust $ cpNotSVars crit
    goals
      | ctype == 1 =
          (hsep . map (\v -> text "all" Doc.<+> pPrint v) $ (fromJust $ ccpVars crit))
          Doc.<+> (hsep . map(\j -> text "exists" Doc.<+> text j) $ (ccpJoinVarNames crit))
          Doc.<+> lparen
          Doc.<> 
            (if Prelude.null (ccpConds cp)
            then joinGoals
            else 
              (hsep . punctuate (text " &") . map (\(cd) ->
                text "reds"
                  Doc.<> lparen
                    Doc.<> pPrint (lcond cd) Doc.<> comma Doc.<> pPrint (rcond cd)
                  Doc.<> rparen
              ) $ (ccpConds cp))
              Doc.<+> text "->"
              Doc.<+> joinGoals)
      | ctype == 2 =
          (hsep . map (\v -> text "all" Doc.<+> pPrint v) $ (fromJust $ ccpVars crit))
          Doc.<+> (hsep . map(\j -> text "exists" Doc.<+> text j) $ (ccpJoinVarNames crit))
          Doc.<+> lparen
          Doc.<> 
            (if Prelude.null (ccpConds cp)
            then joinGoals
            else 
              (hsep . punctuate (text " &") . map (\i ->
                text "reds"
                  Doc.<> lparen
                    Doc.<> pPrint (lcond ((ccpConds cp)!!(i-1))) Doc.<> comma Doc.<> text ((ccpJoinVarNames crit)!!(i))
                  Doc.<> rparen
                Doc.<+> char '&'
                Doc.<+> text "reds"
                  Doc.<> lparen
                    Doc.<> pPrint (rcond ((ccpConds cp)!!(i-1))) Doc.<> comma Doc.<> text ((ccpJoinVarNames crit)!!(i))
                  Doc.<> rparen
              ) $ [1..(length $ ccpConds cp)])
              Doc.<+> text "->"
              Doc.<+> joinGoals)
    joinGoals
      | ctype == 1 =
          text "reds" 
            Doc.<> lparen 
              Doc.<> pPrint (s cp) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
            Doc.<> rparen 
          Doc.<+> char '&'
          Doc.<+> text "reds"
            Doc.<> lparen
              Doc.<> pPrint (t cp) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
            Doc.<> rparen
          Doc.<> rparen
          Doc.<> text ".\n"
      | ctype == 2 =
          text "reds" 
            Doc.<> lparen 
              Doc.<> pPrint (s cp) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
            Doc.<> rparen 
          Doc.<+> char '&'
          Doc.<+> text "reds"
            Doc.<> lparen
              Doc.<> pPrint (t cp) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
            Doc.<> rparen
          Doc.<> rparen
          Doc.<> text ".\n"
      | otherwise = error "Relation not supported. CONDITIONTYPE not specified"
    prov9Problem = bconcat $ [ frmStr "formulas(sos).\n"
                             , frmStr "%Reflexivity\n"
                             , frmStr "reds(x,x).\n\n"
                             , frmStr "%Transitivity\n"
                             , frmStr "(red(x,y) & reds(y,z)) -> reds(x,z).\n\n"
                             , frmStr "%Congruence\n"
                             , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\cg -> pPrint cg) $ (congruence crit)
                             , frmStr "\n%Replacement\n"
                             , frmStr . noRiskyChars $ show replacs
                             , frmStr "\nend_of_list.\n\n"
                             , frmStr "formulas(goals).\n"
                             , frmStr . noRiskyChars $ show goals
                             , frmStr "end_of_list.\n\n"
                             , frmStr "%assign(sos_limit, -1).\n"
                             , frmStr "assign(max_megs, -1).\n"
                             , frmStr "set(quiet).\n"
                           ]
