{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.RightStableCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that checks if all CRules are right-stable
-- and recalculate level confluence
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.RightStableCProcessor (

-- * Exported functions

rightStbCProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..), isLinear, vars)
import TRS.Rules as R (CRule (..), Cond (..), rule, lcond, rcond, lhs, rhs, conds)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermPCProblem (..), 
                           getPSchemeFromProblem, IsCPProblem (..), IsCPPCProblem (..))
import Framework.Problem.Types (InCPTermPCCSCTRS (..), InCPCSCTRS (..), InCPPCCSCTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted, singleP)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.Problem.CPPCCRewriting (CPPCCRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.TRSAux (isGroundTerm)
import Framework.ExtCallsCriteria (termVarsToCons)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, char)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty, member)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Normal RConds Processor
data RightStableCProcessor = RightStableCProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data RightStableInfo problem = 
  RightStableInfo { inProblem            :: problem
                  , inRightStableConds   :: Maybe [RightStableCond]
                  , inRightStable        :: Bool
                  , inProblemConclusions :: ProblemConclusionsCTRS
                  }
  | RightStableDontKnow { inProblem            :: problem
                        , inRightStableConds   :: Maybe [RightStableCond]
                        , inRightStable        :: Bool
                        , inProblemConclusions :: ProblemConclusionsCTRS
                        }

data RightStableCond = 
  RightStableCond { rightStableC     :: Maybe Bool
                  , tiTerm           :: (Term IdFTRS IdVTRS)
                  , rsCondAchieved   :: Bool
                  , linConstructTerm :: Maybe Bool
                  , groundNormForm   :: Maybe Bool
                  , infCheckProof    :: Maybe String
                  }



-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermPCCSCTRS trs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasName idF , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (RightStableInfo (Problem CPTPCCRewriting conftrs))
  ) => Processor info RightStableCProcessor (Problem CPTPCCRewriting conftrs) (Problem CPTPCCRewriting conftrs) where
  apply RightStableCProcessor {timeout = to} inP
    -- | True = success (RightStableInfo inP Nothing True newPconc) inP
    | notHasConds && conf = success (RightStableInfo inP Nothing True newPconc) inP
    | notHasConds = singleP (RightStableInfo inP Nothing True newPconc) inP 
                      (mkProblem (CPTPCCRewriting (getPSchemeFromProblem inP)) (InCPTermPCCSCTRS trs cpis term newPconc))
    | conf = success (RightStableInfo inP (Just rightStableConds) rightStable newPconc) inP
    | otherwise = singleP (RightStableInfo inP (Just rightStableConds) rightStable newPconc) inP 
                    (mkProblem (CPTPCCRewriting (getPSchemeFromProblem inP)) (InCPTermPCCSCTRS trs cpis term newPconc))
    where trs = getConfProblem . getR $ inP
          crs = elems $ getCTRSRules trs
          notHasConds = and . map (\cr -> Prelude.null $ conds cr) $ crs
          cpis = getCPairsCPTPC . getR $ inP
          pconc = getProbConcCPTPC . getR $ inP
          term = terminatingCTRS pconc
          opt = opTerminatingCTRS pconc
          rightStableConds = getRightStable trs to
          rightStable
            | notHasConds = True
            | otherwise = and $ map (isTrue . rightStableC) rightStableConds
          newPconc = problemConclusionCTRS trs cpis Nothing (Just rightStable) term opt (joinableCTRS pconc)
          conf = confluentCTRS pconc

instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , conftrs2 ~ InCPPCCSCTRS trs (CriticalPairInfo idF idV) ProblemConclusionsCTRS
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsCPProblem conftrs1 (CriticalPairInfo idF idV)
  , IsCPPCProblem conftrs2 (CriticalPairInfo idF idV) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasName idF, HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (RightStableInfo (Problem CPCRewriting conftrs1))
  ) => Processor info RightStableCProcessor (Problem CPCRewriting conftrs1) (Problem CPPCCRewriting conftrs2) where
  apply RightStableCProcessor {timeout = to} inP
    -- | True = success (RightStableInfo inP (Just rightStableConds) rightStable pconc) inP
    | notHasConds = singleP (RightStableInfo inP Nothing True pconc) inP 
                      (mkProblem (CPPCCRewriting (getPSchemeFromProblem inP)) (InCPPCCSCTRS trs cpis pconc))
    | levConf = success (RightStableInfo inP (Just rightStableConds) rightStable pconc) inP
    | otherwise = singleP (RightStableInfo inP (Just rightStableConds) rightStable pconc) inP 
                    (mkProblem (CPPCCRewriting (getPSchemeFromProblem inP)) (InCPPCCSCTRS trs cpis pconc))
    where trs = getConfProblem . getR $ inP
          crs = elems $ getCTRSRules trs
          notHasConds = and . map (\cr -> Prelude.null $ conds cr) $ crs
          cpis = getCPairs . getR $ inP
          rightStableConds = getRightStable trs to
          rightStable
            | notHasConds = True
            | otherwise = and $ map (isTrue . rightStableC) rightStableConds
          joinable
            | Prelude.null cpis = Just True
            | otherwise = Nothing
          pconc = problemConclusionCTRS trs cpis Nothing (Just rightStable) Nothing Nothing joinable
          levConf = confluentCTRS pconc

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermPCProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV
  , Pretty (Problem CPTPCCRewriting conftrs)
  ) => Pretty (RightStableInfo (Problem CPTPCCRewriting conftrs)) where
  pPrint RightStableInfo { inProblem = problem
                         , inRightStableConds = (Just rscs)
                         , inRightStable = rs
                         , inProblemConclusions = pconc
                         }
    = pPrint problem
      Doc.<> text "\nRight Stable Processor:\n"
      Doc.<> text (if rs then "Right-stable CTRS" else "Not right-stable CTRS")
      Doc.<> text "\nJustification:"
      Doc.<> (hcat . punctuate (text "\n") . map (\c -> pPrint c) $ rscs)
      Doc.<> pPrint pconc
  pPrint RightStableInfo { inProblem = problem
                         , inRightStableConds = Nothing
                         , inRightStable = rs
                         , inProblemConclusions = pconc
                         }
    = pPrint problem
      Doc.<> text "\nRight Stable Processor:\n"
      Doc.<> text "Right-stable CTRS"
      Doc.<> text "\nJustification:"
      Doc.<> text "No conditional rules"
      Doc.<> pPrint pconc
    where trs = getConfProblem . getR $ problem

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (RightStableInfo (Problem CPCRewriting conftrs)) where
  pPrint RightStableInfo { inProblem = problem
                         , inRightStableConds = (Just rscs)
                         , inRightStable = rs
                         , inProblemConclusions = pconc
                         }
    = pPrint problem
      Doc.<> text "\nRight Stable Processor:\n"
      Doc.<> text (if rs then "Right-stable CTRS" else "Not right-stable CTRS")
      Doc.<> text "\nJustification:"
      Doc.<> (hcat . punctuate (text "\n") . map (\c -> pPrint c) $ rscs)
      Doc.<> pPrint pconc
    where trs = getConfProblem . getR $ problem
  pPrint RightStableInfo { inProblem = problem
                         , inRightStableConds = Nothing
                         , inRightStable = rs
                         , inProblemConclusions = pconc
                         }
    = pPrint problem
      Doc.<> text "\nRight Stable Processor:\n"
      Doc.<> text "Right-stable CTRS"
      Doc.<> text "\nJustification:"
      Doc.<> text "\nNo conditional rules"
      Doc.<> pPrint pconc


instance Pretty (RightStableCond) where
  pPrint (RightStableCond
    { rightStableC = rsc
    , tiTerm = t
    , rsCondAchieved = cAch
    , linConstructTerm = lc
    , groundNormForm = gn
    , infCheckProof = sol
    }) = text "\n\n-> Term right-stability conditions:\n"
    Doc.<+> text "Term:" Doc.<+> pPrint t Doc.<+> char '\n' 
    Doc.<+> text (if isNothing rsc then "Right-stability not checked (not needed)" 
                                   else if isTrue rsc then "Right-stable term" else "Not right-stable term") Doc.<+> char '\n' 
    Doc.<+> text (if isNothing lc then "Don't know if term is a linear constructor form (not needed)" 
                                   else if isTrue lc then "Linear constructor form" else "Not a linear constructor form") Doc.<+> char '\n' 
    Doc.<+> text (if isNothing gn then "Don't know if term is a ground normal form (not needed)" 
                                   else if isTrue gn then "Ground normal form" else "Not a ground normal form") Doc.<+> char '\n'
    Doc.<+> text (if cAch then "Right-stability condition achieved" else "Right-stability condition not achieved") Doc.<+> char '\n'
    Doc.<+> text (if isNothing sol then "A call to InfChecker wasn't needed" else "InfChecker proof:\n" ++ (translateFromRiskyChars $ fromJust sol))

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Right Stable Processor
rightStbCProc :: ( Processor info RightStableCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
rightStbCProc to = apply RightStableCProcessor {timeout = to}


getRightStable :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), Ord idV, Ord idF, HasArity idF, HasName idF, HasName idV, Pretty idF, Pretty idV, Pretty trs) 
                   => trs -> Maybe Int -> [RightStableCond]
getRightStable trs to
  | not $ and condAch = [RightStableCond Nothing (cleanRsTerms!!i) (condAch!!i) Nothing Nothing Nothing | i <- [0..(n-1)]]
  | otherwise = [RightStableCond (Just (tRightSt!!i)) (cleanRsTerms!!i) (condAch!!i) (Just (linConstr!!i)) (normGr!!i) (sols!!i)
                  | i <- [0..(n-1)]]
  where
    crs = elems $ getCTRSRules trs
    constructors = getTRSCSymbols trs
    rsMonad = areRightStable crs
    (rsTerms, condAch) = (map fst rsMonad, map snd rsMonad)
    cleanRsTerms = map termVarsToCons rsTerms
    n = length rsTerms
    linConstrFunc = isLinearConstructor constructors
    linConstr = map linConstrFunc rsTerms
    gnMonad = getNormGroundFormsNotLinC trs to linConstr rsTerms
    (normGr, sols) = (map fst gnMonad, map snd gnMonad)
    tRightSt = [isRightStableTerm (condAch!!i) (Just (linConstr!!i)) (normGr!!i) | i <-[0..(n-1)]]


isRightStableTerm :: Bool -> Maybe Bool -> Maybe Bool -> Bool
isRightStableTerm cAch lc ng = cAch && ((isTrue lc) || (isTrue ng))


isLinearConstructor :: (Ord idV, Ord idF) => Set idF -> Term idF idV -> Bool
isLinearConstructor constructors t@(F f tt) = (isLinear t) && (member f constructors)
isLinearConstructor constructors _ = False


getNormGroundFormsNotLinC :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), Ord idV, HasArity idF, HasName idF, HasName idV, Pretty idF, Pretty idV, Pretty trs) 
                              => trs -> Maybe Int -> [Bool] -> [Term idF idV] -> [(Maybe Bool, Maybe String)]
getNormGroundFormsNotLinC _ _ _ [] = []
getNormGroundFormsNotLinC trs to (lc:lcs) (t:ts)
  | lc = (Nothing, Nothing):(getNormGroundFormsNotLinC trs to lcs ts)
  | not $ isGroundTerm t = (Just False, Just "Not Ground Term"):(replicate n (Nothing, Nothing))
  | not $ isTrue normtjs = (normtjs, sol):(replicate n (Nothing, Nothing))
  | otherwise = (normtjs, sol):(getNormGroundFormsNotLinC trs to lcs ts)
  where n = length ts
        (normtjs, sol) = 
          unsafePerformIO $!
            do { 
              solution <- normGroundInfChecker to trs t
              ; case solution of
                  ("YES", sol) -> return (Just True, sol)
                  ("NO", sol)  -> return (Just False, sol)
                  (_, sol)     -> return (Nothing, sol)
            }


areRightStable :: (Ord idV, Ord idF) => [CRule (Term idF idV)] -> [((Term idF idV), Bool)]
areRightStable [] = []
areRightStable (c:crs)
  | n == 0 = areRightStable crs 
  | and $ map snd rsCSet = rsCSet ++ areRightStable crs
  | otherwise = rsCSet
  where r = R.rule c
        cs = conds c
        lvars = vars $ lhs r
        n = length cs
        rsCSet = [((rcond (cs!!i)), 
                   (S.null $ intersection (union lvars (union (condsVarsSetTo cs i) (vars $ lcond (cs!!i)))) (vars $ rcond (cs!!i)))) 
                    | i <- [0..(n-1)]]


condsVarsSetTo :: (Ord idV, Ord idF) => [Cond (Term idF idV)] -> Int -> Set idV
condsVarsSetTo [] _ = S.empty
condsVarsSetTo cs i
  | i > 0 = union sjTj (condsVarsSetTo cs j)
  | otherwise = S.empty
  where j = i - 1
        sjTj = union (vars $ lcond (cs!!j)) (vars $ rcond (cs!!j))


-- | Generate infChecker input and try to get a solution
normGroundInfChecker :: (HasSignature trs idF idV, HasCRules trs (Term idF idV), HasArity idF, HasName idF, HasName idV, Ord idV, Pretty idF, Pretty idV, Pretty trs) 
                          => Maybe Int -> trs -> Term idF idV -> IO (String, Maybe String)
normGroundInfChecker timeout trs t = callToInfChecker timeout infProblem
  where
    infProblem 
      = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                 , frmStr "(VAR "
                 , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (elems . getTRSVariables $ trs)
                 , frmStr ")\n"
                 , frmStr "(RULES\n"
                 , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (elems . getCTRSRules $ trs)
                 , frmStr ")\n"
                 , frmStr "(VAR z)\n"
                 , frmStr "(CONDITION "
                 , frmStr . noRiskyChars $ show $ pPrint t Doc.<> text " -> " Doc.<> text "z"
                 , frmStr ")"]