{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.TerminationCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that tries to check termination
-- of the CTRS using muterm
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.TerminationCProcessor (

-- * Exported functions

termCProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..))
import Framework.Problem.Types(InConfCSCTRS (..), InTermCSCTRS (..), InCPCSCTRS (..), InCPTermPCCSCTRS (..))
import Framework.Problems(IsConfProblem (..), IsTProblem (..), IsCPProblem (..), CriticalPair (..), CriticalPairInfo (..), 
                          IsCPTermPCProblem (..), getPSchemeFromProblem)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, Proof, Info (..), isSuccess, refuted, runProof)
import TRS.Rules(Rule (..), CRule (..), Cond (..), lcond, rcond)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.Problem.TCRewriting (TCRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Framework.ExternalCalls (callToMuTerm)
import Framework.ExtCallsCriteria (getJoinTerminationCRules)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, char)
import Control.Monad (MonadPlus, when)
import Data.Set as S (Set, elems, null, findMax)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)
import System.IO.Unsafe (unsafePerformIO)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

data TerminationCProcessor = TerminationCProcessor {timeout :: Maybe (Int)}

data TermProcInfo problem = 
  TermProcInfo { inProblem :: problem
               , inSol     :: String
               , inPCs     :: Maybe ProblemConclusionsCTRS
               }
  | TermDontKnow { inProblem :: problem
                 , inSol     :: String
                 , inPCs     :: Maybe ProblemConclusionsCTRS
                 } 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor


instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InCPCSCTRS trs (CriticalPairInfo idF idV)
  , conftrs2 ~ InCPTermPCCSCTRS trs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsCPProblem conftrs1 (CriticalPairInfo idF idV)
  , IsCPTermPCProblem conftrs2 (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (TermProcInfo (Problem CPCRewriting conftrs1))
  ) => Processor info TerminationCProcessor (Problem CPCRewriting conftrs1) (Problem CPTPCCRewriting conftrs2) where
  apply TerminationCProcessor {timeout = to} inP
    | conf = success (TermProcInfo inP sol (Just pconc)) inP --True = success (TermProcInfo inP sol (Just pconc)) inP --
    | otherwise = singleP (TermProcInfo inP sol (Just pconc)) inP 
                 (mkProblem (CPTPCCRewriting (getPSchemeFromProblem inP)) (InCPTermPCCSCTRS trs cpis term pconc))
    where trs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          joinable
            | Prelude.null cpis = Just True
            | otherwise = Nothing
          pconc = problemConclusionCTRS trs cpis Nothing Nothing term Nothing joinable
          (term, sol) =  
            unsafePerformIO $!
              do { 
                solution <- solveMuTerm pconc to trs
               ; case solution of
                    ("YES", Just sol) -> return ((Just True, sol))
                    ("NO", Just sol)  -> return ((Just False, sol))
                    (_, Just sol)     -> return ((Nothing, sol))
               }
          conf = confluentCTRS pconc --((Prelude.null cpis) && (isTrue term) && (isTrue $ almostNormalCTRS pconc)) || (levelConfluent pconc)

          
-- Pretty


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasCRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (TermProcInfo (Problem CPCRewriting conftrs)) where
  pPrint TermProcInfo { inProblem = problem
                      , inSol = sol
                      , inPCs = pcs
                      }
    = text "\nTermination Processor:\n"
      Doc.<> text "Terminating?\n"
      Doc.<> (hcat . punctuate (text "") . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
      Doc.<> pPrint (fromJust pcs)
    where trs = getConfProblem . getR $ problem


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Termination Processor
termCProc :: ( Processor info TerminationCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
termCProc to = apply TerminationCProcessor {timeout = to}


-- | Generate muterm input and try to get a solution
solveMuTerm :: (HasSignature trs idF idV, HasName idV, HasCRules trs (Term idF idV), Pretty idF, Pretty idV, Pretty trs, Ord idV)
            => ProblemConclusionsCTRS -> Maybe Int -> trs -> IO (String, Maybe String)
solveMuTerm pc timeout trs
  = callToMuTerm timeout termProblem
  where vs = elems $ getTRSVariables trs
        cRules = elems $ getCTRSRules trs
        joinCRules = getJoinTerminationCRules cRules vs Nothing
        jvars = Prelude.map snd $ concatMap snd joinCRules
        termProblem
          | oriented pc = 
              bconcat $ [ frmStr "(CHECK TERMINATION)\n"
                        , frmStr "(CONDITIONTYPE ORIENTED)\n"
                        , frmStr "(VAR "
                        , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (vs)
                        , frmStr ")\n"
                        , frmStr "(RULES\n"
                        , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\(r :=> cs) -> 
                                  if Prelude.null cs then pPrint r else pPrint r Doc.<+> char '|' 
                                  Doc.<+> (hsep . punctuate comma . map(\c -> pPrint (lcond c) Doc.<+> text "->" Doc.<+> pPrint (rcond c)) $ cs)
                                  ) $ (cRules)
                        , frmStr ")"
                        ]
          | join pc = 
              bconcat $ [ frmStr "(CHECK TERMINATION)\n"
                        , frmStr "(CONDITIONTYPE ORIENTED)\n"
                        , frmStr "(VAR "
                        , frmStr . noRiskyChars $ show . hsep . map (\v -> text v) $ (jvars)
                        , frmStr " "
                        , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (vs)
                        , frmStr ")\n"
                        , frmStr "(RULES\n"
                        , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\((r,cs)) -> 
                                  if Prelude.null cs then pPrint r else pPrint r Doc.<+> char '|' 
                                  Doc.<+> (hsep . punctuate comma . map(\(c,s) -> pPrint (lcond c) 
                                                                          Doc.<+> text "->" Doc.<+> text s Doc.<> comma 
                                                                        Doc.<+> pPrint (rcond c) 
                                                                          Doc.<+> text "->" Doc.<+> text s
                                                                      ) $ cs)
                                  ) $ (joinCRules)
                        , frmStr ")"
                        ]
          | otherwise = 
              bconcat $ [ frmStr "(CHECK TERMINATION)\n"
                        , frmStr "(CONDITIONTYPE ORIENTED)\n"
                        , frmStr "(VAR "
                        , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (vs)
                        , frmStr ")\n"
                        , frmStr "(RULES\n"
                        , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\(r :=> cs) -> 
                                  if Prelude.null cs then pPrint r else pPrint r Doc.<+> char '|' 
                                  Doc.<+> (hsep . punctuate comma . map(\c -> pPrint (lcond c) Doc.<+> text "->" Doc.<+> pPrint (rcond c)) $ cs)
                                  ) $ (cRules)
                        , frmStr ")"
                        ]
          