{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.UOptTransformCTRSProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that transforms CTRSs if conditions
-- are reached
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.UOptTransformCTRSProcessor (

-- * Exported functions

uOptTransformCTRSProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasCRules (..), HasOSSignature(..), TRS (..), CTRS (..), Signature (..), OSTRS(..))
import TRS.Terms(Term (..), vars, matches)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..), HasMu(..))
import TRS.Rules(Rule(..), CRule(..), Cond(..), lhs, rhs, rcond, lcond, conds, rule)
import Framework.Problem.Types(InCPCSCTRS (..), InConfCSTRS (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem, CriticalPairInfo (..), IsCPProblem (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Framework.TRSAux (getDefinedSymbs, getConstructorSymbs)
import Utils.Auxiliar (isTrue)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (singleP, Proof, Info (..), success, dontKnow)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), hcat, hsep, comma, punctuate)
import Control.Monad (MonadPlus)
import Data.Set as S (Set, elems, fromList, toList, map, unions, elemAt, union, null, empty, insert, singleton, findMax, size, 
                      intersection)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------


data UOptTransformCTRSProcessor  = UOptTransformCTRSProcessor

data UOptTransformCTRSProcInfo problem = 
    UOptTransformCTRSProcInfo { inProblem    :: problem
                           , inTRSString :: Maybe String
                           }
    | UOptTransformCTRSDontKnow { inProblem    :: problem
                             , inTRSString :: Maybe String
                             }

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    ctrs ~ CTRS idF idV (Term idF idV)
  , trs ~ TRS idF idV (Term idF idV)
  , ostrs1 ~ OSTRS ctrs
  , ostrs2 ~ OSTRS trs
  , conftrs1 ~ InCPCSCTRS ostrs1 (CriticalPairInfo idF idV)
  , conftrs2 ~ InConfCSTRS ostrs2
  , IsConfProblem conftrs1 ostrs1
  , IsConfProblem conftrs2 ostrs2
  , IsCPProblem conftrs1 (CriticalPairInfo idF idV)
  , HasSignature ostrs1 idF idV
  , HasCRules ostrs1 (Term idF idV)
  , HasRules ostrs2 (Term idF idV)
  , HasMu idF
  , HasArity idF
  , HasName idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty ostrs1, Pretty ostrs2
  , Pretty (Problem CPCRewriting conftrs1)
  , Pretty (Problem Rewriting conftrs2)
  , Info info (UOptTransformCTRSProcInfo (Problem CPCRewriting conftrs1))
  ) => Processor info UOptTransformCTRSProcessor (Problem CPCRewriting conftrs1) (Problem Rewriting conftrs2) where
  apply UOptTransformCTRSProcessor inP
    -- | True = success (UOptTransformCTRSProcInfo inP (Just showTRS)) inP
    | (determ pconc) && ((ctrsType pconc) <= 3) = singleP (UOptTransformCTRSProcInfo inP (Just showTRS)) inP newP
    | otherwise = dontKnow (UOptTransformCTRSDontKnow inP Nothing) inP
    where ctrs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          joinable
            | Prelude.null cpis = Just True
            | otherwise = Nothing
          pconc = problemConclusionCTRS ctrs cpis Nothing Nothing Nothing Nothing joinable
          trs = uTransformCTRS ctrs
          newP = mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS trs)
          showTRS = show $ pPrint trs

-- Pretty

instance {-# OVERLAPPABLE #-} (
    ctrs ~ CTRS idF idV (Term idF idV)
  , ostrs ~ OSTRS ctrs
  , conftrs ~ InCPCSCTRS ostrs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs ostrs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature ostrs idF idV
  , HasCRules ostrs (Term idF idV)
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty conftrs
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (UOptTransformCTRSProcInfo (Problem CPCRewriting conftrs)) where
  pPrint UOptTransformCTRSProcInfo { inProblem = problem
                                , inTRSString = Nothing
                                }
    = text "\n\nU_opt Transform CTRS Processor:\n"
      Doc.<> pPrint problem
  pPrint UOptTransformCTRSProcInfo { inProblem = problem
                                , inTRSString = (Just trsStr)
                                }
    = text "\n\nU_opt Transform CTRS Processor:\n"
      Doc.<> pPrint problem
      Doc.<> text "Resulting U_opt(R):\n"
      Doc.<> text trsStr

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Transform CTRS Processor
uOptTransformCTRSProc :: (Processor info UOptTransformCTRSProcessor o d
          , MonadPlus mp
          , Info info o
          , Info info d) => o -> Proof info mp d
uOptTransformCTRSProc  = apply UOptTransformCTRSProcessor


uTransformCTRS :: (
    ctrs ~ CTRS idF idV (Term idF idV)
  , trs ~ TRS idF idV (Term idF idV)
  , ostrs1 ~ OSTRS ctrs
  , ostrs2 ~ OSTRS trs
  , HasSignature ctrs idF idV
  , HasCRules ctrs (Term idF idV)
  , Eq idF, Eq idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , Ord idF, Ord idV) => ostrs1 -> ostrs2
uTransformCTRS cProblem = ostrs2
  where ostrs1@(OSTRS { ostrsprev = prev
                      , ossignature = ossig
                      }) = cProblem

        ctrs@(CTRS { 
                ctrsSignature
                  = Signature { defined      = cdefined
                              , constructors = cconstructors
                              , relations    = relations
                              }
              , ctrsVariables = variables
              , ctrsRules     = crules
              , ctrsLabel     = label
              }) = prev

        allSymbs = cdefined `union` cconstructors
        optRs = uTransformCRules crules (findMax allSymbs)

        newDefined = cdefined `union` (getDefinedSymbs optRs)
        newConstructors = cconstructors `union` (getConstructorSymbs optRs)
        trs = TRS { trsSignature
                         = Signature { defined      = newDefined
                                     , constructors = newConstructors
                                     , relations    = relations
                                     }
                   , trsVariables = variables
                   , trsRules     = optRs
                   , trsLabel     = label
                   }
        ostrs2 = OSTRS { ostrsprev = trs
                       , ossignature = ossig
                       }
         

uTransformCRules :: (HasName idF, HasArity idF, HasMu idF, Ord idF, Ord idV) => Set (CRule (Term idF idV)) -> idF -> Set (Rule (Term idF idV))
uTransformCRules crules s
  | S.null crules = S.empty
  | n == 0 = insert rl (uTransformCRules crs s)
  | otherwise = union newRs (uTransformCRules crs (findMax $ fromList uSymbs))
  where cr = elemAt 0 crules
        crs = S.fromList . tail . S.elems $ crules
        cds = conds cr
        n = length cds
        rl = rule cr
        m = getId s
        newM = m
        uSymbs = [(setId (newM + i) . (setName $ Just ("U" ++ (show newM) ++ "_" ++ (show i)))) s | i <- [1..n]]
        newRs = fromList $ createUOptRules uSymbs Nothing rl cds 0



createUOptRules :: (HasArity idF, HasMu idF, Ord idF, Ord idV) 
  => [idF] -> Maybe (Term idF idV) -> Rule (Term idF idV) -> [Cond (Term idF idV)] -> Int -> [Rule (Term idF idV)]
createUOptRules [] Nothing _ _ _ = []
createUOptRules [] (Just t) (l :-> r) cds i = [ti :-> r]
  where F f (_:vs) = t
        cd = cds !! i
        subt = (rcond cd):vs
        n = length subt
        ti = F ((setFArity n . setFMu [1..n]) f) subt
createUOptRules (uf:ufs) Nothing rl cds i = ((lhs rl) :-> ti):(createUOptRules ufs (Just ti) rl cds i)
  where subt = (lcond $ head cds):yi
        n = length subt
        vs1 = vars $ lhs rl
        vsti = unions . S.elems $ S.map (vars . rcond) (fromList $ Prelude.drop i cds)
        vssip1 = unions . S.elems $ S.map (vars . lcond) (fromList $ Prelude.drop (i + 1) cds)
        vsr = vars $ rhs rl
        vs2 = vsti `union` vssip1 `union` vsr
        yi = toList $ S.map (\v -> V v) (vs1 `intersection` vs2)
        ti = F ((setFArity n . setFMu [1..n]) uf) subt     
createUOptRules (uf:ufs) (Just t) rl cds i = (t1 :-> t2):(createUOptRules ufs (Just t2) rl cds j)
  where F f (_:vs) = t
        cd1 = cds !! i
        j = i + 1
        subt1 = (rcond cd1):vs
        n1 = length subt1
        t1 = F (setFArity n1 f) subt1
        cd2 = cds !! j
        vsl = vars $ lhs rl
        vstitk = unions . S.elems $ S.map (vars . rcond) (fromList $ Prelude.take j cds)
        vs1 = vsl `union` vstitk
        vstidr = unions . S.elems $ S.map (vars . rcond) (fromList $ Prelude.drop j cds)
        vssip1 = unions . S.elems $ S.map (vars . lcond) (fromList $ Prelude.drop (j + 1) cds)
        vsr = vars $ rhs rl
        vs2 = vstidr `union` vssip1 `union` vsr
        yi = toList $ S.map (\v -> V v) (vs1 `intersection` vs2)
        subt2 = (lcond cd2):yi
        n2 = length subt2
        t2 = F ((setFArity n2 . setFMu [1..n2]) uf) subt2
