{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.CRewriting.UnderlyingTRSTermCProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that tries to check termination
-- of the underlying TRS of a CS-TRS using muterm
--
-----------------------------------------------------------------------------

module Framework.Processor.CRewriting.UnderlyingTRSTermCProcessor (

-- * Exported functions

undRTermCProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..), TRS (..), CTRS (..), Signature (..), OSTRS(..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity (..), HasMu (..))
import TRS.Rules(Rule (..), CRule (..), Cond (..), rule)
import Framework.Problem.Types(InConfCSCTRS (..), InCPCSCTRS (..), InCPTermPCCSCTRS (..))
import Framework.Problems(IsConfProblem (..), IsCPProblem (..), CriticalPair (..), CriticalPairInfo (..), 
                          IsCPTermPCProblem (..), getPSchemeFromProblem)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, Proof, Info (..), isSuccess, refuted, runProof)
import Framework.Problem.CRewriting (CRewriting (..))
import Framework.Problem.CPCRewriting (CPCRewriting (..))
import Framework.Problem.CPTPCCRewriting (CPTPCCRewriting (..))
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..), problemConclusionCTRS)
import Framework.TRSAux (getDefinedSymbs, getConstructorSymbs)
import Framework.ExternalCalls (callToMuTerm)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, char)
import Control.Monad (MonadPlus, when)
import Data.Set as S (Set, elems, null, findMax, map, union)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)
import System.IO.Unsafe (unsafePerformIO)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

data UnderlyingTRSTermCProcessor = UnderlyingTRSTermCProcessor {timeout :: Maybe (Int)}

data UndRTermProcInfo problem = 
  UndRTermProcInfo { inProblem :: problem
               , inSol       :: Maybe String
               , inPCs       :: Maybe ProblemConclusionsCTRS
               , inTRSString :: Maybe String
               }
  | UndRTermDontKnow { inProblem :: problem
                 , inSol       :: Maybe String
                 , inPCs       :: Maybe ProblemConclusionsCTRS
                 , inTRSString :: Maybe String
                 } 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor


instance {-# OVERLAPPABLE #-} (
    ctrs ~ CTRS idF idV (Term idF idV)
  , ostrs ~ OSTRS ctrs
  , conftrs1 ~ InCPCSCTRS ostrs (CriticalPairInfo idF idV)
  , conftrs2 ~ InCPTermPCCSCTRS ostrs (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , IsConfProblem conftrs1 ostrs
  , IsConfProblem conftrs2 ostrs
  , IsCPProblem conftrs1 (CriticalPairInfo idF idV)
  , IsCPTermPCProblem conftrs2 (CriticalPairInfo idF idV) (Maybe(Bool)) ProblemConclusionsCTRS
  , HasSignature ostrs idF idV
  , HasCRules ostrs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty ostrs, Pretty (Signature idF)
  , HasArity idF
  , HasName idF, HasName idV
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (UndRTermProcInfo (Problem CPCRewriting conftrs1))
  ) => Processor info UnderlyingTRSTermCProcessor (Problem CPCRewriting conftrs1) (Problem CPTPCCRewriting conftrs2) where
  apply UnderlyingTRSTermCProcessor {timeout = to} inP
    | conf = success (UndRTermProcInfo inP (Just sol) (Just pconc) (Just showTRS)) inP --True = success (UndRTermProcInfo inP (Just sol) (Just pconc) (Just showTRS)) inP --
    | ctrsType pconc > 2 = dontKnow (UndRTermProcInfo inP Nothing (Just pconc) Nothing) inP
    | otherwise = singleP (UndRTermProcInfo inP (Just sol) (Just pconc) (Just showTRS)) inP 
                 (mkProblem (CPTPCCRewriting (getPSchemeFromProblem inP)) (InCPTermPCCSCTRS ctrs cpis undrterm pconc))
    where ctrs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          joinable
            | Prelude.null cpis = Just True
            | otherwise = Nothing
          pconc = problemConclusionCTRS ctrs cpis Nothing Nothing undrterm Nothing joinable
          trs = toUnderlyingTRS ctrs
          (undrterm, sol) =  
            unsafePerformIO $!
              do { 
                solution <- solveMuTerm to trs
               ; case solution of
                    ("YES", Just sol) -> return ((Just True, sol))
                    ("NO", Just sol)  -> return ((Just False, sol))
                    (_, Just sol)     -> return ((Nothing, sol))
               }
          conf = confluentCTRS pconc --((Prelude.null cpis) && (isTrue undrterm) && (isTrue $ almostNormalCTRS pconc)) || (levelConfluent pconc)
          showTRS = show $ pPrint trs

          
-- Pretty


instance {-# OVERLAPPABLE #-} (
    ctrs ~ CTRS idF idV (Term idF idV)
  , ostrs ~ OSTRS ctrs
  , conftrs ~ InCPCSCTRS ostrs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs ostrs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature ostrs idF idV
  , HasCRules ostrs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty ostrs
  , Pretty (Problem CPCRewriting conftrs)
  ) => Pretty (UndRTermProcInfo (Problem CPCRewriting conftrs)) where
  pPrint UndRTermProcInfo { inProblem = problem
                          , inSol = (Just sol)
                          , inPCs = (Just pcs)
                          , inTRSString = (Just trsStr)
                          }
    = text "\nUnderlying TRS Termination Processor:\n"
      Doc.<> text "\nResulting Underlying TRS:\n"
      Doc.<> text trsStr
      Doc.<> text "\nUnderlying TRS terminating?\n"
      Doc.<> (hcat . punctuate (text "") . Prelude.map (\c -> pPrint c) $ (translateFromRiskyChars sol))
      Doc.<> pPrint pcs
    where trs = getConfProblem . getR $ problem


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Underlying TRS Termination Processor
undRTermCProc :: ( Processor info UnderlyingTRSTermCProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
undRTermCProc to = apply UnderlyingTRSTermCProcessor {timeout = to}


-- | Generate muterm input and try to get a solution
solveMuTerm :: (HasSignature trs idF idV, HasRules trs (Term idF idV), Pretty idF, Pretty idV, Pretty trs) => Maybe Int -> trs -> IO (String, Maybe String)
solveMuTerm timeout trs
  = callToMuTerm timeout termProblem
  where
      termProblem = bconcat $ [frmStr "(VAR "
                              , frmStr . noRiskyChars $ show . hsep . Prelude.map (\v -> pPrint v) $ (elems . getTRSVariables $ trs)
                              , frmStr ")\n"
                              , frmStr "(RULES\n"
                              , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . Prelude.map (\r -> pPrint r 
                                        ) $ (elems . getTRSRules $ trs)
                              , frmStr ")"
                              ]


toUnderlyingTRS :: (
    ctrs ~ CTRS idF idV (Term idF idV)
  , trs ~ TRS idF idV (Term idF idV)
  , ostrs1 ~ OSTRS ctrs
  , ostrs2 ~ OSTRS trs
  , HasSignature ctrs idF idV
  , HasCRules ctrs (Term idF idV)
  , Eq idF, Eq idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , Ord idF, Ord idV) => ostrs1 -> ostrs2
toUnderlyingTRS cProblem = ostrs2
  where ostrs1@(OSTRS { ostrsprev = prev
                      , ossignature = ossig
                      }) = cProblem

        ctrs@(CTRS { 
                ctrsSignature
                  = Signature { defined      = cdefined
                              , constructors = cconstructors
                              , relations    = relations
                              }
              , ctrsVariables = variables
              , ctrsRules     = crules
              , ctrsLabel     = label
              }) = prev

        allSymbs = cdefined `union` cconstructors
        undRules = S.map rule crules
        newDefined = getDefinedSymbs undRules
        newConstructors = getConstructorSymbs undRules
        trs = TRS { trsSignature
                         = Signature { defined      = newDefined
                                     , constructors = newConstructors
                                     , relations    = relations
                                     }
                   , trsVariables = variables
                   , trsRules     = undRules
                   , trsLabel     = label
                   }
        ostrs2 = OSTRS { ostrsprev = trs
                       , ossignature = ossig
                       }