{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.CSRConverterFromCanonicalProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that converts a TRS to a CSR if conditions
-- from Gramlich and Lucas 2006 are achieved starting from canonical 
-- \mu-replacement map (most restrictive canonical \mu-replacement map)
-- [CSUR20]
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.CSRConverterFromCanonicalProcessor (

-- * Exported functions

csrcfCanProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), isLeftLinear)
import TRS.Terms(Term(..), Substitution (..), isLinear, applySigma, matches, replace, vars)
import TRS.Identifiers (HasName (..), HasArity (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem, IsTProblem (..))
import Framework.Problem.Types (InConfCSTRS (..), InTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, Proof, Info (..))
import TRS.Rules(Rule(..), lhs, rhs, freshRule)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.TRSAux (isCSTRSSymb, getMRCanonRepMap, setRepMap, getTopRepMap)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Set as S (Set, elems, empty)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data CSRConverterFromCanonicalProcessor = CSRConverterFromCanonicalProcessor

data CSRCFCanProcInfo problem idF = 
  CSRCFCanProcInfo { inProblem :: problem
                   , inConditionsReached :: Bool
                   , inReplacementMap1 :: Map idF [Int]
                   , inReplacementMap2 :: Map idF [Int]
                   }
  | CSRCFCanDontKnow { inProblem :: problem
                     , inConditionsReached :: Bool
                     , inReplacementMap1 :: Map idF [Int]
                     , inReplacementMap2 :: Map idF [Int]
                     }



-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InConfCSTRS trs
  , IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (CSRCFCanProcInfo (Problem Rewriting conftrs) idF)
  ) => Processor info CSRConverterFromCanonicalProcessor (Problem Rewriting conftrs) (Problem Rewriting conftrs) where
  apply CSRConverterFromCanonicalProcessor inP
    -- | True = success (CSRCFCanProcInfo inP True rm1 rm2) inP
    | isCstrs = dontKnow (CSRCFCanDontKnow inP False rm1 rm1) inP
    | isNothing tc = dontKnow (CSRCFCanDontKnow inP False rm1 rm1) inP
    | isLeftLinear trs = singleP (CSRCFCanProcInfo newP True rm1 rm2) inP newP
    | otherwise = dontKnow (CSRCFCanDontKnow inP False rm1 rm1) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          isCstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          rm1 = getTopRepMap trs
          tc = toCSR trs
          (csr,rm2) = fromJust tc
          newP = mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS csr)


instance {-# OVERLAPPABLE #-} (
    conftrs ~ InTermCSTRS trs (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsTProblem conftrs (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (CSRCFCanProcInfo (Problem TRewriting conftrs) idF)
  ) => Processor info CSRConverterFromCanonicalProcessor (Problem TRewriting conftrs) (Problem TRewriting conftrs) where
  apply CSRConverterFromCanonicalProcessor inP
    -- | True = success (CSRCFCanProcInfo inP True rm1 rm2) inP
    | isCstrs = dontKnow (CSRCFCanDontKnow inP False rm1 rm1) inP
    | isNothing tc = dontKnow (CSRCFCanDontKnow inP False rm1 rm1) inP
    | isLeftLinear trs = singleP (CSRCFCanProcInfo newP True rm1 rm2) inP newP
    | otherwise = dontKnow (CSRCFCanDontKnow inP False rm1 rm1) inP
    where trs = getConfProblem . getR $ inP
          term = getTermination . getR $ inP
          rules = elems . getTRSRules $ trs
          isCstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          rm1 = getTopRepMap trs
          tc = toCSR trs
          (csr,rm2) = fromJust tc
          newP = mkProblem (TRewriting (getPSchemeFromProblem inP)) (InTermCSTRS csr term)

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Eq idF
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (CSRCFCanProcInfo (Problem Rewriting conftrs) idF) where
  pPrint CSRCFCanProcInfo { inProblem = problem
                          , inConditionsReached = cr
                          , inReplacementMap1 = rm1
                          , inReplacementMap2 = rm2
                          }
    = text "\nCSR Converter From Canonical u-Replacement Map Processor [CSUR20]:"
      Doc.<> (
      if not cr
      then text "\n\nConditions not reached\n"
      else text "\n\nOriginal Replacement Map:\n"
      Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList rm1))
      Doc.<> text "\n\nNew Replacement Map:\n"
      Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList rm2))
      Doc.<> (
        if rm1 == rm2
        then text "\n\nA more restrictive canonical u-Replacement Map cannot be found\n"
        else text "\n\nNew problem:\n"
             Doc.<> pPrint problem
        )
      )

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsTProblem conftrs (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Eq idF
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem TRewriting conftrs)
  ) => Pretty (CSRCFCanProcInfo (Problem TRewriting conftrs) idF) where
  pPrint CSRCFCanProcInfo { inProblem = problem
                          , inConditionsReached = cr
                          , inReplacementMap1 = rm1
                          , inReplacementMap2 = rm2
                          }
    = text "\nCSR Converter From Canonical u-Replacement Map Processor [CSUR20]:"
      Doc.<> (
      if not cr
      then text "\n\nConditions not reached\n"
      else text "\n\nOriginal Replacement Map:\n"
      Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList rm1))
      Doc.<> text "\n\nNew Replacement Map:\n"
      Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList rm2))
      Doc.<> (
        if rm1 == rm2
        then text "\n\nA more restrictive canonical u-Replacement Map cannot be found\n"
        else text "\n\nNew problem:\n"
             Doc.<> pPrint problem
        )
      )


      


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the CSR Converter From Canonical u-Replacement Map Processor
csrcfCanProc :: ( Processor info CSRConverterFromCanonicalProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
csrcfCanProc = apply CSRConverterFromCanonicalProcessor


toCSR :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasArity idF, HasMu idF, Ord idF, Ord idV) => trs -> Maybe (trs, Map idF [Int])
toCSR trs
  | isNothing rm = Nothing
  | otherwise = Just (newTRS, (fromJust rm))
  where rm = getMRCanonRepMap trs
        newTRS = setRepMap (fromJust rm) trs