{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.CSRConverterFromTopProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that converts a TRS to some CSRs if conditions
-- from Gramlich and Lucas 2006 are achieved starting from top \mu-replacement map
-- (less restrictive canonical \mu-replacement maps) [CSUR20]
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.CSRConverterFromTopProcessor (

-- * Exported functions

csrcfTopProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), addRepMapTRS, isLeftLinear)
import TRS.Terms(Term(..), Substitution (..), isLinear, applySigma, matches, replace, vars)
import TRS.Identifiers (HasName (..), HasArity (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem, IsTProblem (..))
import Framework.Problem.Types (InConfCSTRS (..), InTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, andP, Proof, Info (..))
import TRS.Rules(Rule(..), lhs, rhs, freshRule)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.TRSAux (isCSTRSSymb, getLRCanoneRepMaps, setRepMap, getTopRepMap, getRepMap)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes, int)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Set as S (Set, elems, empty, member, delete)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data CSRConverterFromTopProcessor = CSRConverterFromTopProcessor

data CSRCFTopProcInfo problem idF = 
  CSRCFTopProcInfo { inProblem :: problem
                   , inConditionsReached :: Bool
                   , inReplacementMapOr :: Map idF [Int]
                   , inNewReplacementMaps :: [Map idF [Int]]
                   }
  | CSRCFTopDontKnow { inProblem :: problem
                     , inConditionsReached :: Bool
                     , inReplacementMapOr :: Map idF [Int]
                     , inNewReplacementMaps :: [Map idF [Int]]
                     }



-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InConfCSTRS trs
  , IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (CSRCFTopProcInfo (Problem Rewriting conftrs) idF)
  ) => Processor info CSRConverterFromTopProcessor (Problem Rewriting conftrs) (Problem Rewriting conftrs) where
  apply CSRConverterFromTopProcessor inP
    -- | True = success (CSRCFTopProcInfo inP True rmor nrms) inP
    | isCstrs = dontKnow (CSRCFTopDontKnow inP False rmor []) inP
    | isLeftLinear trs && (not $ null csrs) = singleP (CSRCFTopProcInfo newP True rmor nrms) inP newP
    | otherwise = dontKnow (CSRCFTopProcInfo inP False rmor []) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          isCstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          rmor = getTopRepMap trs
          csrs = toCSRs trs
          nrms = Prelude.map getRepMap csrs
          index = div (length csrs) 2
          csrChosen = csrs !! index
          newP = mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS (csrChosen))
          --newPs = [mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS csr) | csr <- csrs]

instance (
    conftrs ~ InTermCSTRS trs (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsTProblem conftrs (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (CSRCFTopProcInfo (Problem TRewriting conftrs) idF)
  ) => Processor info CSRConverterFromTopProcessor (Problem TRewriting conftrs) (Problem TRewriting conftrs) where
  apply CSRConverterFromTopProcessor inP
    -- | True = success (CSRCFTopProcInfo inP True rmor nrms) inP
    | isCstrs = dontKnow (CSRCFTopDontKnow inP False rmor []) inP
    | isLeftLinear trs && (not $ null csrs) = singleP (CSRCFTopProcInfo newP True rmor nrms) inP newP
    | otherwise = dontKnow (CSRCFTopProcInfo inP False rmor []) inP
    where trs = getConfProblem . getR $ inP
          term = getTermination . getR $ inP
          rules = elems . getTRSRules $ trs
          isCstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          rmor = getTopRepMap trs
          csrs = toCSRs trs
          nrms = Prelude.map getRepMap csrs
          index = div (length csrs) 2
          csrChosen = csrs !! index
          newP = mkProblem (TRewriting (getPSchemeFromProblem inP)) (InTermCSTRS csrChosen term)
          --newPs = [mkProblem (Rewriting (getPSchemeFromProblem inP)) (InTermCSTRS csr term) | csr <- csrs]

          
-- Pretty

instance (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (CSRCFTopProcInfo (Problem Rewriting conftrs) idF) where
  pPrint CSRCFTopProcInfo { inProblem = problem
                          , inConditionsReached = cr
                          , inReplacementMapOr = rmor
                          , inNewReplacementMaps = nrms
                          }
    = text "\nCSR Converter From Top u-Replacement Map Processor [CSUR20]:\n"
      Doc.<> (
        if not cr
        then text "\n\nConditions not reached\n"
        else text "\n\nOriginal Replacement Map:\n"
        Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList rmor))
        Doc.<> (
          if null nrms
          then text "\n\nA more restrictive canonical u-Replacement Map cannot be found\n"
          else text "\n\nNew Replacement Maps:\n"
               Doc.<> (
                  (hcat . punctuate (text "\n") . map (\n -> text "\nReplacement Map number" Doc.<+> int n Doc.<> text ":\n"
                    Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList (nrms!!(n-1))))
                  ) $ ns)
               )
          )
      )
    where ns = [1..(length nrms)]

instance (
    IsConfProblem conftrs trs
  , IsTProblem conftrs (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem TRewriting conftrs)
  ) => Pretty (CSRCFTopProcInfo (Problem TRewriting conftrs) idF) where
  pPrint CSRCFTopProcInfo { inProblem = problem
                          , inConditionsReached = cr
                          , inReplacementMapOr = rmor
                          , inNewReplacementMaps = nrms
                          }
    = text "\nCSR Converter From Top u-Replacement Map Processor [CSUR20]:\n"
      Doc.<> (
        if not cr
        then text "\n\nConditions not reached\n"
        else text "\n\nOriginal Replacement Map:\n"
        Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList rmor))
        Doc.<> (
          if null nrms
          then text "\n\nA more restrictive canonical u-Replacement Map cannot be found\n"
          else text "\n\nNew Replacement Maps:\n"
               Doc.<> (
                  (hcat . punctuate (text "\n") . map (\n -> text "\nReplacement Map number" Doc.<+> int n Doc.<> text ":\n"
                    Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map (\m -> pPrint m) $ mu) Doc.<> rparen) $ (M.toList (nrms!!(n-1))))
                  ) $ ns)
               )
          )
      )
    where ns = [1..(length nrms)]

      


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the CSR Converter From Top u-Replacement Map Processor
csrcfTopProc :: ( Processor info CSRConverterFromTopProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
csrcfTopProc = apply CSRConverterFromTopProcessor


toCSRs :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasArity idF, HasMu idF, Ord idF, Ord idV) => trs -> [trs]
toCSRs trs = newTRSs
  where rms = elems $ getLRCanoneRepMaps trs
        setrm a = setRepMap a trs 
        newTRSs = map setrm rms