{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.DifferentNormalCPTermsProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- if cp terms are distinct and not ground. Analysed CPs are not trivial
-- before entering this processor
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.DifferentNormalCPTermsProcessor (

-- * Exported functions

dncptProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..), vars)
import TRS.Identifiers (HasName (..), HasArity(..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPProblem (..), IsCPTermProblem (..))
import Framework.Problem.Types (InCPCSTRS (..), InCPTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..))
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, andP, Proof, Info (..), refuted)
import TRS.Rules(Rule(..), freshRule)
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.TRSAux (RuleId (..), getRulesIds, isNormalTerm)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Set as S (Set, elems, null, findMax)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Different Normal CP Terms Processor
data DiffNCPTsProcessor = DiffNCPTsProcessor

-- | The information obtained are the rules
data DiffNCPTsProcInfo problem idF idV = 
  DiffNCPTsProcInfo { inProblem   :: problem
                    , inCpi       :: CriticalPairInfo idF idV
                    }
  | DiffNCPTsDontKnow { inProblem   :: problem
                      , inCpi       :: CriticalPairInfo idF idV
                      }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (DiffNCPTsProcInfo (Problem CPRewriting conftrs) idF idV)
  ) => Processor info DiffNCPTsProcessor (Problem CPRewriting conftrs) (Problem CPRewriting conftrs) where
  apply DiffNCPTsProcessor inP
    -- | True = refuted (DiffNCPTsProcInfo inP cpi) inP
    | dn = refuted (DiffNCPTsProcInfo inP cpi) inP
    | otherwise = dontKnow (DiffNCPTsDontKnow inP cpi) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          rids = getRulesIds freshRules
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          dn = (isNormalTerm (s cp) rids) && (isNormalTerm (t cp) rids)

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (DiffNCPTsProcInfo (Problem CPTRewriting conftrs) idF idV)
  ) => Processor info DiffNCPTsProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply DiffNCPTsProcessor inP
    -- | True = refuted (DiffNCPTsProcInfo inP cpi) inP
    | dn = refuted (DiffNCPTsProcInfo inP cpi) inP
    | otherwise = dontKnow (DiffNCPTsDontKnow inP cpi) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          rids = getRulesIds freshRules
          cpis = getCPairsCPT . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          dn = (isNormalTerm (s cp) rids) && (isNormalTerm (t cp) rids)

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPRewriting conftrs)
  ) => Pretty (DiffNCPTsProcInfo (Problem CPRewriting conftrs) idF idV) where
  pPrint DiffNCPTsProcInfo { inProblem = problem
                           , inCpi = cpi
                           }
    = text "\nDifferent Normal CP Terms Processor:\n"
    Doc.<> pPrint cpi Doc.<> comma Doc.<+> text "Normal and not trivial cp"
    where trs = getConfProblem . getR $ problem

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPTRewriting conftrs)
  ) => Pretty (DiffNCPTsProcInfo (Problem CPTRewriting conftrs) idF idV) where
  pPrint DiffNCPTsProcInfo { inProblem = problem
                           , inCpi = cpi
                           }
    = text "\nDifferent Normal CP Terms Processor:\n"
    Doc.<> pPrint cpi Doc.<> comma Doc.<+> text "Normal and not trivial cp"
    where trs = getConfProblem . getR $ problem



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Different Normal CP Terms Processor
dncptProc :: ( Processor info DiffNCPTsProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
dncptProc = apply DiffNCPTsProcessor
