{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.ECPsNoLHRVProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that gets Extended u-Critical Pairs
-- and checks no convergence [JLAMP21]
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.ECPsNoLHRVProcessor (

-- * Exported functions

ecpnlhrvProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..), Substitution (..), isLinear, applySigma, matches, replace, vars)
import TRS.Identifiers (HasName (..), HasArity (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem, IsCPProblem (..), CriticalPair (..), CriticalPairInfo (..))
import Framework.Problem.Types (InConfCSTRS (..), InCPCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, Proof, Info (..))
import TRS.Rules(Rule(..), lhs, rhs, freshRule)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.TRSAux (RuleId (..), PosSubt (..), getRulesIds, isLHRV, activeVars, frozenVars, quicksortRuleIDByNumOfVars, isCSTRSSymb, posMuSubterm,
                         pairVarsCreation, createCPs, cpsInfo, getVarsRules, normalWeight)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes, int)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList, fromList)
import Data.Set as S (Set, elems, null, findMax, elemAt, intersection, insert, map, empty, union, delete, singleton, unions, fromList)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data ECPsNoLHRVProcessor = ECPsNoLHRVProcessor

-- | The information obtained are the rules
data ECPsNoLHRVProcInfo problem idF idV = 
  ECPsNoLHRVProcInfo { inProblem :: problem
                     , inECPs    :: [ExtendedCP idF idV]
                     }
  | ECPsNoLHRVDontKnow { inProblem :: problem
                       , inECPs    :: [ExtendedCP idF idV]
                       }


data ExtendedCP idF idV =
  ExtendedCP { ruleid1  :: RuleId idF idV
             , ruleid2  :: RuleId idF idV
             , var      :: idV
             , pECP     :: [Int]
             , sigmaECP :: Map idV (Term idF idV)
             , sECP     :: Term idF idV
             , tECP     :: Term idF idV
             , normWECP :: Int
             }
    deriving (Eq,Ord) 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs1 ~ InConfCSTRS trs
  , conftrs2 ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsCPProblem conftrs2 (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF 
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV, Ord (RuleId idF idV)
  , Info info (ECPsNoLHRVProcInfo (Problem Rewriting conftrs1) idF idV)
  ) => Processor info ECPsNoLHRVProcessor (Problem Rewriting conftrs1) (Problem CPRewriting conftrs2) where
  apply ECPsNoLHRVProcessor inP
    | not cstrs = dontKnow (ECPsNoLHRVDontKnow inP ecps) inP
    | otherwise = singleP (ECPsNoLHRVProcInfo inP ecps) inP (mkProblem (CPRewriting (getPSchemeFromProblem inP)) (InCPCSTRS trs cpis))
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          cstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          rIds = getRulesIds rules
          ecps = elems $ getECPs (getRulesIds rules) maxid 
          cps = Prelude.map toCP ecps
          cpis = cpsInfo cps 1

          
-- Pretty

instance (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (ECPsNoLHRVProcInfo (Problem Rewriting conftrs) idF idV) where
  pPrint ECPsNoLHRVProcInfo { inProblem = problem
                            , inECPs = ecps
                            }
    = pPrint problem
      Doc.<> text "\nExtended u-Critical Pairs NonLHRV Processor [JLAMP21]:\n"
      Doc.<> (if Prelude.null ecps then text "No ECPs found" else (hcat . punctuate (text "\n\n") . Prelude.map (\e -> pPrint e) $ ecps))


instance (
  Pretty idF, Pretty idV
  ) => Pretty (ExtendedCP idF idV) where
  pPrint ExtendedCP { ruleid1 = rid1
                    , ruleid2 = rid2
                    , var = v
                    , pECP = p
                    , sigmaECP = sig
                    , sECP = s
                    , tECP = t
                    , normWECP = nw
                    }
    = text "->Extended u-Critical Pair:\n"
      Doc.<+> text "Rule" Doc.<+> int (tidr rid1) Doc.<+> text "(l :-> r) =>" Doc.<+> pPrint (trule rid1) Doc.<> text "\n"
      Doc.<+> text "Rule" Doc.<+> int (tidr rid2) Doc.<+> text "(l' :-> r') =>" Doc.<+> pPrint (trule rid2) Doc.<> text "\n"
      Doc.<+> text "Var =>" Doc.<+> pPrint v Doc.<> text "\n"
      Doc.<+> text "Pos" Doc.<+> pPrint v Doc.<+> text "in l =>" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map pPrint $ p) Doc.<> text "\n"
      Doc.<+> text "Sigma =>" 
        Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v) 
          $ (M.toList sig)) Doc.<> text "\n"
      Doc.<+> text "s =>" Doc.<+> pPrint s Doc.<> text "\n"
      Doc.<+> text "t =>" Doc.<+> pPrint t Doc.<> text "\n"
      Doc.<+> text "NW =>" Doc.<+> pPrint nw Doc.<> text "\n"
      


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Extended u-Critical Pairs NonLHRV Processor
ecpnlhrvProc :: ( Processor info ECPsNoLHRVProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
ecpnlhrvProc = apply ECPsNoLHRVProcessor


getECPs :: (HasName idV, Pretty idV, HasMu idF, HasArity idF, Ord idF, Ord idV, Ord (RuleId idF idV), Eq idV) => [RuleId idF idV] -> Int -> Set (ExtendedCP idF idV)
getECPs [] _ = S.empty
getECPs (rid1:rids) maxid = S.union ecps (getECPs rids nmaxid)
  where (frl1, nmaxid) = runState (sequence . Prelude.map freshRule $ [trule rid1]) maxid
        frid1 = head $ getRulesIds frl1
        rids2 = (rid1:rids)
        ecps = S.unions . S.elems $ S.fromList [getECP frid1 rid2 rids2 | rid2 <- rids2]


getECP :: (HasMu idF, HasArity idF, Ord idF, Ord idV, Eq idV) => RuleId idF idV -> RuleId idF idV -> [RuleId idF idV] -> Set (ExtendedCP idF idV)
getECP rid1 rid2 rids
  | S.null inters = S.empty
  | otherwise = S.singleton (ExtendedCP rid1 rid2 v p sigma s t nw)
  where (l1 :-> r1) = trule rid1
        (l2 :-> r2) = trule rid2
        lAct = activeVars l1
        froz = (frozenVars l1) `union` (frozenVars r1)
        inters = lAct `intersection` froz
        v = elemAt 0 inters
        psubt = head $ posSubtActiveOcurrence v [] l1
        p = pos psubt
        subt = subterm psubt
        sigma = M.fromList [(v, l2)]
        s = applySigma sigma r1
        t = replace (applySigma sigma r2) p (applySigma sigma l1)
        nw = (normalWeight s rids) + (normalWeight t rids)


posSubtActiveOcurrence :: (HasMu idF, Eq idV) => idV -> [Int] -> Term idF idV -> [PosSubt idF idV]
posSubtActiveOcurrence v p t2@(F f ts) = concat [posSubtActiveOcurrence v (p++[i]) (ts!!(i-1)) | i <- mu]
  where mu = getFMu f
posSubtActiveOcurrence v p t2@(V v2)
  | v == v2 = [PosSubt p (V v)]
  | otherwise = []



toCP :: ExtendedCP idF idV -> CriticalPair idF idV
toCP (ExtendedCP _ _ _ p _ sEcp tEcp nw) = CriticalPair sEcp tEcp p [] nw
