{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.GroundNormalCPInfProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- of a criticial pair with a ground term and a normal form using infChecker
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.GroundNormalCPInfProcessor (

-- * Exported functions

gncpiProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(freshRule)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..), HasArity(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPProblem (..), IsCPTermProblem (..))
import Framework.Problem.Types (InConfCSTRS (..), InCPCSTRS (..), InCPTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteria)
import Framework.TRSAux (RuleId (..), getRulesIds, groundNormalTerms)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, hsep, lparen, rparen)
import Data.Set as S (Set, elems, null, findMax)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data GroundNormalCPInfProcessor = GroundNormalCPInfProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data GNCPIProcInfo problem = 
  GNCPIProcInfo { inProblem :: problem 
                , inICCrit  :: InfCheckerCriteria
                , inSol     :: String
                }
  | GNCPIDontKnow { inProblem :: problem 
                  , inICCrit  :: InfCheckerCriteria
                  , inSol     :: String  
                  }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (GNCPIProcInfo (Problem CPRewriting conftrs))
  ) => Processor info GroundNormalCPInfProcessor (Problem CPRewriting conftrs) (Problem CPRewriting conftrs) where
  apply GroundNormalCPInfProcessor {timeout = to} inP = 
{-    
           if gncp then refuted (GNCPIProcInfo inP icc "sol") inP
                   else dontKnow (GNCPIProcInfo inP icc "sol") inP
    -}   
        
    if gncp
    then unsafePerformIO $!
        do { 
            solution <- solveInfChecker to icc
            ; case solution of
                ("YES", Just sol) -> return (refuted (GNCPIProcInfo inP icc sol) inP)
                (_, Just sol)     -> return (dontKnow (GNCPIDontKnow inP icc sol) inP)
           }
    else dontKnow (GNCPIDontKnow inP icc "MAYBE") inP

    where trs = getConfProblem . getR $ inP
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          rids = getRulesIds freshRules
          cp = critPair cpi
          ts = ((s cp), (t cp))
          (gncp, t1, t2) = groundNormalTerms ts rids
          newCP = CriticalPair t1 t2 (cpPos cp) [] (normWeight cp)
          icc = buildInfCheckerCriteria trs newCP

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (GNCPIProcInfo (Problem CPTRewriting conftrs))
  ) => Processor info GroundNormalCPInfProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply GroundNormalCPInfProcessor {timeout = to} inP = 
{-    
           if gncp then refuted (GNCPIProcInfo inP icc "sol") inP
                   else dontKnow (GNCPIProcInfo inP icc "sol") inP
    -}   
        
    if gncp
    then unsafePerformIO $!
        do { 
            solution <- solveInfChecker to icc
            ; case solution of
                ("YES", Just sol) -> return (refuted (GNCPIProcInfo inP icc sol) inP)
                (_, Just sol)     -> return (dontKnow (GNCPIDontKnow inP icc sol) inP)
           }
    else dontKnow (GNCPIDontKnow inP icc "MAYBE") inP

    where trs = getConfProblem . getR $ inP
          cpis = getCPairsCPT . getR $ inP
          cpi = head cpis
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          rids = getRulesIds freshRules
          cp = critPair cpi
          ts = ((s cp), (t cp))
          (gncp, t1, t2) = groundNormalTerms ts rids
          newCP = CriticalPair t1 t2 (cpPos cp) [] (normWeight cp)
          icc = buildInfCheckerCriteria trs newCP

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idV
  , Pretty (Problem CPRewriting conftrs)
  ) => Pretty (GNCPIProcInfo (Problem CPRewriting conftrs)) where
  pPrint GNCPIProcInfo { inProblem = problem
                       , inICCrit = crit
                       , inSol = sol
                       }
    = text "\nGround Normal CP InfChecker Processor:\n"
    Doc.<> text "(PROBLEM INFEASIBILITY)\n"
    Doc.<> text "(VAR "
    Doc.<> (hsep . map (\v -> pPrint v) $ (trsVars crit))
    Doc.<> text ")\n"
    Doc.<> text "(RULES\n"
    Doc.<> (hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (rulesICC crit))
    Doc.<> text "\n)\n"
    Doc.<> text "(CONDITION "
    Doc.<> pPrint (fst $ cpsConst crit) Doc.<> text " == " Doc.<> pPrint (snd $ cpsConst crit)
    Doc.<> text ")\n"
    Doc.<> text "\nInfeasible convergence?\n"
      Doc.<> (hcat . punctuate (text "") . map (\c -> pPrint c) $ sol)
    where trs = getConfProblem . getR $ problem

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idV
  , Pretty (Problem CPTRewriting conftrs)
  ) => Pretty (GNCPIProcInfo (Problem CPTRewriting conftrs)) where
  pPrint GNCPIProcInfo { inProblem = problem
                       , inICCrit = crit
                       , inSol = sol
                       }
    = text "\nGround Normal CP InfChecker Processor:\n"
    Doc.<> text "(PROBLEM INFEASIBILITY)\n"
    Doc.<> text "(VAR "
    Doc.<> (hsep . map (\v -> pPrint v) $ (trsVars crit))
    Doc.<> text ")\n"
    Doc.<> text "(RULES\n"
    Doc.<> (hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (rulesICC crit))
    Doc.<> text "\n)\n"
    Doc.<> text "(CONDITION "
    Doc.<> pPrint (fst $ cpsConst crit) Doc.<> text " == " Doc.<> pPrint (snd $ cpsConst crit)
    Doc.<> text ")\n"
    Doc.<> text "\nInfeasible convergence?\n"
      Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Ground Normal CP InfChecker Processor
gncpiProc :: ( Processor info GroundNormalCPInfProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
gncpiProc to = apply GroundNormalCPInfProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveInfChecker :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
solveInfChecker timeout crit
  = callToInfChecker timeout infProblem
  where
    infProblem = bconcat $ [frmStr "(PROBLEM INFEASIBILITY)\n"
                           , frmStr "(VAR "
                           , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (trsVars crit)
                           , frmStr ")\n"
                           , frmStr "(RULES\n"
                           , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (rulesICC crit)
                           , frmStr "\n)\n"
                           , frmStr "(CONDITION "
                           , frmStr . noRiskyChars $ show $ pPrint (fst $ cpsConst crit) Doc.<> text " == " Doc.<> pPrint (snd $ cpsConst crit)
                           , frmStr ")"]
