{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.HuetBFJProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check convergence with
-- a width first algorithm with graphs. Depth is enclosed
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.HuetBFJProcessor (

-- * Exported functions

huetBFJProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), HasArity(..), HasMu (..))
import TRS.Rules(Rule(..), freshRule)
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermProblem (..), IsCPNormProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermCSTRS (..), InCPNormCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, Proof, Info (..), refuted)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.Problem.CPNRewriting (CPNRewriting (..))
import Framework.TRSAux (RuleId (..), JoinabilityInfo (..), RewTerms (..), getRulesIds, isCSTRSSymb, isLHRV, maxDepthReached, createRewritingGraph, 
  haveJoinTerm)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Set as S (Set, elems, null, findMax)
import Data.Map as M (toList, empty)
import Data.Maybe (isJust, fromJust, isNothing)

import Data.Graph.Inductive.Graph as G
import Data.Graph.Inductive.PatriciaTree (Gr (..))

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data HuetBFJProcessor = HuetBFJProcessor  { maxDepth :: Maybe (Int), maxCPOut :: Maybe (Int) }

-- | The information obtained are the rules
data HuetBFJProcInfo problem idF idV = 
  HuetBFJProcInfo { inProblem   :: problem
                  , inCpi       :: CriticalPairInfo idF idV
                  , inJoinInfo  :: JoinabilityInfo idF idV
                  , inSGraph    :: Gr (RewTerms idF idV) () 
                  , inTGraph    :: Gr (RewTerms idF idV) ()
                  , inIsLHCP    :: Bool
                  }
  | HuetBFJDontKnow { inProblem   :: problem
                    , inCpi       :: CriticalPairInfo idF idV
                    , inJoinInfo  :: JoinabilityInfo idF idV
                    , inSGraph    :: Gr (RewTerms idF idV) () 
                    , inTGraph    :: Gr (RewTerms idF idV) ()
                    , inIsLHCP    :: Bool
                    }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (HuetBFJProcInfo (Problem CPTRewriting conftrs) idF idV)
  ) => Processor info HuetBFJProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply HuetBFJProcessor {maxDepth = md, maxCPOut = mcp} inP
    | not . Prelude.null $ ccpConds cp = 
        singleP (HuetBFJProcInfo inP cpi joinInfo sg tg True) inP (mkProblem (CPTRewriting (getPSchemeFromProblem inP)) (InCPTermCSTRS trs cpis term))
    | join = success (HuetBFJProcInfo inP cpi joinInfo sg tg False) inP
    | otherwise = dontKnow (HuetBFJProcInfo inP cpi joinInfo sg tg False) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          symbs =  elems . getTRSSymbols $ trs
          cstrs = or $ map isCSTRSSymb symbs
          lhrv = and $ map isLHRV rules
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          cpis = getCPairsCPT . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          term = getTermCPT . getR $ inP
          ncp = cpNum cpi
          (joinInfo, sg, tg) = if isTrue term then isLocallyConfluent (getRulesIds freshRules) cp (md,mcp,ncp)
                               else (JoinabilityInfo False Nothing (Just True) [] [], G.empty, G.empty)
          join = (cstrs && lhrv && (isTrue term) && (joinable joinInfo)) || ((not cstrs) && (isTrue term) && (joinable joinInfo))


instance (
    conftrs ~ InCPNormCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPNormProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (HuetBFJProcInfo (Problem CPNRewriting conftrs) idF idV)
  ) => Processor info HuetBFJProcessor (Problem CPNRewriting conftrs) (Problem CPNRewriting conftrs) where
  apply HuetBFJProcessor {maxDepth = md, maxCPOut = mcp} inP
    -- | True = success (HuetBFJProcInfo inP cpi joinInfo sg tg False) inP
    | join = success (HuetBFJProcInfo inP cpi joinInfo sg tg False) inP
    | otherwise = dontKnow (HuetBFJProcInfo inP cpi joinInfo sg tg False) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          symbs =  elems . getTRSSymbols $ trs
          cstrs = or $ map isCSTRSSymb symbs
          lhrv = and $ map isLHRV rules
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          cpis = getCPairsCPN . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          norm = getNormCPN . getR $ inP
          ncp = cpNum cpi
          (joinInfo, sg, tg) = if isTrue norm then isLocallyConfluent (getRulesIds freshRules) cp (md,mcp,ncp)
                               else (JoinabilityInfo False Nothing (Just True) [] [], G.empty, G.empty)
          join = (cstrs && lhrv && (isTrue norm) && (joinable joinInfo)) || ((not cstrs) && (isTrue norm) && (joinable joinInfo))


instance (
    IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPTRewriting conftrs)
  ) => Pretty (HuetBFJProcInfo (Problem CPTRewriting conftrs) idF idV) where
  pPrint HuetBFJProcInfo { inProblem = problem
                         , inCpi = cpi
                         , inJoinInfo = j
                         , inSGraph = sg
                         , inTGraph = tg
                         , inIsLHCP = False
                         }
    = pPrint problem
    Doc.<> text "\nHuet Brute Force Joinability Processor:\n"
    Doc.<> text "-> Rewritings:\n"
    Doc.<> text "s: " Doc.<> pPrint (s cp) Doc.<> text "\n" 
    Doc.<> text "Nodes: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes sg)) Doc.<> text "\n" 
    Doc.<> text "Edges: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges sg)) Doc.<> text "\n"
      Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes sg)) Doc.<> text "\n"
    Doc.<> text "t: " Doc.<> pPrint (t cp) Doc.<> text "\n" 
    Doc.<> text "Nodes: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes tg)) Doc.<> text "\n" 
    Doc.<> text "Edges: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges tg)) Doc.<> text "\n"
      Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes tg)) Doc.<> text "\n"
    Doc.<> text "SNodesPath: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (sNodesPath j)) Doc.<> text "\n"
      Doc.<> text "TNodesPath: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (tNodesPath j)) Doc.<> text "\n"
      
    Doc.<> pPrint (s cp) Doc.<> text " ->* "
    Doc.<> (if isNothing (union j) then text "no union" else pPrint (fromJust $ union j))
    Doc.<> text " *<- " Doc.<> pPrint (t cp) Doc.<> text "\n"
    Doc.<> pPrint (if joinable j then  "Joinable" else "Not joinable")
    where trs = getConfProblem . getR $ problem
          cp = critPair cpi
  pPrint HuetBFJProcInfo { inProblem = problem
                         , inCpi = cpi
                         , inJoinInfo = j
                         , inSGraph = sg
                         , inTGraph = tg
                         , inIsLHCP = True
                         }
    = pPrint problem
    Doc.<> text "\nHuet Brute Force Joinability Processor:\n"
    Doc.<> text "That is a LHCP. Trying another procedure."

instance (
    IsConfProblem conftrs trs
  , IsCPNormProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPNRewriting conftrs)
  ) => Pretty (HuetBFJProcInfo (Problem CPNRewriting conftrs) idF idV) where
  pPrint HuetBFJProcInfo { inProblem = problem
                         , inCpi = cpi
                         , inJoinInfo = j
                         , inSGraph = sg
                         , inTGraph = tg
                         , inIsLHCP = False
                         }
    = pPrint problem
    Doc.<> text "\nHuet Brute Force Joinability Processor:\n"
    Doc.<> text "-> Rewritings:\n"
    Doc.<> text "s: " Doc.<> pPrint (s cp) Doc.<> text "\n" 
    Doc.<> text "Nodes: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes sg)) Doc.<> text "\n" 
    Doc.<> text "Edges: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges sg)) Doc.<> text "\n"
      Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes sg)) Doc.<> text "\n"
    Doc.<> text "t: " Doc.<> pPrint (t cp) Doc.<> text "\n" 
    Doc.<> text "Nodes: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes tg)) Doc.<> text "\n" 
    Doc.<> text "Edges: "
      Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges tg)) Doc.<> text "\n"
      Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes tg)) Doc.<> text "\n"
    Doc.<> text "SNodesPath: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (sNodesPath j)) Doc.<> text "\n"
      Doc.<> text "TNodesPath: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (tNodesPath j)) Doc.<> text "\n"
      
    Doc.<> pPrint (s cp) Doc.<> text " ->* "
    Doc.<> (if isNothing (union j) then text "no union" else pPrint (fromJust $ union j))
    Doc.<> text " *<- " Doc.<> pPrint (t cp) Doc.<> text "\n"
    Doc.<> pPrint (if joinable j then  "Joinable" else "Not joinable")
    where trs = getConfProblem . getR $ problem
          cp = critPair cpi
  pPrint HuetBFJProcInfo { inProblem = problem
                         , inCpi = cpi
                         , inJoinInfo = j
                         , inSGraph = sg
                         , inTGraph = tg
                         , inIsLHCP = True
                         }
    = pPrint problem
    Doc.<> text "\nHuet Brute Force Joinability Processor:\n"
    Doc.<> text "That is a LHCP. Trying another procedure."

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Huet Brute Force Joinability Processor
huetBFJProc :: ( Processor info HuetBFJProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe(Int) -> Maybe(Int) -> o -> Proof info mp d
huetBFJProc md mcp = apply HuetBFJProcessor { maxDepth = md
                                            , maxCPOut = mcp}


isLocallyConfluent :: (HasMu idF, Eq idF, Ord idV) => [RuleId idF idV] -> CriticalPair idF idV -> (Maybe(Int),Maybe(Int),Int) 
  -> (JoinabilityInfo idF idV, Gr (RewTerms idF idV) (), Gr (RewTerms idF idV) ())
isLocallyConfluent rs cp (md,mcp,ncp) = areJoinable (s cp) (t cp) rs (md,mcp,ncp)


areJoinable :: (HasMu idF, Eq idF, Ord idV) => Term idF idV -> Term idF idV -> [RuleId idF idV] -> (Maybe(Int),Maybe(Int),Int) 
  -> (JoinabilityInfo idF idV, Gr (RewTerms idF idV) (), Gr (RewTerms idF idV) ())
areJoinable s t rs (md,(Just mcp),ncp) 
  | mdr = (JoinabilityInfo False Nothing (Just mdr) [] [], sGraph, tGraph)
  | ncp >= mcp = (JoinabilityInfo j u (Just mdr) [] [], G.empty, G.empty)
  | otherwise = (JoinabilityInfo j u (Just mdr) sNodesPath tNodesPath, sGraph, tGraph)
  where sGraph = createRewritingGraph s rs md
        tGraph = createRewritingGraph t rs md
        sRewTerms = map snd (labNodes sGraph)
        tRewTerms = map snd (labNodes tGraph)
        mdr = (maxDepthReached md sGraph) || (maxDepthReached md tGraph)
        (j, u, sNodesPath, tNodesPath) = haveJoinTerm sGraph tGraph
