{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.HuetLevyProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that calculate TRS CP Pairs and applies
-- Huet Levy theorem
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.HuetLevyProcessor (

-- * Exported data


-- * Exported functions

huetLevyProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), HasMu (..), HasArity (..), IdFTRS(..), IdVTRS(..))
import Framework.Problem.Types(InConfCSTRS, InCPCSTRS (..), InTermCSTRS (..), InCPTermCSTRS (..), InCPLinCSTRS (..), InLinCSTRS (..),
                               InNormCSTRS (..), InCPNormCSTRS (..))
import Framework.Problems(IsConfProblem (..), IsCPProblem (..), CriticalPair (..), CriticalPairInfo (..), getPSchemeFromProblem, IsTProblem (..), 
      IsCPTermProblem (..), IsCPLinProblem (..), IsLProblem (..), IsNProblem (..), IsCPNormProblem (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, andP, Proof, Info (..))
import TRS.Rules(Rule(..), freshRule, rcond)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.Problem.NRewriting (NRewriting (..))
import Framework.Problem.LRewriting (LRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.Problem.CPNRewriting (CPNRewriting (..))
import Framework.Problem.CPLRewriting (CPLRewriting (..))
import Framework.TRSAux (PosSubt (..), RuleId (..), UnificationsCPBuilder (..), RuleLhsPosSubterms (..), UVar (..), 
                         maybeConvergentCPs, cpsInfo, getRulesIds, uVarsCreation, criticalPairUnification, 
                         quicksortCPByNumOfNormalTerms, cleanUnifications, createCPs, checkUnify, posMuSubtermsRule,
                         pairVarsCreation, getVarsRules, getLHCPPairVarYFromCPIs, criticalPairsInfoVars)
import Framework.ProblemConclusions (ProblemConclusionsTRS (..), problemConclusionTRS)
import Utils.Auxiliar (isTrue, shuffle)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, int, hsep, comma, brackets, lparen, rparen, braces)
import Control.Monad (MonadPlus)
import Data.Set as S (Set, elems, null, findMax, map, toList, fromList, union)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Levy Processor
data HuetLevyProcessor = HuetLevyProcessor

-- | The information obtained are the rules
data HuetLevyProcInfo problem idF idV = 
  HuetLevyProcInfo { inProblem            :: problem
                   , inRLhsPosSbts        :: [RuleLhsPosSubterms idF idV]
                   , inUnifications       :: [UnificationsCPBuilder idF idV]
                   , inCriticalPairsInfo  :: [CriticalPairInfo idF idV]
                   , inProblemConclusions :: ProblemConclusionsTRS
                   , inVars               :: [idV]
                   , inUVars              :: [UVar idV]
                   }
  | HuetLevyDontKnow { inProblem            :: problem
                     , inRLhsPosSbts        :: [RuleLhsPosSubterms idF idV]
                     , inUnifications       :: [UnificationsCPBuilder idF idV]
                     , inCriticalPairsInfo  :: [CriticalPairInfo idF idV]
                     , inProblemConclusions :: ProblemConclusionsTRS
                     , inVars               :: [idV]
                     , inUVars              :: [UVar idV]
                     }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InTermCSTRS trs (Maybe(Bool))
  , conftrs2 ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsTProblem conftrs1 (Maybe(Bool))
  , IsCPTermProblem conftrs2 (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (HuetLevyProcInfo (Problem TRewriting conftrs1) idF idV)
  ) => Processor info HuetLevyProcessor (Problem TRewriting conftrs1) (Problem CPTRewriting conftrs2) where
  apply HuetLevyProcessor inP
    -- | True = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (isCSTRS pconc) && (isTrue $ leftHomogMuRepVars pconc) && ((huetLevyConfluent pconc) || (newman pconc)) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (not $ isCSTRS pconc) && ((huetLevyConfluent pconc) || (newman pconc)) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | Prelude.null potConvCps = dontKnow (HuetLevyDontKnow inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | otherwise = andP (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP 
                    [mkProblem (CPTRewriting (getPSchemeFromProblem inP)) (InCPTermCSTRS newTRS [cp] termination) | cp <- potConvCps]
    where trs = getConfProblem . getR $ inP
          termination = getTermination . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshRules,newMaxId) = runState (sequence . Prelude.map freshRule $ rules) maxid
          rIds = getRulesIds rules
          vs = getVarsRules rules
          uvars = Prelude.map uVarsCreation rIds
          rLhsPosSbts = Prelude.map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = shuffle $ createCPs [criticalPairUnification u | u <- unifications]
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          pconc = problemConclusionTRS newTRS cpis termination


instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InNormCSTRS trs (Maybe(Bool))
  , conftrs2 ~ InCPNormCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsNProblem conftrs1 (Maybe(Bool))
  , IsCPNormProblem conftrs2 (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (HuetLevyProcInfo (Problem NRewriting conftrs1) idF idV)
  ) => Processor info HuetLevyProcessor (Problem NRewriting conftrs1) (Problem CPNRewriting conftrs2) where
  apply HuetLevyProcessor inP
    -- | True = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs fVars pvars uvars) inP
    | (isCSTRS pconc) && (isTrue $ leftHomogMuRepVars pconc) && ((huetLevyConfluent pconc) || (newman pconc)) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (not $ isCSTRS pconc) && ((huetLevyConfluent pconc) || (newman pconc)) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | Prelude.null potConvCps = dontKnow (HuetLevyDontKnow inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | otherwise = andP (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP 
                    [mkProblem (CPNRewriting (getPSchemeFromProblem inP)) (InCPNormCSTRS newTRS [cp] normalization) | cp <- potConvCps]
    where trs = getConfProblem . getR $ inP
          normalization = getNormalization . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshRules,newMaxId) = runState (sequence . Prelude.map freshRule $ rules) maxid
          rIds = getRulesIds rules
          vs = getVarsRules rules
          uvars = Prelude.map uVarsCreation rIds
          rLhsPosSbts = Prelude.map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = shuffle $ createCPs [criticalPairUnification u | u <- unifications]
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          pconc = problemConclusionTRS newTRS cpis Nothing
          

instance {-# OVERLAPPABLE #-}(
    conftrs1 ~ InConfCSTRS trs
  , conftrs2 ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsCPProblem conftrs2 (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (HuetLevyProcInfo (Problem Rewriting conftrs1) idF idV)
  ) => Processor info HuetLevyProcessor (Problem Rewriting conftrs1) (Problem CPRewriting conftrs2) where
  apply HuetLevyProcessor inP
    -- | True = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (isCSTRS pconc) && (isTrue $ leftHomogMuRepVars pconc) && (huetLevyConfluent pconc) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (not $ isCSTRS pconc) && (huetLevyConfluent pconc) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | Prelude.null potConvCps = dontKnow (HuetLevyDontKnow inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | otherwise = andP (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP 
                    [mkProblem (CPRewriting (getPSchemeFromProblem inP)) (InCPCSTRS newTRS [cp]) | cp <- potConvCps]
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshRules,newMaxId) = runState (sequence . Prelude.map freshRule $ rules) maxid
          rIds = getRulesIds rules
          vs = getVarsRules rules
          uvars = Prelude.map uVarsCreation rIds
          rLhsPosSbts = Prelude.map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = shuffle $ createCPs [criticalPairUnification u | u <- unifications]
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          pconc = problemConclusionTRS newTRS cpis Nothing


instance {-# OVERLAPPABLE #-}(
    conftrs1 ~ InLinCSTRS trs Bool
  , conftrs2 ~ InCPLinCSTRS trs (CriticalPairInfo idF idV) Bool
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsLProblem conftrs1 Bool
  , IsCPLinProblem conftrs2 (CriticalPairInfo idF idV) Bool
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (HuetLevyProcInfo (Problem LRewriting conftrs1) idF idV)
  ) => Processor info HuetLevyProcessor (Problem LRewriting conftrs1) (Problem CPLRewriting conftrs2) where
  apply HuetLevyProcessor inP
    -- | True = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (isCSTRS pconc) && (isTrue $ leftHomogMuRepVars pconc) && (huetLevyConfluent pconc) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (not $ isCSTRS pconc) && (huetLevyConfluent pconc) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | Prelude.null potConvCps = dontKnow (HuetLevyDontKnow inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | otherwise = andP (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP 
                    [mkProblem (CPLRewriting (getPSchemeFromProblem inP)) (InCPLinCSTRS newTRS [cp] l) | cp <- potConvCps]
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshRules,newMaxId) = runState (sequence . Prelude.map freshRule $ rules) maxid
          rIds = getRulesIds rules
          vs = getVarsRules rules
          uvars = Prelude.map uVarsCreation rIds
          rLhsPosSbts = Prelude.map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = shuffle $ createCPs [criticalPairUnification u | u <- unifications]
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          pconc = problemConclusionTRS newTRS cpis Nothing
          l = linearTRS pconc


instance {-# OVERLAPPABLE #-}(
    conftrs ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (HuetLevyProcInfo (Problem CPRewriting conftrs) idF idV)
  ) => Processor info HuetLevyProcessor (Problem CPRewriting conftrs) (Problem CPRewriting conftrs) where
  apply HuetLevyProcessor inP
    -- | True = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | Prelude.null potConvCps = dontKnow (HuetLevyDontKnow inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | otherwise = andP (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP 
                    [mkProblem (CPRewriting (getPSchemeFromProblem inP)) (InCPCSTRS newTRS [cp]) | cp <- potConvCps]
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          oldCpis = getCPairs . getR $ inP
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshRules,newMaxId) = runState (sequence . Prelude.map freshRule $ rules) maxid
          rIds = getRulesIds rules
          --frIds = getRulesIds freshRules
          --fVars = getVarsRules freshRules
          vs = getVarsRules rules
          --pvy = getLHCPPairVarYFromCPIs oldCpis
          --pvars
          --  | isNothing pvy = pairVarsCreation vs fVars
          --  | otherwise = (fromJust pvy):(pairVarsCreation vs fVars)
          uvars = Prelude.map uVarsCreation rIds
          rLhsPosSbts = Prelude.map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = shuffle $ (Prelude.map critPair oldCpis) ++
                (createCPs [criticalPairUnification u | u <- unifications])
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          pconc = problemConclusionTRS newTRS cpis Nothing


instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (HuetLevyProcInfo (Problem CPTRewriting conftrs) idF idV)
  ) => Processor info HuetLevyProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply HuetLevyProcessor inP
    -- | True = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (isCSTRS pconc) && (isTrue $ leftHomogMuRepVars pconc) && ((huetLevyConfluent pconc) || (newman pconc)) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | (not $ isCSTRS pconc) && ((huetLevyConfluent pconc) || (newman pconc)) 
        = success (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | Prelude.null potConvCps = dontKnow (HuetLevyDontKnow inP rLhsPosSbts unifications cpis pconc vs uvars) inP
    | otherwise = andP (HuetLevyProcInfo inP rLhsPosSbts unifications cpis pconc vs uvars) inP 
                    [mkProblem (CPTRewriting (getPSchemeFromProblem inP)) (InCPTermCSTRS newTRS [cp] termination) | cp <- potConvCps]
    where trs = getConfProblem . getR $ inP
          termination = getTermCPT . getR $ inP
          oldCpis = getCPairsCPT . getR $ inP
          rules = elems . getTRSRules $ trs
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          --(freshRules,newMaxId) = runState (sequence . Prelude.map freshRule $ rules) maxid
          rIds = getRulesIds rules
          --frIds = getRulesIds freshRules
          --fVars = getVarsRules freshRules
          vs = getVarsRules rules
          --pvy = getLHCPPairVarYFromCPIs oldCpis
          --pvars
          --  | isNothing pvy = pairVarsCreation vs fVars
          --  | otherwise = (fromJust pvy):(pairVarsCreation vs fVars)
          uvars = Prelude.map uVarsCreation rIds
          rLhsPosSbts = Prelude.map posMuSubtermsRule rIds
          unifications = cleanUnifications $ checkUnify rLhsPosSbts rIds maxid
          cps = shuffle $ (Prelude.map critPair oldCpis) ++
                (createCPs [criticalPairUnification u | u <- unifications])
          cpis = cpsInfo cps 1
          potConvCps = maybeConvergentCPs cpis
          newVs = (S.fromList vs) `S.union` criticalPairsInfoVars potConvCps
          newTRS = setTRSVariables newVs trs
          pconc = problemConclusionTRS newTRS cpis termination


-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem TRewriting conftrs)
  ) => Pretty (HuetLevyProcInfo (Problem TRewriting conftrs) idF idV) where
  pPrint HuetLevyProcInfo { inProblem = problem
                          , inRLhsPosSbts = rlps
                          , inUnifications = unfs
                          , inCriticalPairsInfo = cpis
                          , inProblemConclusions = pconc
                          , inVars = vs
                          , inUVars = uvars}
    = pPrint problem
      Doc.<> text "\nHuet Levy Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map pPrint . elems . getTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<+> (hsep . punctuate comma . Prelude.map(\v-> text (getName v)) $ vs)
      Doc.<> (if isCSTRS pconc 
              then text "\n-> UVars:\n"
                      Doc.<+> (hsep . punctuate (text "\n") $ Prelude.map pPrint uvars)
              else text "")
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map (\r -> lparen
        Doc.<> text "rule:" Doc.<+> pPrint (trule $ rid r) Doc.<> comma
        Doc.<+> text "id:" Doc.<+> pPrint (tidr $ rid r) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . Prelude.map(\ps-> pPrint (subterm ps) Doc.<> text "->"
          Doc.<> (brackets . hsep . punctuate comma . Prelude.map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))  Doc.<> rparen) $ rlps)
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\u -> lparen
        Doc.<> text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<> text "p:" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l:" Doc.<+> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp:" Doc.<+> pPrint (lp u) Doc.<> comma
        Doc.<+> text "sig:" Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l':" Doc.<+> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r:" Doc.<+> pPrint (r u) Doc.<> comma
        Doc.<+> text "r':" Doc.<+> pPrint (r1 u) Doc.<> rparen) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\cpi -> pPrint cpi) $ cpis)
      Doc.<> pPrint pconc Doc.<> text "\n"
    where trs = getConfProblem . getR $ problem


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem NRewriting conftrs)
  ) => Pretty (HuetLevyProcInfo (Problem NRewriting conftrs) idF idV) where
  pPrint HuetLevyProcInfo { inProblem = problem
                          , inRLhsPosSbts = rlps
                          , inUnifications = unfs
                          , inCriticalPairsInfo = cpis
                          , inProblemConclusions = pconc
                          , inVars = vs
                          , inUVars = uvars}
    = pPrint problem
      Doc.<> text "\nHuet Levy Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map pPrint . elems . getTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<+> (hsep . punctuate comma . Prelude.map(\v-> text (getName v)) $ vs)
      Doc.<> (if isCSTRS pconc 
              then text "\n-> UVars:\n"
                      Doc.<+> (hsep . punctuate (text "\n") $ Prelude.map pPrint uvars)
              else text "")
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map (\r -> lparen
        Doc.<> text "rule:" Doc.<+> pPrint (trule $ rid r) Doc.<> comma
        Doc.<+> text "id:" Doc.<+> pPrint (tidr $ rid r) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . Prelude.map(\ps-> pPrint (subterm ps) Doc.<> text "->"
          Doc.<> (brackets . hsep . punctuate comma . Prelude.map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))  Doc.<> rparen) $ rlps)
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\u -> lparen
        Doc.<> text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<> text "p:" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l:" Doc.<+> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp:" Doc.<+> pPrint (lp u) Doc.<> comma
        Doc.<+> text "sig:" Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l':" Doc.<+> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r:" Doc.<+> pPrint (r u) Doc.<> comma
        Doc.<+> text "r':" Doc.<+> pPrint (r1 u) Doc.<> rparen) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\cpi -> pPrint cpi) $ cpis)
      Doc.<> pPrint pconc Doc.<> text "\n"
    where trs = getConfProblem . getR $ problem


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem LRewriting conftrs)
  ) => Pretty (HuetLevyProcInfo (Problem LRewriting conftrs) idF idV) where
  pPrint HuetLevyProcInfo { inProblem = problem
                          , inRLhsPosSbts = rlps
                          , inUnifications = unfs
                          , inCriticalPairsInfo = cpis
                          , inProblemConclusions = pconc
                          , inVars = vs
                          , inUVars = uvars}
    = pPrint problem
      Doc.<> text "\nHuet Levy Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map pPrint . elems . getTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<+> (hsep . punctuate comma . Prelude.map(\v-> text (getName v)) $ vs)
      Doc.<> (if isCSTRS pconc 
              then text "\n-> UVars:\n"
                      Doc.<+> (hsep . punctuate (text "\n") $ Prelude.map pPrint uvars)
              else text "")
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map (\r -> lparen
        Doc.<> text "rule:" Doc.<+> pPrint (trule $ rid r) Doc.<> comma
        Doc.<+> text "id:" Doc.<+> pPrint (tidr $ rid r) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . Prelude.map(\ps-> pPrint (subterm ps) Doc.<> text "->"
          Doc.<> (brackets . hsep . punctuate comma . Prelude.map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))  Doc.<> rparen) $ rlps)
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\u -> lparen
        Doc.<> text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<> text "p:" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l:" Doc.<+> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp:" Doc.<+> pPrint (lp u) Doc.<> comma
        Doc.<+> text "sig:" Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l':" Doc.<+> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r:" Doc.<+> pPrint (r u) Doc.<> comma
        Doc.<+> text "r':" Doc.<+> pPrint (r1 u) Doc.<> rparen) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\cpi -> pPrint cpi) $ cpis)
      Doc.<> pPrint pconc Doc.<> text "\n"
    where trs = getConfProblem . getR $ problem


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (HuetLevyProcInfo (Problem Rewriting conftrs) idF idV) where
  pPrint HuetLevyProcInfo { inProblem = problem
                          , inRLhsPosSbts = rlps
                          , inUnifications = unfs
                          , inCriticalPairsInfo = cpis
                          , inProblemConclusions = pconc
                          , inVars = vs
                          , inUVars = uvars}
    = pPrint problem
      Doc.<> text "\nHuet Levy Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map pPrint . elems . getTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<+> (hsep . punctuate comma . Prelude.map(\v-> text (getName v)) $ vs)
      Doc.<> (if isCSTRS pconc 
              then text "\n-> UVars:\n"
                      Doc.<+> (hsep . punctuate (text "\n") $ Prelude.map pPrint uvars)
              else text "")
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map (\r -> lparen
        Doc.<> text "rule:" Doc.<+> pPrint (trule $ rid r) Doc.<> comma
        Doc.<+> text "id:" Doc.<+> pPrint (tidr $ rid r) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . Prelude.map(\ps-> pPrint (subterm ps) Doc.<> text "->"
          Doc.<> (brackets . hsep . punctuate comma . Prelude.map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))  Doc.<> rparen) $ rlps)
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\u -> lparen
        Doc.<> text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<> text "p:" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l:" Doc.<+> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp:" Doc.<+> pPrint (lp u) Doc.<> comma
        Doc.<+> text "sig:" Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l':" Doc.<+> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r:" Doc.<+> pPrint (r u) Doc.<> comma
        Doc.<+> text "r':" Doc.<+> pPrint (r1 u) Doc.<> rparen) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\cpi -> pPrint cpi) $ cpis)
      Doc.<> pPrint pconc Doc.<> text "\n"
    where trs = getConfProblem . getR $ problem


instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPRewriting conftrs)
  ) => Pretty (HuetLevyProcInfo (Problem CPRewriting conftrs) idF idV) where
  pPrint HuetLevyProcInfo { inProblem = problem
                          , inRLhsPosSbts = rlps
                          , inUnifications = unfs
                          , inCriticalPairsInfo = cpis
                          , inProblemConclusions = pconc
                          , inVars = vs
                          , inUVars = uvars}
    = pPrint problem
      Doc.<> text "\nHuet Levy Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map pPrint . elems . getTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<+> (hsep . punctuate comma . Prelude.map(\v-> text (getName v)) $ vs)
      Doc.<> (if isCSTRS pconc 
              then text "\n-> UVars:\n"
                      Doc.<+> (hsep . punctuate (text "\n") $ Prelude.map pPrint uvars)
              else text "")
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map (\r -> lparen
        Doc.<> text "rule:" Doc.<+> pPrint (trule $ rid r) Doc.<> comma
        Doc.<+> text "id:" Doc.<+> pPrint (tidr $ rid r) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . Prelude.map(\ps-> pPrint (subterm ps) Doc.<> text "->"
          Doc.<> (brackets . hsep . punctuate comma . Prelude.map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))  Doc.<> rparen) $ rlps)
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\u -> lparen
        Doc.<> text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<> text "p:" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l:" Doc.<+> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp:" Doc.<+> pPrint (lp u) Doc.<> comma
        Doc.<+> text "sig:" Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l':" Doc.<+> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r:" Doc.<+> pPrint (r u) Doc.<> comma
        Doc.<+> text "r':" Doc.<+> pPrint (r1 u) Doc.<> rparen) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\cpi -> pPrint cpi) $ cpis)
      Doc.<> pPrint pconc Doc.<> text "\n"
    where trs = getConfProblem . getR $ problem

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPTRewriting conftrs)
  ) => Pretty (HuetLevyProcInfo (Problem CPTRewriting conftrs) idF idV) where
  pPrint HuetLevyProcInfo { inProblem = problem
                          , inRLhsPosSbts = rlps
                          , inUnifications = unfs
                          , inCriticalPairsInfo = cpis
                          , inProblemConclusions = pconc
                          , inVars = vs
                          , inUVars = uvars}
    = pPrint problem
      Doc.<> text "\nHuet Levy Processor:\n"
      Doc.<> text "-> Rules:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map pPrint . elems . getTRSRules $ trs)
      Doc.<> text "\n-> Vars:\n"
        Doc.<+> (hsep . punctuate comma . Prelude.map(\v-> text (getName v)) $ vs)
      Doc.<> (if isCSTRS pconc 
              then text "\n-> UVars:\n"
                      Doc.<+> (hsep . punctuate (text "\n") $ Prelude.map pPrint uvars)
              else text "")
      Doc.<> text "\n\n-> Rlps:\n"
        Doc.<+> (hsep . punctuate (text "\n") . Prelude.map (\r -> lparen
        Doc.<> text "rule:" Doc.<+> pPrint (trule $ rid r) Doc.<> comma
        Doc.<+> text "id:" Doc.<+> pPrint (tidr $ rid r) Doc.<> comma
        Doc.<+> text "possubterms: " Doc.<> (hsep . punctuate comma . Prelude.map(\ps-> pPrint (subterm ps) Doc.<> text "->"
          Doc.<> (brackets . hsep . punctuate comma . Prelude.map(\p-> pPrint p) $ (pos ps))) $ (pSubts r))  Doc.<> rparen) $ rlps)
      Doc.<> text "\n\n-> Unifications:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\u -> lparen
        Doc.<> text "R" Doc.<> pPrint (id1 u) Doc.<> text " unifies with "
        Doc.<> text "R" Doc.<> pPrint (id2 u) Doc.<> text " at "
        Doc.<> text "p:" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map (\pos -> pPrint pos) $ (p u)) Doc.<> comma
        Doc.<+> text "l:" Doc.<+> pPrint (l u) Doc.<> comma
        Doc.<+> text "lp:" Doc.<+> pPrint (lp u) Doc.<> comma
        Doc.<+> text "sig:" Doc.<+> (braces . hcat . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v ) $ (M.toList $ fromJust $ sigma u)) Doc.<> comma
        Doc.<+> text "l':" Doc.<+> pPrint (l1 u) Doc.<> comma
        Doc.<+> text "r:" Doc.<+> pPrint (r u) Doc.<> comma
        Doc.<+> text "r':" Doc.<+> pPrint (r1 u) Doc.<> rparen) $ unfs) 
      Doc.<> text "\n\n-> Critical pairs info:\n"
        Doc.<> (hcat . punctuate (text "\n") . Prelude.map (\cpi -> pPrint cpi) $ cpis)
      Doc.<> pPrint pconc Doc.<> text "\n"
    where trs = getConfProblem . getR $ problem


instance (HasName idV) 
  => Pretty (UVar idV) where
  pPrint (UVar
    { uvRuleId = i
    , uvLActive = la
    , uvRActive = ra
    , uvLFrozen = lf
    , uvRFrozen = rf
    }) = lparen Doc.<> text "UV-RuleId:" Doc.<+> int i Doc.<> comma
          Doc.<+> text "UV-LActive:" Doc.<+> (brackets . hsep . punctuate comma . Prelude.map(\a -> text (getName a)) $ (S.toList la)) Doc.<> comma
          Doc.<+> text "UV-RActive:" Doc.<+> (brackets . hsep . punctuate comma . Prelude.map(\a -> text (getName a)) $ (S.toList ra)) Doc.<> comma
          Doc.<+> text "UV-LFrozen:" Doc.<+> (brackets . hsep . punctuate comma . Prelude.map(\f -> text (getName f)) $ (S.toList lf)) Doc.<> comma
          Doc.<+> text "UV-RFrozen:" Doc.<+> (brackets . hsep . punctuate comma . Prelude.map(\f -> text (getName f)) $ (S.toList rf)) Doc.<> rparen


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Huet Levy Processor
huetLevyProc :: ( Processor info HuetLevyProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
huetLevyProc  = apply HuetLevyProcessor


-----------------------------------------------------------------------------
-- Examples
-----------------------------------------------------------------------------

{--
ff = IdFTRS { fid = 1, fname = "f", arity = 2}
cc = IdFTRS { fid = 2, fname = "c", arity = 0}
v1 = IdVTRS { vid = 1, vname = Just "x"}
v2 = IdVTRS { vid = 2, vname = Just "y"}

t1 :: Term IdFTRS IdVTRS
t1 = F ff [V v1, F ff [F ff [V v2, V v1], F cc []]]

t2 :: Term IdFTRS IdVTRS
t2 = F ff [V v2, V v1]

t3 :: Term IdFTRS IdVTRS
t3 = F ff [V v1, F ff [F ff [V v2, V v1], F cc []]]

t4 :: Term IdFTRS IdVTRS
t4 = F ff [V v2, V v1]

rule1 :: Rule (Term IdFTRS IdVTRS)
rule1 = t1 :-> t2

rule2 :: Rule (Term IdFTRS IdVTRS)
rule2 = t3 :-> t4

rule3 :: Rule (Term IdFTRS IdVTRS)
rule3 = t3 :-> t3

--}