{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.LHCPsProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that gets the LH_\mu-critical pairs
-- [JLAMP21-Def12]
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.LHCPsProcessor (

-- * Exported functions

lhcpsProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..), Substitution (..), isLinear, applySigma, matches, replace, vars, fresh)
import TRS.Identifiers (HasName (..), HasArity (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..))
import Framework.Problem.Types (InConfCSTRS (..), InCPCSTRS (..), InTermCSTRS (..), InCPTermCSTRS (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem, IsTProblem (..), IsCPProblem (..), IsCPTermProblem (..), 
                           CriticalPair (..), CriticalPairInfo (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, singleP, andP, Proof, Info (..))
import TRS.Rules(Rule(..), Cond(..), lhs, rhs, freshRule, ruleVars)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.TRSAux (RuleId (..), PosSubt (..), getRulesIds, isLHRV, activeVars, frozenVars, activeVarsPosByVars, isCSTRSSymb, cpsInfo, normalWeight)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes, int)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList, fromList)
import Data.Set as S (Set, elems, null, findMax, elemAt, intersection, insert, map, empty, union, delete, singleton, unions, fromList, member)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data LHCPsProcessor = LHCPsProcessor

-- | The information obtained are the rules
data LHCPsProcInfo problem idF idV = 
  LHCPsProcInfo { inProblem     :: problem
                , inLHCPs       :: [LHCP idF idV]
                }
  | LHCPsDontKnow { inProblem     :: problem
                  , inLHCPs       :: [LHCP idF idV]
                  }


data LHCP idF idV =
  LHCP { ruleidLHCP :: RuleId idF idV
       , varXLHCP    :: idV
       , varYLHCP    :: idV
       , pLHCP      :: [Int]
       , sLHCP      :: Term idF idV
       , tLHCP      :: Term idF idV
       , normWLHCP  :: Int
       }
    deriving (Eq,Ord) 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InConfCSTRS trs
  , conftrs2 ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsCPProblem conftrs2 (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF 
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV, Ord (RuleId idF idV)
  , Info info (LHCPsProcInfo (Problem Rewriting conftrs1) idF idV)
  ) => Processor info LHCPsProcessor (Problem Rewriting conftrs1) (Problem CPRewriting conftrs2) where
  apply LHCPsProcessor inP
    -- | True = success (LHCPsProcInfo inP lhcps) inP
    | not cstrs = dontKnow (LHCPsDontKnow inP lhcps) inP
    | otherwise = singleP (LHCPsProcInfo inP lhcps) inP (mkProblem (CPRewriting (getPSchemeFromProblem inP)) (InCPCSTRS newTrs cpis))
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          cstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          variables = getTRSVariables trs
          yid = getNonRepeatedVarIdForY variables Nothing
          rIds = getRulesIds rules
          lhcps = elems $ getLHCPs rIds yid
          lhcpsNW = [getLHCPNW rIds lhcp | lhcp <- lhcps]
          cps = Prelude.map toCP lhcpsNW
          cpis = cpsInfo cps 1
          insertedVar
            | Prelude.null lhcps = S.empty
            | otherwise = S.singleton . varYLHCP $ head lhcps
          newTrs = setTRSVariables (insertedVar `S.union` variables) trs

instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InTermCSTRS trs (Maybe(Bool))
  , conftrs2 ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsTProblem conftrs1 (Maybe(Bool))
  , IsCPTermProblem conftrs2 (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF 
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV, Ord (RuleId idF idV)
  , Info info (LHCPsProcInfo (Problem TRewriting conftrs1) idF idV)
  ) => Processor info LHCPsProcessor (Problem TRewriting conftrs1) (Problem CPTRewriting conftrs2) where
  apply LHCPsProcessor inP
    -- | True = success (LHCPsProcInfo inP lhcps) inP
    | not cstrs = dontKnow (LHCPsDontKnow inP lhcps) inP
    | otherwise = singleP (LHCPsProcInfo inP lhcps) inP (mkProblem (CPTRewriting (getPSchemeFromProblem inP)) (InCPTermCSTRS newTrs cpis termination))
    where trs = getConfProblem . getR $ inP
          termination = getTermination . getR $ inP
          rules = elems . getTRSRules $ trs
          cstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          variables = getTRSVariables trs
          yid = getNonRepeatedVarIdForY variables Nothing
          rIds = getRulesIds rules
          lhcps = elems $ getLHCPs rIds yid
          lhcpsNW = [getLHCPNW rIds lhcp | lhcp <- lhcps]
          cps = Prelude.map toCP lhcpsNW
          cpis = cpsInfo cps 1
          insertedVar
            | Prelude.null lhcps = S.empty
            | otherwise = S.singleton . varYLHCP $ head lhcps
          newTrs = setTRSVariables (insertedVar `S.union` variables) trs


          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (LHCPsProcInfo (Problem Rewriting conftrs) idF idV) where
  pPrint LHCPsProcInfo { inProblem = problem
                       , inLHCPs = lhcps
                       }
    = pPrint problem
      Doc.<> text "\nLH u-Critical Pairs Processor [JLAMP21-Def12]:\n"
      Doc.<> (if Prelude.null lhcps then text "No LHCPs found" else (hcat . punctuate (text "\n\n") . Prelude.map (\lh -> pPrint lh) $ lhcps))

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem TRewriting conftrs)
  ) => Pretty (LHCPsProcInfo (Problem TRewriting conftrs) idF idV) where
  pPrint LHCPsProcInfo { inProblem = problem
                       , inLHCPs = lhcps
                       }
    = pPrint problem
      Doc.<> text "\nLH u-Critical Pairs Processor [JLAMP21-Def12]:\n"
      Doc.<> (if Prelude.null lhcps then text "No LHCPs found" else (hcat . punctuate (text "\n\n") . Prelude.map (\lh -> pPrint lh) $ lhcps))


instance (
  Pretty idF, Pretty idV
  ) => Pretty (LHCP idF idV) where
  pPrint LHCP { ruleidLHCP = rid
              , varXLHCP = vx
              , varYLHCP = vy
              , pLHCP = p
              , sLHCP = s
              , tLHCP = t
              , normWLHCP = nw
              }
    = text "->LH u-Critical Pair:\n"
      Doc.<+> text "Rule" Doc.<+> int (tidr rid) Doc.<+> text "(l :-> r) =>" Doc.<+> pPrint (trule rid) Doc.<> text "\n"
      Doc.<+> text "Var x =>" Doc.<+> pPrint vx Doc.<> text "\n"
      Doc.<+> text "Var y =>" Doc.<+> pPrint vy Doc.<> text "\n"
      Doc.<+> text "Pos" Doc.<+> pPrint vx Doc.<+> text "in l =>" Doc.<+> (brackets . hcat . punctuate comma . Prelude.map pPrint $ p) Doc.<> text "\n"
      Doc.<+> text "s =>" Doc.<+> pPrint s Doc.<> text "\n"
      Doc.<+> text "t =>" Doc.<+> pPrint t Doc.<> text "\n"
      Doc.<+> text "C =>" Doc.<+> pPrint vx Doc.<+> text "\\->" Doc.<+> pPrint vy Doc.<+> text "\n"
      Doc.<+> text "NW =>" Doc.<+> pPrint nw Doc.<+> text "\n"
      


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the LH u-Critical Pairs Processor
lhcpsProc :: ( Processor info LHCPsProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => o -> Proof info mp d
lhcpsProc = apply LHCPsProcessor


getLHCPs :: (HasName idV, Pretty idV, HasMu idF, HasArity idF, Ord idF, Ord idV, Ord (RuleId idF idV), Eq idV) => [RuleId idF idV] -> Maybe Int -> Set (LHCP idF idV)
getLHCPs [] _ = S.empty
getLHCPs (rid:rids) yid
  | S.null validVsPos = getLHCPs rids yid
  | not . isLHRV $ trule rid = S.union lhcps (getLHCPs rids yid)
  | otherwise = getLHCPs rids yid
  where (l :-> r) = trule rid
        frozVs = (frozenVars l) `union` (frozenVars r)
        validVsPos = activeVarsPosByVars l frozVs
        vy = if isJust yid
              then setName (Just ("y" ++ (show $ fromJust yid))) (fst $ elemAt 0 validVsPos)
              else setName (Just "y") (fst $ elemAt 0 validVsPos)
        fvy = V vy
        lhcps = S.map (\(vx,p) -> (LHCP rid vx vy p (replace fvy p l) r 0)) $ validVsPos


toCP :: (HasName idV) => LHCP idF idV -> CriticalPair idF idV
toCP (LHCP _ vx vy p sEcp tEcp nw) = CriticalPair sEcp tEcp p [cs] nw
  where cs = x :\-> y
        --x = V ((setName (Just "x") . (setId $ getId v)) v)
        --y = V ((setName (Just "y") . (setId $ getId v)) v)
        x = V vx
        y = V vy

getLHCPNW :: (Ord idV, Eq idF, HasMu idF) => [RuleId idF idV] -> LHCP idF idV -> LHCP idF idV
getLHCPNW rIds (LHCP rid vx vy p sEcp tEcp _) = (LHCP rid vx vy p sEcp tEcp nw)
  where nw = ((normalWeight sEcp rIds) + (normalWeight tEcp rIds))


getNonRepeatedVarIdForY :: (HasName idV) => Set idV -> Maybe Int -> Maybe Int
getNonRepeatedVarIdForY vs Nothing
    | member newN names = getNonRepeatedVarIdForY vs (Just 1)
    | otherwise = Nothing
    where names = S.map getName vs
          newN = "y"
getNonRepeatedVarIdForY vs (Just n)
    | member newN names = getNonRepeatedVarIdForY vs (Just (n + 1))
    | otherwise = Just n
    where names = S.map getName vs
          newN = "y" ++ (show n)
