{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.LHRVProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that checks Left-Homogeneous \mu-Replacing Variables
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.LHRVProcessor (

-- * Exported functions

lhrvProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term)
import TRS.Identifiers (HasArity (..), HasMu (..))
import TRS.Rules(Rule(..), lhs, rhs)
import TRS.Terms(Term(..))
import Framework.Problem.Types(InConfCSTRS (..),)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem)
import Framework.TRSAux (isLHRV, isCSTRSSymb)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (dontKnow, success, singleP, Proof, Info (..), refuted)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>))
import Control.Monad (MonadPlus)
import Data.Set as S (Set, elems)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | The extra variables processor check if there is a rule with extra
-- variables
data LHRVProcessor  = LHRVProcessor

-- | The information obtained are the rules
data LHRVProcInfo problem = 
    LHRVProcInfo { inProblem :: problem 
                  , inCSTRS  :: Bool
                  , inLHRV   :: Bool
                 }
  | LHRVProcDontKnow { inProblem :: problem 
                     , inCSTRS  :: Bool
                     , inLHRV   :: Bool
                     }

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
  conftrs ~ InConfCSTRS trs
  , IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasArity idF
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  , Info info (LHRVProcInfo (Problem Rewriting conftrs))
  ) => Processor info LHRVProcessor (Problem Rewriting conftrs) (Problem Rewriting conftrs) where
  apply LHRVProcessor inP
    | cstrs && lhrv = singleP (LHRVProcInfo inP cstrs lhrv) inP inP
    | otherwise = dontKnow (LHRVProcDontKnow inP cstrs lhrv) inP --success (LHRVProcInfo inP cstrs lhrv) inP --
    -- | not cstrs = success (LHRVProcInfo inP cstrs lhrv) inP --success (LHRVProcInfo inP cstrs lhrv) inP --
    -- | lhrv = success (LHRVProcInfo inP cstrs lhrv) inP
    -- | otherwise = refuted (LHRVProcInfo inP cstrs lhrv) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          cstrs = or . map isCSTRSSymb . elems $ getTRSSymbols trs
          lhrv = and $ map isLHRV rules

-- Pretty

instance (
  IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (LHRVProcInfo (Problem Rewriting conftrs)) where
  pPrint LHRVProcInfo { inProblem = problem
                      , inCSTRS = False
                      , inLHRV = lhrv
                      }
    = text "\nLeft-Homogeneous u-Replacing Variables Processor:\n"
      Doc.<> text "R is not a CSTRS"
  pPrint LHRVProcInfo { inProblem = problem
                      , inCSTRS = cstrs
                      , inLHRV = True
                      }
    = text "\nLeft-Homogeneous u-Replacing Variables Processor:\n"
      Doc.<> text "R satisfies LHRV condition"
  pPrint LHRVProcInfo { inProblem = problem
                      , inCSTRS = cstrs
                      , inLHRV = False
                      }
    = text "\nLeft-Homogeneous u-Replacing Variables Processor:\n"
      Doc.<> text "R does not satisfy LHRV condition"

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Clean TRS Processor
lhrvProc :: (Processor info LHRVProcessor o d
          , MonadPlus mp
          , Info info o
          , Info info d) => o -> Proof info mp d
lhrvProc  = apply LHRVProcessor
