{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.ModularConfCombsTRSProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the mccTRSProc processor
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.ModularConfCombsTRSProcessor (

-- * Exported functions

mccTRSProc

) where

import TRS.TRSs(TRS (..), Signature (..), HasSignature (..), HasRules (..), HasOSSignature(..), isLeftLinear)
import TRS.Terms(Term, root)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity (..), HasMu (..))
import TRS.Rules(Rule(..), extractCompRule, ruleVars, ruleSymbols, isCollapsing, rhs, lhs)
import TRS.Terms(Term(..))
import Framework.Problem.Types(InConfCSTRS (..),)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problems (IsConfProblem (..), getPSchemeFromProblem)
import Framework.TRSAux (isCSTRSSymb)
import Utils.Auxiliar (boolGroups)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (singleP, Proof, Info (..), andP, success, dontKnow, refuted)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..))
import Data.Set as S (Set, elems, fromList, unions, union, map, intersection, toList, empty, isSubsetOf, singleton, (\\), member, size)
import Data.Map as M (Map (..), fromList)
import Data.Maybe (isJust, fromJust, isNothing)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------


data ModularConfCombsTRSProcessor  = ModularConfCombsTRSProcessor

data MCCTRSProcInfo problem trs idF idV = MCCTRSProcInfo { inProblem :: problem 
                                                         , trs1  :: Maybe (trs)
                                                         , trs2  :: Maybe (trs)
                                                         , trsc  :: [TrsCombination idF idV]
                                                         , cstrs :: Bool
                                                         }

data TrsCombination idF idV =
  TrsCombination { rules1          :: Set (Rule (Term idF idV))
                 , rules2          :: Set (Rule (Term idF idV))
                 , disjoint        :: Bool
                 , constshar       :: Bool
                 --, composable      :: Bool
                 , leftlinear      :: Bool
                 , layerpreserving :: Bool
                 }

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
  conftrs ~ InConfCSTRS trs
  , IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasArity idF
  , HasMu idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  , Info info (MCCTRSProcInfo (Problem Rewriting conftrs) trs idF idV)
  ) => Processor info ModularConfCombsTRSProcessor (Problem Rewriting conftrs) (Problem Rewriting conftrs) where
  apply ModularConfCombsTRSProcessor inP
    | cstrs = singleP (MCCTRSProcInfo inP t1 t2 trsc True) inP (mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS trs))--success (MCCTRSProcInfo inP t1 t2 trsc True) inP--
    | isNothing t1 || isNothing t2 = singleP (MCCTRSProcInfo inP t1 t2 trsc False) inP (mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS trs))--refuted (MCCTRSProcInfo inP t1 t2 trsc False) inP--
    | otherwise = andP (MCCTRSProcInfo inP t1 t2 trsc False) inP newPs --success (MCCTRSProcInfo inP t1 t2 trsc False) inP--
    where trs = getConfProblem . getR $ inP
          cstrs = or . Prelude.map isCSTRSSymb . elems $ getTRSSymbols trs
          (t1, t2, trsc) =  disjointTRSs trs
          trss = t1:[t2]
          newPs = [mkProblem (Rewriting (getPSchemeFromProblem inP)) (InConfCSTRS (fromJust t)) | t <- trss]

-- Pretty

instance (
  IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (MCCTRSProcInfo (Problem Rewriting conftrs) trs idF idV) where
  pPrint MCCTRSProcInfo { inProblem = problem
                        , trs1 = t1
                        , trs2 = t2
                        , trsc = trsc
                        , cstrs = False
                        }
    = text "\n\nModular Confluence Combinations Decomposition Processor:\n"
      Doc.<> text "\n" Doc.<> (hcat . punctuate ( text "\n\n") . Prelude.map(\t -> text "TRS combination:\n"
        Doc.<> (braces . hsep . punctuate (text "\n") . Prelude.map(\rr1 -> pPrint rr1) $ (S.toList $ rules1 t)) Doc.<> text "\n"
        Doc.<> (braces . hsep . punctuate (text "\n") . Prelude.map(\rr2 -> pPrint rr2) $ (S.toList $ rules2 t)) Doc.<> text "\n"
        Doc.<> (if (disjoint t) then text "Disjoint" else text "Not disjoint") Doc.<> text "\n"
        Doc.<> (if (constshar t) then text "Constructor-sharing" else text "Not constructor-sharing") Doc.<> text "\n"
        --Doc.<> (if (composable t) then text "Composable" else text "Not composable") Doc.<> text "\n"
        Doc.<> (if (leftlinear t) then text "Left linear" else text "Not left linear") Doc.<> text "\n"
        Doc.<> (if (layerpreserving t) then text "Layer-preserving" else text "Not layer-preserving") Doc.<> text "\n"
      ) $ trsc)
      Doc.<> (
      if (isNothing t1) || (isNothing t2) then text "No usable combinations"
      else text "TRS1\n" Doc.<> pPrint t1 Doc.<> text "\n\nTRS2\n" Doc.<> pPrint t2
      )
  pPrint MCCTRSProcInfo { inProblem = problem
                        , trs1 = t1
                        , trs2 = t2
                        , trsc = trsc
                        , cstrs = True
                        }
    = text "\n\nModular Confluence Combinations Decomposition Processor:\n"
      Doc.<> text "It is a CS-TRS -> No modular confluence"

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Modular Confluence Combinations Decomposition Processor
mccTRSProc :: (Processor info ModularConfCombsTRSProcessor o d
          , MonadPlus mp
          , Info info o
          , Info info d) => o -> Proof info mp d
mccTRSProc  = apply ModularConfCombsTRSProcessor


disjointTRSs :: (HasArity idF, Ord idF, Ord idV, HasSignature trs idF idV, HasRules trs (Term idF idV)) => trs -> (Maybe (trs),Maybe (trs), [TrsCombination idF idV])
disjointTRSs trs
  | null r1 || null r2 = (Nothing, Nothing,trsc)
  | otherwise = (t1,t2,trsc)
  where rules = elems . getTRSRules $ trs
        bgs = boolGroups (length rules)
        ll = isLeftLinear trs
        trsc = modularityConfluent $ disjointRulesBGs ll rules bgs
        (r1,r2) = if length trsc > 0 then ((rules1 (trsc!!0)), (rules2 (trsc!!0))) else (S.empty, S.empty)
        sig = getTRSSignature trs
        lab = getTRSLabel trs
        v1 = unions . S.elems $ S.map ruleVars r1
        v2 = unions . S.elems $ S.map ruleVars r2
        f1 = unions . S.elems $ S.map ruleSymbols r1
        f2 = unions . S.elems $ S.map ruleSymbols r2
        d1 = unions . S.elems $ S.map ruleDefinedSymbol r1
        d2 = unions . S.elems $ S.map ruleDefinedSymbol r2
        c1 = f1 \\ d1
        c2 = f2 \\ d2
        t1 = Just (setTRSCSymbols c1 . setTRSDSymbols d1 . setTRSRules r1 $ setTRSVariables v1 trs)
        t2 = Just (setTRSCSymbols c2 . setTRSDSymbols d2 . setTRSRules r2 $ setTRSVariables v2 trs)


modularityConfluent :: [TrsCombination idF idV] -> [TrsCombination idF idV]
modularityConfluent [] = []
modularityConfluent (tc:trsc)
  | (disjoint tc) || ((constshar tc) && ((leftlinear tc) || (layerpreserving tc))) = tc:(modularityConfluent trsc) -- || ((composable tc) && (layerpreserving tc)) = tc:(modularityConfluent trsc)
  | otherwise = modularityConfluent trsc


disjointRulesBGs :: (HasArity idF, Ord idF, Ord idV) => Bool -> [Rule (Term idF idV)] -> [[Bool]] -> [TrsCombination idF idV]
disjointRulesBGs _ _ [] = []
disjointRulesBGs ll rs (bg:bgs) = trsc:(disjointRulesBGs ll rs bgs)
  where (r1,r2) = extractRulesBG ([],[]) rs bg
        (sr1, sr2) = ((S.fromList r1), (S.fromList r2))
        f1 = unions . S.elems $ S.map ruleSymbols sr1
        f2 = unions . S.elems $ S.map ruleSymbols sr2
        d1 = unions . S.elems $ S.map ruleDefinedSymbol sr1
        d2 = unions . S.elems $ S.map ruleDefinedSymbol sr2
        c1 = f1 \\ d1
        c2 = f2 \\ d2
        symbInters = intersection f1 f2
        disj = null symbInters
        cSharing = (null $ intersection c1 d2) && (null $ intersection d1 c2) && (null $ intersection d1 d2)
        --comp = (null $ intersection c1 d2) && (null $ intersection c2 d1) 
        --    && (areFsInRootL (S.toList $ intersection d1 d2) (S.toList sr1)) && (areFsInRootL (S.toList $ intersection d1 d2) (S.toList sr2))
        noCollap = not . or $ Prelude.map isCollapsing rs
        lp = (disj && noCollap)
          || (cSharing && noCollap && (not $ isContructorLifting (S.toList (intersection c1 c2)) rs))
        --  || (comp && isCompLayerPreserving f1 f2 r1 r2)
        trsc = TrsCombination sr1 sr2 disj cSharing ll lp --TrsCombination sr1 sr2 disj cSharing comp ll lp

{-
isCompLayerPreserving :: (HasArity idF, Ord idF, Ord idV) => Set idF -> Set idF -> [Rule (Term idF idV)] -> [Rule (Term idF idV)] -> Bool
isCompLayerPreserving f1 f2 r1 r2 = (isCompLayerPreserving' diff12 r1) && (isCompLayerPreserving' diff21 r2)
    where diff12 = f1 \\ f2
          diff21 = f2 \\ f1


isCompLayerPreserving' :: (HasArity idF, Ord idF, Ord idV) => Set idF -> [Rule (Term idF idV)] -> Bool
isCompLayerPreserving' _ [] = True
isCompLayerPreserving' d (r:rs)
  | null d = True
  | ((not $ isNothing rlt) && (member (fromJust rlt) d) && ((isNothing rrt) || (not $ member (fromJust rrt) d))) = False
  | otherwise = isCompLayerPreserving' d rs
    where lt = lhs r
          rt = rhs r
          rlt = root lt
          rrt = root rt
-}

extractRulesBG :: (HasArity idF, Ord idF) 
                    => ([Rule (Term idF idV)], [Rule (Term idF idV)]) -> [Rule (Term idF idV)] -> [Bool] -> ([Rule (Term idF idV)], [Rule (Term idF idV)])
extractRulesBG (a1,a2) [] [] = (a1,a2)
extractRulesBG (a1,a2) (r:rs) (b:bs)
  | b = extractRulesBG (r:a1, a2) rs bs
  | otherwise = extractRulesBG (a1, r:a2) rs bs


ruleDefinedSymbol :: (HasArity idF, Ord idF, Ord idV, Eq idF) => Rule (Term idF idV) -> Set idF
ruleDefinedSymbol (l :-> r) = termDefinedSymbol l


termDefinedSymbol :: (HasArity idF, Ord idF, Ord idV, Eq idF) => Term idF idV -> Set idF
termDefinedSymbol (F f ts) = singleton f
termDefinedSymbol _ = S.empty


isContructorLifting :: (HasArity idF, Ord idF, Ord idV, Eq idF) => [idF] -> [Rule (Term idF idV)] -> Bool
isContructorLifting _ [] = False
isContructorLifting cs (r:rs)
  | isRoot cs (rhs r) = True
  | otherwise = isContructorLifting cs rs


isRoot :: (HasArity idF, Ord idF, Ord idV, Eq idF) => [idF] -> Term idF idV -> Bool
isRoot [] _ = False
isRoot (c:cs) (F f tt)
  | c == f = True
  | otherwise = isRoot cs (F f tt)
isRoot _ _ = False


{-
areFsInRootL :: (HasArity idF, Ord idF, Ord idV, Eq idF) => [idF] -> [Rule (Term idF idV)] -> Bool
areFsInRootL [] _ = True
areFsInRootL (d:ds) rs
  | or [isRoot [d] (lhs (rs!!i)) | i<-[0..(length rs)]] = areFsInRootL ds rs
  | otherwise = False
-}
