{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.NoConvBFProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence
-- with maybe terminating problems
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.NoConvBFProcessor (

-- * Exported functions

ncbfProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), HasArity (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPProblem (..), IsCPTermProblem (..))
import Framework.Problem.Types (InCPCSTRS (..), InCPTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..))
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, andP, Proof, Info (..), refuted)
import TRS.Rules(Rule(..), freshRule)
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.TRSAux (RuleId (..), JoinabilityInfo (..), RewTerms (..), maxDepthReached, createRewritingGraph, haveJoinTerm, getRulesIds)
import Utils.Auxiliar (isTrue)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Set as S (Set, elems, null, findMax)
import Data.Maybe (isJust, fromJust, isNothing)

import Data.Graph.Inductive.Graph as G
import Data.Graph.Inductive.PatriciaTree (Gr (..))

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data NoConvBFProcessor = NoConvBFProcessor  { maxDepth :: Maybe (Int), maxCPOut :: Maybe (Int) }

-- | The information obtained are the rules
data NoConvBFProcInfo problem idF idV = 
  NoConvBFProcInfo { inProblem   :: problem
                   , inCpi        :: CriticalPairInfo idF idV
                   , inJoinInfo  :: JoinabilityInfo idF idV
                   , inSGraph    :: Gr (RewTerms idF idV) () 
                   , inTGraph    :: Gr (RewTerms idF idV) () 
                   }
  | NoConvBFDontKnow { inProblem   :: problem
                     , inCpi        :: CriticalPairInfo idF idV
                     , inJoinInfo  :: JoinabilityInfo idF idV
                     , inSGraph    :: Gr (RewTerms idF idV) () 
                     , inTGraph    :: Gr (RewTerms idF idV) () 
                     }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-}(
    conftrs ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF 
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (NoConvBFProcInfo (Problem CPRewriting conftrs) idF idV)
  ) => Processor info NoConvBFProcessor (Problem CPRewriting conftrs) (Problem CPRewriting conftrs) where
  apply NoConvBFProcessor {maxDepth = md, maxCPOut = mcp} inP
    = if noconv
      then refuted (NoConvBFProcInfo inP cpi j sg tg) inP
      else dontKnow (NoConvBFDontKnow inP cpi j sg tg) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          ncp = cpNum cpi
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          (j,sg, tg) = isNoConvergent (getRulesIds freshRules) cp (md,mcp,ncp)
          mdReached = isTrue (maxDReached j)
          noconv = (not $ joinable j) && (not mdReached)

instance {-# OVERLAPPABLE #-}(
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF 
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (NoConvBFProcInfo (Problem CPTRewriting conftrs) idF idV)
  ) => Processor info NoConvBFProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply NoConvBFProcessor {maxDepth = md, maxCPOut = mcp} inP
    = if noconv
      then refuted (NoConvBFProcInfo inP cpi j sg tg) inP
      else dontKnow (NoConvBFDontKnow inP cpi j sg tg) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          cpis = getCPairsCPT . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          ncp = cpNum cpi
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          (j,sg, tg) = isNoConvergent (getRulesIds freshRules) cp (md,mcp,ncp)
          mdReached = isTrue (maxDReached j)
          noconv = (not $ joinable j) && (not mdReached)

          
-- Pretty

instance {-# OVERLAPPABLE #-}(
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPRewriting conftrs)
  ) => Pretty (NoConvBFProcInfo (Problem CPRewriting conftrs) idF idV) where
  pPrint NoConvBFProcInfo { inProblem = problem
                          , inCpi = cpi
                          , inJoinInfo = j
                          , inSGraph = sg
                          , inTGraph = tg
                          }
    = text "\nNo Convergence Brute Force Processor:\n"
    Doc.<> text "-> Rewritings:\n"
      Doc.<> text "s: " Doc.<> pPrint (s cp) Doc.<> text "\n"
      Doc.<> text "Nodes: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes sg)) Doc.<> text "\n" 
      Doc.<> text "Edges: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges sg)) Doc.<> text "\n"
        Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes sg)) Doc.<> text "\n"
      Doc.<> text "t: " Doc.<> pPrint (t cp) Doc.<> text "\n" 
      Doc.<> text "Nodes: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes tg)) Doc.<> text "\n" 
      Doc.<> text "Edges: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges tg)) Doc.<> text "\n"
        Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes tg)) Doc.<> text "\n"
      Doc.<> pPrint (s cp) Doc.<> text " ->* "
      Doc.<> (if isNothing (union j) then text "no union" else pPrint (fromJust $ union j))
      Doc.<> text " *<- " Doc.<> pPrint (t cp) Doc.<> text "\n"
      Doc.<> pPrint (if joinable j then  "Joinable" else "Not joinable")
    where trs = getConfProblem . getR $ problem
          cp = critPair cpi

instance {-# OVERLAPPABLE #-}(
    IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPRewriting conftrs)
  ) => Pretty (NoConvBFProcInfo (Problem CPTRewriting conftrs) idF idV) where
  pPrint NoConvBFProcInfo { inProblem = problem
                          , inCpi = cpi
                          , inJoinInfo = j
                          , inSGraph = sg
                          , inTGraph = tg
                          }
    = text "\nNo Convergence Brute Force Processor:\n"
    Doc.<> text "-> Rewritings:\n"
      Doc.<> text "s: " Doc.<> pPrint (s cp) Doc.<> text "\n"
      Doc.<> text "Nodes: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes sg)) Doc.<> text "\n" 
      Doc.<> text "Edges: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges sg)) Doc.<> text "\n"
        Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes sg)) Doc.<> text "\n"
      Doc.<> text "t: " Doc.<> pPrint (t cp) Doc.<> text "\n" 
      Doc.<> text "Nodes: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes tg)) Doc.<> text "\n" 
      Doc.<> text "Edges: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges tg)) Doc.<> text "\n"
        Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes tg)) Doc.<> text "\n"
      Doc.<> pPrint (s cp) Doc.<> text " ->* "
      Doc.<> (if isNothing (union j) then text "no union" else pPrint (fromJust $ union j))
      Doc.<> text " *<- " Doc.<> pPrint (t cp) Doc.<> text "\n"
      Doc.<> pPrint (if joinable j then  "Joinable" else "Not joinable")
    where trs = getConfProblem . getR $ problem
          cp = critPair cpi



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the No Convergence Brute Force Processor
ncbfProc :: ( Processor info NoConvBFProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe(Int) -> Maybe(Int) -> o -> Proof info mp d
ncbfProc md mcp = apply NoConvBFProcessor { maxDepth = md
                                          , maxCPOut = mcp
                                          }


isNoConvergent :: (HasMu idF, Eq idF, Ord idV) => [RuleId idF idV] -> CriticalPair idF idV -> (Maybe(Int),Maybe(Int),Int)
  -> (JoinabilityInfo idF idV, Gr (RewTerms idF idV) (), Gr (RewTerms idF idV) ())
isNoConvergent rs cp (md,mcp,ncp) = areNoConvergent (s cp) (t cp) rs (md,mcp,ncp)


areNoConvergent :: (HasMu idF, Eq idF, Ord idV) => Term idF idV -> Term idF idV -> [RuleId idF idV] -> (Maybe(Int),Maybe(Int),Int)
  -> (JoinabilityInfo idF idV, Gr (RewTerms idF idV) (), Gr (RewTerms idF idV) ())
areNoConvergent s t rs (md,(Just mcp),ncp)
  | mdr = (JoinabilityInfo False Nothing (Just mdr) [] [], sGraph, tGraph)
  | ncp >= mcp = (JoinabilityInfo j u (Just mdr) [] [], G.empty, G.empty)
  | otherwise = (JoinabilityInfo j u (Just mdr) [] [], sGraph, tGraph)
  where sGraph = createRewritingGraph s rs md
        tGraph = createRewritingGraph t rs md
        sRewTerms = map snd (labNodes sGraph)
        tRewTerms = map snd (labNodes tGraph)
        mdr = (maxDepthReached md sGraph) || (maxDepthReached md tGraph)
        --(j, u) = haveJoinTerm (map term sRewTerms) (map term tRewTerms)
        (j, u, un1, un2) = haveJoinTerm sGraph tGraph