{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.NoConvInfProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check no convergence using ground 
-- terms calling to InfChecker
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.NoConvInfProcessor (

-- * Exported functions

ncIProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPProblem (..), IsCPTermProblem (..))
import Framework.Problem.Types (InConfCSTRS (..), InCPCSTRS (..), InCPTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteria)
import Framework.TRSAux (isCSTRSSymb, isLHRV, getRepMap)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, rparen, lparen, hsep, comma)
import Data.Set as S (Set, elems)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data NoConvInfProcessor = NoConvInfProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data NoConvIProcInfo problem = 
  NoConvIProcInfo { inProblem :: problem 
                  , inICCrit  :: InfCheckerCriteria
                  , inSol     :: String
                  }
  | NoConvIDontKnow { inProblem :: problem 
                    , inICCrit :: InfCheckerCriteria
                    , inSol    :: String  
                    }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPCSTRS trs (CriticalPairInfo idF idV)
  , IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (NoConvIProcInfo (Problem CPRewriting conftrs))
  ) => Processor info NoConvInfProcessor (Problem CPRewriting conftrs) (Problem CPRewriting conftrs) where
  apply NoConvInfProcessor {timeout = to} inP
    = unsafePerformIO $!
        do { 
            solution <- solveInfChecker to trs icc
            ; case solution of
                ("YES", Just sol) -> return (refuted (NoConvIProcInfo inP icc sol) inP)
                (_, Just sol)     -> return (dontKnow (NoConvIDontKnow inP icc sol) inP)
           }
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          symbs =  elems . getTRSSymbols $ trs
          cstrs = or $ map isCSTRSSymb symbs
          lhrv = and $ map isLHRV rules
          cpis = getCPairs . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          icc = buildInfCheckerCriteria trs cp

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (NoConvIProcInfo (Problem CPTRewriting conftrs))
  ) => Processor info NoConvInfProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply NoConvInfProcessor {timeout = to} inP
    = unsafePerformIO $!
        do { 
            solution <- solveInfChecker to trs icc
            ; case solution of
                ("YES", Just sol) -> return (refuted (NoConvIProcInfo inP icc sol) inP)
                (_, Just sol)     -> return (dontKnow (NoConvIDontKnow inP icc sol) inP)
           }
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          symbs =  elems . getTRSSymbols $ trs
          cstrs = or $ map isCSTRSSymb symbs
          lhrv = and $ map isLHRV rules
          cpis = getCPairsCPT . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          icc = buildInfCheckerCriteria trs cp

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPProblem conftrs (CriticalPairInfo idF idV)
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPRewriting conftrs)
  ) => Pretty (NoConvIProcInfo (Problem CPRewriting conftrs)) where
  pPrint NoConvIProcInfo { inProblem = problem
                         , inICCrit = crit
                         , inSol = sol
                         }
    = pPrint problem
    Doc.<> text "\nNo Convergence InfChecker Processor:\n"
    Doc.<> text "Infeasible convergence?\n"
      Doc.<> (hcat . punctuate (text "") . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPTRewriting conftrs)
  ) => Pretty (NoConvIProcInfo (Problem CPTRewriting conftrs)) where
  pPrint NoConvIProcInfo { inProblem = problem
                         , inICCrit = crit
                         , inSol = sol
                         }
    = pPrint problem
    Doc.<> text "\nNo Convergence InfChecker Processor:\n"
    Doc.<> text "Infeasible convergence?\n"
      Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the No Convergence InfChecker Processor
ncIProc :: ( Processor info NoConvInfProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
ncIProc to = apply NoConvInfProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveInfChecker :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasMu idF, Ord idF, Ord idV, Pretty idF, Pretty idV) 
                 => Maybe Int -> trs -> InfCheckerCriteria -> IO (String, Maybe String)
solveInfChecker timeout trs crit
  = callToInfChecker timeout infProblem
  where
    replMap = getRepMap trs
    infProblem = bconcat $ [ frmStr "(PROBLEM INFEASIBILITY)\n"
                           , frmStr "(VAR "
                           , frmStr . noRiskyChars $ show . hcat . punctuate (text " ") . map (\v -> pPrint v) $ (trsVars crit)
                           , frmStr ")\n"
                           , frmStr "(STRATEGY CONTEXTSENSITIVE\n"
                           , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map pPrint $ mu) Doc.<> rparen) $ (M.toList replMap)
                           , frmStr "\n)\n"
                           , frmStr "(RULES\n"
                           , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (rulesICC crit)
                           , frmStr "\n)\n"
                           , frmStr "(VAR "
                           , frmStr . noRiskyChars $ show . text $ infCheckerJVN crit 
                           , frmStr ")\n"
                           , frmStr "(CONDITION "
                           , frmStr . noRiskyChars $ show $ pPrint (fst $ cpsConst crit) Doc.<+> text "==" Doc.<+> text (infCheckerJVN crit) Doc.<> comma
                                      Doc.<+> pPrint (snd $ cpsConst crit) Doc.<+> text "==" Doc.<+> text (infCheckerJVN crit)
                           , frmStr ")"]
