{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.Prov9LHCPsConvProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check convergence of LHCPs
-- calling to Prover9 [JLAMP21-ProvingAndDisprovingConfluenceOfCSR-Prop24]
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.Prov9LHCPsConvProcessor (

-- * Exported functions

prov9LHCPCvProc

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Rules(Rule (..), CRule (..), lcond, rcond, lhs, rhs, conds, rule, ruleVars)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..), HasMu(..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPTermProblem (..), getPSchemeFromProblem)
import Framework.Problem.Types (InCPTermCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.ExternalCalls (callToProver9)
import Framework.ExtCallsCriteria (Prover9Criteria (..), buildProver9CriteriaCSR)
import Framework.ProblemConclusions (ProblemConclusionsCTRS (..))
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, comma, hsep, lparen, rparen, char)
import Data.Set as S (Set, elems, null, findMax, union, unions, intersection, isSubsetOf, fromList, empty)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map)
import Data.Maybe (isJust, fromJust, isNothing)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data Prov9LHCPsConvProcessor = Prov9LHCPsConvProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data Prov9LHCPsCvProcInfo problem idF idV = 
  Prov9LHCPsCvProcInfo { inProblem            :: problem 
                       , inP9Crit             :: Prover9Criteria idF idV
                       , inSol                :: String
                       }
  | Prov9LHCPsCvDontKnow { inProblem            :: problem 
                         , inP9Crit             :: Prover9Criteria idF idV
                         , inSol                :: String 
                         }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPTermCSTRS trs (CriticalPairInfo idF idV) (Maybe(Bool))
  , IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasMu idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV, Ord String
  , Info info (Prov9LHCPsCvProcInfo (Problem CPTRewriting conftrs) idF idV)
  ) => Processor info Prov9LHCPsConvProcessor (Problem CPTRewriting conftrs) (Problem CPTRewriting conftrs) where
  apply Prov9LHCPsConvProcessor {timeout = to} inP
   | isTrue term = unsafePerformIO $!
                do { 
                    solution <- solveProver9 to trs p9c
                    ; case solution of
                        ("YES", Just sol) -> return (success (Prov9LHCPsCvProcInfo inP p9c sol) inP)
                        ("NO", Just sol)  -> return (dontKnow (Prov9LHCPsCvDontKnow inP p9c sol) inP)--return (success (Prov9LHCPsCvProcInfo inP p9c sol) inP)--
                        (_, Just sol)     -> return (dontKnow (Prov9LHCPsCvDontKnow inP p9c sol) inP)--return (success (Prov9LHCPsCvProcInfo inP p9c sol) inP)--
                  }
   | otherwise = dontKnow (Prov9LHCPsCvProcInfo inP p9c "MAYBE") inP
    where trs = getConfProblem . getR $ inP
          cpis = getCPairsCPT . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          term = getTermCPT . getR $ inP
          p9c = buildProver9CriteriaCSR trs cp

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPTermProblem conftrs (CriticalPairInfo idF idV) (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Ord idF, Ord idV, Ord String
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPTRewriting conftrs)
  ) => Pretty (Prov9LHCPsCvProcInfo (Problem CPTRewriting conftrs) idF idV) where
  pPrint Prov9LHCPsCvProcInfo { inProblem = problem
                              , inP9Crit = crit
                              , inSol = sol
                              }
    = pPrint problem
      Doc.<> text "\nProver9 LHCP Convergence Checker Processor:\n"
      Doc.<> text "formulas(sos).\n"
      Doc.<+> text "%Reflexivity\n"
      Doc.<+> text "reds(x,x).\n\n"
      Doc.<+> text "%Transitivity\n"
      Doc.<+> text "(red(x,y) & reds(y,z)) -> reds(x,z).\n\n"
      Doc.<+> text "%Congruence\n"
      Doc.<+> (hsep . punctuate (text "\n") . map (\cg -> pPrint cg) $ (congruence crit))
      Doc.<+> text "%Replacement\n"
      Doc.<> replacs
      Doc.<> text "end_of_list.\n\n"
      Doc.<> text "formulas(goals).\n"
      Doc.<+> goals
      Doc.<> text "end_of_list.\n\n"
      Doc.<> text "Proof:\n"
        Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem
          replacs = (hcat . map (\rl ->
                        (hcat . map (\v -> text " all" Doc.<+> pPrint v) $ (elems $ ruleVars rl))
                        Doc.<+> lparen
                          Doc.<+> text "red" Doc.<> lparen Doc.<> pPrint (lhs rl) Doc.<> comma 
                          Doc.<> pPrint (rhs rl) Doc.<> rparen 
                        Doc.<+> rparen
                      Doc.<> text ".\n"
                    ) $ (fromJust $ rulesNotSVars crit))
          Just (x,y) = csrCondNotSVars crit
          goals = text "all" Doc.<+> pPrint x Doc.<+> text "all" Doc.<+> pPrint y
                  Doc.<+> text "exists" Doc.<+> text (head $ ccpJoinVarNames crit)
                  Doc.<+> lparen
                  Doc.<+> text "red"
                    Doc.<> lparen
                      Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y
                    Doc.<> rparen
                  Doc.<+> text "->"
                  Doc.<+> lparen
                  Doc.<> text "reds"
                    Doc.<> lparen 
                      Doc.<> pPrint (fst $ cpConsNotSVars crit) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
                    Doc.<> rparen Doc.<+> char '&'
                  Doc.<+> text "reds" 
                    Doc.<> lparen 
                      Doc.<> pPrint (snd $ cpConsNotSVars crit) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
                    Doc.<> rparen
                  Doc.<> rparen
                  Doc.<+> rparen Doc.<> text ".\n"

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Prover9 CCP Convergence Checker Processor
prov9LHCPCvProc :: ( Processor info Prov9LHCPsConvProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
prov9LHCPCvProc to = apply Prov9LHCPsConvProcessor {timeout = to}


-- | Generate Prover9 input and try to get a solution
solveProver9 :: (HasSignature trs idF idV, HasRules trs (Term idF idV), Pretty idF, Pretty idV, Pretty trs, Ord String) => Maybe Int -> trs -> Prover9Criteria idF idV -> IO (String, Maybe String)
solveProver9 timeout trs crit
  = callToProver9 timeout prov9Problem
  where
    replacs = (hcat . map (\rl ->
                  (hcat . map (\v -> text " all" Doc.<+> pPrint v) $ (elems $ ruleVars rl))
                  Doc.<+> lparen
                    Doc.<+> text "red" Doc.<> lparen Doc.<> pPrint (lhs rl) Doc.<> comma 
                    Doc.<> pPrint (rhs rl) Doc.<> rparen 
                  Doc.<+> rparen
                Doc.<> text ".\n"
              ) $ (fromJust $ rulesNotSVars crit))
    Just (x,y) = csrCondNotSVars crit
    goals = text "all" Doc.<+> pPrint x Doc.<+> text "all" Doc.<+> pPrint y
            Doc.<+> text "exists" Doc.<+> text (head $ ccpJoinVarNames crit)
            Doc.<+> lparen
            Doc.<+> text "red"
              Doc.<> lparen
                Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y
              Doc.<> rparen
            Doc.<+> text "->"
            Doc.<+> lparen
            Doc.<> text "reds"
              Doc.<> lparen 
                Doc.<> pPrint (fst $ cpConsNotSVars crit) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
              Doc.<> rparen Doc.<+> char '&'
            Doc.<+> text "reds" 
              Doc.<> lparen 
                Doc.<> pPrint (snd $ cpConsNotSVars crit) Doc.<> comma Doc.<> text (head $ ccpJoinVarNames crit) 
              Doc.<> rparen 
            Doc.<> rparen
            Doc.<+> rparen Doc.<> text ".\n"
    prov9Problem = bconcat $ [ frmStr "formulas(sos).\n"
                             , frmStr "%Reflexivity\n"
                             , frmStr "reds(x,x).\n\n"
                             , frmStr "%Transitivity\n"
                             , frmStr "(red(x,y) & reds(y,z)) -> reds(x,z).\n\n"
                             , frmStr "%Congruence\n"
                             , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\cg -> pPrint cg) $ (congruence crit)
                             , frmStr "\n%Replacement\n"
                             , frmStr . noRiskyChars $ show replacs
                             , frmStr "\nend_of_list.\n\n"
                             , frmStr "formulas(goals).\n"
                             , frmStr . noRiskyChars $ show goals
                             , frmStr "end_of_list.\n\n"
                             , frmStr "%assign(sos_limit, -1).\n"
                             , frmStr "assign(max_megs, -1).\n"
                             , frmStr "set(quiet).\n"
                           ]
