{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.SCRInfProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check strong joinable
-- critical pairs with infChecker
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.SCRInfProcessor (

-- * Exported functions

scrIProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPLinProblem (..))
import Framework.Problem.Types (InConfCSTRS (..),  InCPLinCSTRS (..))
import MuTerm.Framework.Problem (Problem (..), MkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Proof, Info (..), isSuccess, refuted)
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPLRewriting (CPLRewriting (..))
import Framework.ExternalCalls (callToInfChecker)
import Framework.ExtCallsCriteria (InfCheckerCriteria (..), buildInfCheckerCriteria)
import Framework.TRSAux (isCSTRSSymb, isLHRV)
import Utils.Auxiliar (bconcat, frmStr, isTrue, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, rparen, hsep, lparen)
import Data.Set as S (Set, elems)
import Control.Monad.State (MonadState (..), runState)
import Control.Monad (MonadPlus, when)
import System.IO.Unsafe (unsafePerformIO)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust)


-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data SCRInfProcessor = SCRInfProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data SCRIProcInfo problem = 
  SCRIProcInfo { inProblem :: problem 
               , inICCrit :: InfCheckerCriteria
               , inSol    :: String
               }
  | SCRIProcDontKnow { inProblem :: problem 
                     , inICCrit :: InfCheckerCriteria
                     , inSol    :: String  
                     }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs ~ InCPLinCSTRS trs (CriticalPairInfo idF idV) Bool
  , IsConfProblem conftrs trs
  , IsCPLinProblem conftrs (CriticalPairInfo idF idV) Bool
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (SCRIProcInfo (Problem CPLRewriting conftrs))
  ) => Processor info SCRInfProcessor (Problem CPLRewriting conftrs) (Problem CPLRewriting conftrs) where
  apply SCRInfProcessor {timeout = to} inP
    | scr = unsafePerformIO $!
            do { 
                solution <- solveInfChecker to icc
                ; case solution of
                    ("NO", Just sol)  -> return (success (SCRIProcInfo inP icc sol) inP)
                    (_, Just sol)     -> return (dontKnow (SCRIProcDontKnow inP icc sol) inP)
               }
    | otherwise = dontKnow (SCRIProcDontKnow inP icc "MAYBE") inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          symbs =  elems . getTRSSymbols $ trs
          cstrs = or $ map isCSTRSSymb symbs
          lhrv = and $ map isLHRV rules
          cpis = getCPairsCPL . getR $ inP
          cpi = head cpis
          lin = getLinCPL . getR $ inP
          cp = critPair cpi
          icc = buildInfCheckerCriteria trs cp
          scr = (cstrs && lhrv && lin) || ((not cstrs) && lin)

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , IsCPLinProblem conftrs cps Bool
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPLRewriting conftrs)
  ) => Pretty (SCRIProcInfo (Problem CPLRewriting conftrs)) where
  pPrint SCRIProcInfo { inProblem = problem
                       , inICCrit = crit
                       , inSol = sol
                       }
    = pPrint problem
    Doc.<> text "\nStrong Confluence InfChecker Processor:\n"
    Doc.<> text "Infeasible convergence?\n"
      Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Strong Confluence Processor
scrIProc :: ( Processor info SCRInfProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
scrIProc to = apply SCRInfProcessor {timeout = to}


-- | Generate infChecker input and try to get a solution
solveInfChecker :: Maybe Int -> InfCheckerCriteria -> IO (String, Maybe String)
solveInfChecker timeout crit
  = callToInfChecker timeout infProblem
  where
    infProblem = bconcat $ [ frmStr "(PROBLEM INFEASIBILITY)\n"
                           , frmStr "(VAR "
                           , frmStr . noRiskyChars $ show . hcat . punctuate (text " ") . map (\v -> pPrint v) $ (trsVars crit)
                           , frmStr ")\n"
                           , frmStr "(RULES\n"
                           , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (rulesICC crit)
                           , frmStr "\n)\n"
                           , frmStr "(VAR z)\n"
                           , frmStr "(CONDITION "
                           , frmStr . noRiskyChars $ show $ pPrint (fst $ cpsConst crit) Doc.<> text " ->= z, " Doc.<> pPrint (snd $ cpsConst crit) Doc.<> text " == z"
                           , frmStr ", "
                           , frmStr . noRiskyChars $ show $ pPrint (snd $ cpsConst crit) Doc.<> text " ->= z, " Doc.<> pPrint (fst $ cpsConst crit) Doc.<> text " == z"
                           , frmStr ")"]
