{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE DataKinds #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.HuetBFJProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module applies the processor that tries to check strong joinable
-- critical pairs
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.SCRProcessor (

-- * Exported functions

scrProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), HasArity(..), HasMu (..))
import Framework.Problems (IsConfProblem (..), CriticalPairInfo(..), CriticalPair (..), IsCPLinProblem (..))
import Framework.Problem.Types (InCPLinCSTRS (..))
import MuTerm.Framework.Problem (Problem (..))
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, Info (..), Proof)
import TRS.Rules(Rule(..), freshRule)
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.CPLRewriting (CPLRewriting (..))
import Framework.TRSAux (RuleId (..), JoinabilityInfo (..), RewTerms (..), maxDepthReached, createRewritingGraph, haveJoinTerm, getRulesIds, isCSTRSSymb, 
  isLHRV, reflexiveClosureGraph)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Control.Monad (MonadPlus)
import Control.Monad.State (MonadState (..), runState)
import Data.Set as S (Set, elems, null, findMax)
import Data.Maybe (isJust, fromJust, isNothing)

import Data.Graph.Inductive.Graph as G
import Data.Graph.Inductive.PatriciaTree (Gr (..))

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Brute Force Joinability Processor
data SCRProcessor = SCRProcessor  { maxDepth :: Maybe (Int), maxCPOut :: Maybe (Int) }

-- | The information obtained are the rules
data SCRProcInfo problem idF idV = 
  SCRProcInfo { inProblem   :: problem
              , inCpi       :: CriticalPairInfo idF idV
              , inJoinInfo1 :: JoinabilityInfo idF idV
              , inJoinInfo2 :: JoinabilityInfo idF idV
              , inSGraph    :: Gr (RewTerms idF idV) () 
              , inTGraph    :: Gr (RewTerms idF idV) () 
              }
  | SCRDontKnow { inProblem   :: problem
                , inCpi       :: CriticalPairInfo idF idV
                , inJoinInfo1 :: JoinabilityInfo idF idV
                , inJoinInfo2 :: JoinabilityInfo idF idV
                , inSGraph    :: Gr (RewTerms idF idV) () 
                , inTGraph    :: Gr (RewTerms idF idV) () 
                }


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance (
    conftrs ~ InCPLinCSTRS trs (CriticalPairInfo idF idV) Bool
  , IsConfProblem conftrs trs
  , IsCPLinProblem conftrs (CriticalPairInfo idF idV) Bool
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasMu idF
  , HasArity idF
  , HasName idF
  , HasName idV
  , Eq idF, Eq idV
  , Ord idV
  , Info info (SCRProcInfo (Problem CPLRewriting conftrs) idF idV)
  ) => Processor info SCRProcessor (Problem CPLRewriting conftrs) (Problem CPLRewriting conftrs) where
  apply SCRProcessor {maxDepth = md, maxCPOut = mcp} inP
    = if scr
      then success (SCRProcInfo inP cpi ji1 ji2 sg tg) inP
      else dontKnow (SCRProcInfo inP cpi ji1 ji2 sg tg) inP
    where trs = getConfProblem . getR $ inP
          rules = elems . getTRSRules $ trs
          symbs =  elems . getTRSSymbols $ trs
          cpis = getCPairsCPL . getR $ inP
          cpi = head cpis
          cp = critPair cpi
          ncp = cpNum cpi
          cstrs = or $ map isCSTRSSymb symbs
          lhrv = and $ map isLHRV rules
          variables = getTRSVariables trs
          maxid = if S.null variables then 0 else ((+1) . getId . findMax $ variables)
          (freshRules,newMaxId) = runState (sequence . map freshRule $ rules) maxid
          (ji1, ji2, sg, tg) = isStronglyClosed (getRulesIds freshRules) cp (md,mcp,ncp)
          lin = getLinCPL . getR $ inP
          scrj = lin && (joinable ji1) && (joinable ji2)
          scr = (cstrs && lhrv && scrj) || ((not cstrs) && scrj)

          
-- Pretty

instance (
    IsConfProblem conftrs trs
  , IsCPLinProblem conftrs (CriticalPairInfo idF idV) Bool
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem CPLRewriting conftrs)
  ) => Pretty (SCRProcInfo (Problem CPLRewriting conftrs) idF idV) where
  pPrint SCRProcInfo { inProblem = problem
                     , inCpi = cpi
                     , inJoinInfo1 = j1
                     , inJoinInfo2 = j2
                     , inSGraph = sg
                     , inTGraph = tg
                     }
    = pPrint problem
    Doc.<> text "\nStrong Confluence Processor:\n"
    Doc.<>text "-> Rewritings:\n"
      Doc.<> text "s: " Doc.<> pPrint (s cp) Doc.<> text "\n"
      Doc.<> text "Nodes: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes sg)) Doc.<> text "\n" 
      Doc.<> text "Edges: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges sg)) Doc.<> text "\n"
        Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes sg)) Doc.<> text "\n"
      Doc.<> text "t: " Doc.<> pPrint (t cp) Doc.<> text "\n" 
      Doc.<> text "Nodes: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y) -> pPrint x ) $ (labNodes tg)) Doc.<> text "\n" 
      Doc.<> text "Edges: "
        Doc.<> (brackets . hcat . punctuate comma . map (\(x,y,z) -> lparen Doc.<> pPrint x Doc.<> comma Doc.<> pPrint y Doc.<> rparen) $ (labEdges tg)) Doc.<> text "\n"
        Doc.<> (hcat . punctuate (char '\n') . map (\(x,y) -> pPrint y) $ (labNodes tg)) Doc.<> text "\n"
      Doc.<> text "SNodesPath1: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (sNodesPath j1)) Doc.<> text "\n"
      Doc.<> text "TNodesPath1: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (tNodesPath j1)) Doc.<> text "\n"
      Doc.<> text "SNodesPath2: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (sNodesPath j2)) Doc.<> text "\n"
      Doc.<> text "TNodesPath2: "
        Doc.<> (hsep . punctuate (text " ->") . map (\(t) -> pPrint t) $ (tNodesPath j2)) Doc.<> text "\n"
      
      Doc.<> pPrint (s cp) Doc.<> text " ->= "
        Doc.<> (if isNothing (union j1) then text "no union" else pPrint (fromJust $ union j1))
        Doc.<> text " *<- " Doc.<> pPrint (t cp) Doc.<> text "\n"
      Doc.<> pPrint (t cp) Doc.<> text " ->= "
        Doc.<> (if isNothing (union j2) then text "no union" else pPrint (fromJust $ union j2))
        Doc.<> text " *<- " Doc.<> pPrint (s cp) Doc.<> text "\n"
      Doc.<> pPrint (if (joinable j1) && (joinable j2) then "Strongly closed critical pair" else "Not strongly closed critical pair")

    where trs = getConfProblem . getR $ problem
          cp = critPair cpi

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Strong Confluence Processor
scrProc :: ( Processor info SCRProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe(Int) -> Maybe(Int) -> o -> Proof info mp d
scrProc md mcp = apply SCRProcessor {maxDepth = md, maxCPOut = mcp}


isStronglyClosed :: (HasMu idF, Eq idF, Ord idV) => [RuleId idF idV] -> CriticalPair idF idV -> (Maybe(Int),Maybe(Int),Int)
  -> (JoinabilityInfo idF idV, JoinabilityInfo idF idV, Gr (RewTerms idF idV) (), Gr (RewTerms idF idV) ())
isStronglyClosed rs cp (md,mcp,ncp) = areStronglyJoinable (s cp) (t cp) rs (md,mcp,ncp)


areStronglyJoinable :: (HasMu idF, Eq idF, Ord idV) => Term idF idV -> Term idF idV -> [RuleId idF idV] -> (Maybe(Int),Maybe(Int),Int) 
  -> (JoinabilityInfo idF idV, JoinabilityInfo idF idV, Gr (RewTerms idF idV) (), Gr (RewTerms idF idV) ())
areStronglyJoinable s t rs (md,(Just mcp),ncp) 
  | ncp >= mcp = ((JoinabilityInfo j1 u1 Nothing [] []), (JoinabilityInfo j2 u2 Nothing [] []), G.empty, G.empty)
  | otherwise = ((JoinabilityInfo j1 u1 Nothing sNodesPath1 tNodesPath1), (JoinabilityInfo j2 u2 Nothing sNodesPath2 tNodesPath2), sGraph, tGraph)
  where sGraph = createRewritingGraph s rs md
        tGraph = createRewritingGraph t rs md
        sRewTerms = map snd (labNodes sGraph)
        tRewTerms = map snd (labNodes tGraph)
        sRCGr = reflexiveClosureGraph sGraph
        tRCGr = reflexiveClosureGraph tGraph
        ((j1, u1, sNodesPath1, tNodesPath1),(j2, u2, tNodesPath2, sNodesPath2)) = (haveJoinTerm sRCGr tGraph, haveJoinTerm tRCGr sGraph)
