{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.TerminationProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that tries to check termination calling to
-- muterm
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.TerminationProcessor (

-- * Exported functions

termProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..), HasArity (..))
import Framework.Problem.Types(InConfCSTRS (..), InTermCSTRS (..))
import Framework.Problems(IsConfProblem (..), IsTProblem (..), getPSchemeFromProblem)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, andP, Proof, Info (..), isSuccess, refuted, runProof)
import TRS.Rules(Rule(..))
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.CPRewriting (CPRewriting (..))
import Framework.Problem.TRewriting (TRewriting (..))
import Framework.Problem.CPTRewriting (CPTRewriting (..))
import Framework.ExternalCalls (callToMuTerm)
import Framework.ExtCallsCriteria (sVarsToVars, ruleSVarsToVars)
import Framework.TRSAux (getRepMap)
import Utils.Auxiliar (bconcat, frmStr, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, hsep, lparen, rparen)
import Control.Monad (MonadPlus, when)
import Data.Set as S (Set, elems, null, findMax)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)
import System.IO.Unsafe (unsafePerformIO)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data TerminationProcessor = TerminationProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data TermProcInfo problem = 
  TermProcInfo { inProblem :: problem
               , inSol     :: String
               }
  | TermDontKnow { inProblem :: problem
                 , inSol     :: String
                 } 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InConfCSTRS trs
  , conftrs2 ~ InTermCSTRS trs (Maybe(Bool))
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsTProblem conftrs2 (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasArity idF
  , HasName idF
  , HasName idV
  , HasMu idF
  , HasArity idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (TermProcInfo (Problem Rewriting conftrs1))
  ) => Processor info TerminationProcessor (Problem Rewriting conftrs1) (Problem TRewriting conftrs2) where
  apply TerminationProcessor {timeout = to} inP
    = unsafePerformIO $!
      do { 
          solution <- solveMuTerm to trs
         ; case solution of
              ("YES", Just sol) -> return (andP (TermProcInfo inP sol) inP [mkProblem (TRewriting (getPSchemeFromProblem inP)) (InTermCSTRS trs (Just True))])
              ("NO", Just sol)  -> return (andP (TermProcInfo inP sol) inP [mkProblem (TRewriting (getPSchemeFromProblem inP)) (InTermCSTRS trs (Just False))])
              (_, Just sol)     -> return (andP (TermProcInfo inP sol) inP [mkProblem (TRewriting (getPSchemeFromProblem inP)) (InTermCSTRS trs Nothing)])
         }
    where trs = getConfProblem . getR $ inP

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idV
  , HasMu idF
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (TermProcInfo (Problem Rewriting conftrs)) where
  pPrint TermProcInfo { inProblem = problem
                      , inSol = sol
                      }
    = text "\nTermination Processor:\n"
      Doc.<> text "(STRATEGY CONTEXTSENSITIVE\n"
      Doc.<> (hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map pPrint $ mu) Doc.<> rparen) $ (M.toList replMap))
      Doc.<> text "\n)\n"
      Doc.<> text "Terminating?\n"
        Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem
          replMap = getRepMap trs


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Termination Processor
termProc :: ( Processor info TerminationProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
termProc to = apply TerminationProcessor {timeout = to}


-- | Generate muterm input and try to get a solution
solveMuTerm :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasName idF, HasName idV, HasMu idF, HasArity idF, Ord idF, Ord idV, Pretty idF, Pretty idV, Pretty trs) => Maybe Int -> trs -> IO (String, Maybe String)
solveMuTerm timeout trs
  = callToMuTerm timeout termProblem
  where
      replMap = getRepMap trs
      termProblem = bconcat $ [ frmStr "(VAR "
                              , frmStr . noRiskyChars $ show . hcat . punctuate (text " ") . map (\v -> pPrint $ sVarsToVars v) $ (elems . getTRSVariables $ trs)
                              , frmStr ")\n"
                              , frmStr "(STRATEGY CONTEXTSENSITIVE\n"
                              , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\(f,mu) -> lparen Doc.<> pPrint f Doc.<+> (hsep . map pPrint $ mu) Doc.<> rparen) $ (M.toList replMap)
                              , frmStr "\n)\n"
                              , frmStr "(RULES\n"
                              , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint $ ruleSVarsToVars r) $ (elems . getTRSRules $ trs)
                              , frmStr "\n)"
                              ]