{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
{-# LANGUAGE CPP #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Processor.Rewriting.WNormalizationFortProcessor
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module apply the processor that tries to check weakly-normalization calling to
-- fort
--
-----------------------------------------------------------------------------

module Framework.Processor.Rewriting.WNormalizationFortProcessor (

-- * Exported functions

wnFortProc

) where

import TRS.TRSs(HasSignature (..), HasRules (..), HasOSSignature(..))
import TRS.Terms(Term(..))
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasMu (..), HasArity (..))
import Framework.Problem.Types(InConfCSTRS (..), InNormCSTRS (..))
import Framework.Problems(IsConfProblem (..), IsNProblem (..), getPSchemeFromProblem)
import MuTerm.Framework.Problem (Problem (..), MkProblem, mkProblem)
import MuTerm.Framework.Processor (Processor (..))
import MuTerm.Framework.Proof (success, dontKnow, andP, Proof, Info (..), isSuccess, refuted, runProof)
import TRS.Rules(Rule(..))
import Framework.Problem.Rewriting (Rewriting (..))
import Framework.Problem.NRewriting (NRewriting (..))
import Framework.ExternalCalls (callToWeakNormFort)
import Framework.ExtCallsCriteria (sVarsToVars, ruleSVarsToVars)
import Framework.TRSAux (getRepMap)
import Utils.Auxiliar (bconcat, frmStr, noRiskyChars, translateFromRiskyChars)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), text, (<+>), (<>), punctuate, hcat, hsep, lparen, rparen)
import Control.Monad (MonadPlus, when)
import Data.Set as S (Set, elems, null, findMax)
import Control.Monad.State (MonadState (..), runState)
import Data.Map as M (Map, toList)
import Data.Maybe (isJust, fromJust, isNothing)
import System.IO.Unsafe (unsafePerformIO)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Huet Processor
data WNormalizationFortProcessor = WNormalizationFortProcessor {timeout :: Maybe (Int)}

-- | The information obtained are the rules
data WNFortProcInfo problem = 
  WNFortProcInfo { inProblem :: problem
                 , inSol     :: String
                 }
  | WNFortDontKnow { inProblem :: problem
                   , inSol     :: String
                   } 


-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- Processor

instance {-# OVERLAPPABLE #-} (
    conftrs1 ~ InConfCSTRS trs
  , conftrs2 ~ InNormCSTRS trs (Maybe(Bool))
  , IsConfProblem conftrs1 trs
  , IsConfProblem conftrs2 trs
  , IsNProblem conftrs2 (Maybe(Bool))
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , Pretty idF, Pretty idV, Pretty trs
  , HasName idF, HasName idV
  , HasMu idF
  , HasArity idF
  , Eq idF, Eq idV
  , Ord idF, Ord idV
  , Info info (WNFortProcInfo (Problem Rewriting conftrs1))
  ) => Processor info WNormalizationFortProcessor (Problem Rewriting conftrs1) (Problem NRewriting conftrs2) where
  apply WNormalizationFortProcessor {timeout = to} inP -- = success (WNFortProcInfo inP "sol") inP
    
    = unsafePerformIO $!
      do { 
          solution <- solveFORT to trs
         ; case solution of
              ("YES", Just sol) -> return (andP (WNFortProcInfo inP sol) inP [mkProblem (NRewriting (getPSchemeFromProblem inP)) (InNormCSTRS trs (Just True))])
              ("NO", Just sol)  -> return (andP (WNFortProcInfo inP sol) inP [mkProblem (NRewriting (getPSchemeFromProblem inP)) (InNormCSTRS trs (Just False))])
              (_, Just sol)     -> return (andP (WNFortProcInfo inP sol) inP [mkProblem (NRewriting (getPSchemeFromProblem inP)) (InNormCSTRS trs Nothing)])
         }

    where trs = getConfProblem . getR $ inP

          
-- Pretty

instance {-# OVERLAPPABLE #-} (
    IsConfProblem conftrs trs
  , HasSignature trs idF idV
  , HasRules trs (Term idF idV)
  , HasName idF, HasName idV
  , HasMu idF
  , HasArity idF
  , Ord idF, Ord idV
  , Pretty idF, Pretty idV, Pretty trs
  , Pretty (Problem Rewriting conftrs)
  ) => Pretty (WNFortProcInfo (Problem Rewriting conftrs)) where
  pPrint WNFortProcInfo { inProblem = problem
                        , inSol = sol
                        }
    = text "\nWeak Normalization FORT Processor:\n"
      Doc.<> text "(VAR "
      Doc.<> (hsep . map (\v -> pPrint v) $ (map sVarsToVars . elems $ getTRSVariables trs))
      Doc.<> text ")\n"
      Doc.<> text "(RULES\n"
      Doc.<> (hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (map ruleSVarsToVars . elems $ getTRSRules trs))
      Doc.<> text "\n)"
      Doc.<> text "\nNormalizating?\n"
        Doc.<> (hcat . map (\c -> pPrint c) $ (translateFromRiskyChars sol))
    where trs = getConfProblem . getR $ problem
          replMap = getRepMap trs


-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Applies the Weak Normalization FORT Processor
wnFortProc :: ( Processor info WNormalizationFortProcessor o d
            , MonadPlus mp
            , Info info o
            , Info info d) => Maybe (Int) -> o -> Proof info mp d
wnFortProc to = apply WNormalizationFortProcessor {timeout = to}


-- | Generate FORT input and try to get a solution
solveFORT :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasArity idF, HasName idF, HasName idV, HasMu idF, Ord idF, Ord idV, Pretty idF, Pretty idV, Pretty trs) => Maybe Int -> trs -> IO (String, Maybe String)
solveFORT timeout trs
  = callToWeakNormFort timeout normProblem
  where
      replMap = getRepMap trs
      normProblem = bconcat $ [ frmStr "(VAR "
                              , frmStr . noRiskyChars $ show . hsep . map (\v -> pPrint v) $ (map sVarsToVars . elems $ getTRSVariables trs)
                              , frmStr ")\n"
                              , frmStr "(RULES\n"
                              , frmStr . noRiskyChars $ show . hcat . punctuate (text "\n") . map (\r -> pPrint r) $ (map ruleSVarsToVars . elems $ getTRSRules trs)
                              , frmStr ")"
                              ]