{-# LANGUAGE RecordWildCards #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.Proofs
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
--
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable
--
-- This module contains the proof functor.
--
-----------------------------------------------------------------------------

module Framework.Proofs (

-- * Exported data

Solution (..)

-- * Exported function

, runProofSol, shead

) where

import MuTerm.Framework.Proof (Proof, ProofF (..))

import Control.Applicative (Applicative (..), Alternative (..))
import Control.Monad.Free (Free (..), foldFree)
import Control.Monad (MonadPlus (..), join, liftM)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------

-- | Solution is the result of the evaluation
data Solution a = YES a (Solution a) -- We need the second argument to
                                     -- allow backtracking
                | NO  a
                | MAYBE
                deriving (Show)

-----------------------------------------------------------------------------
-- Instances
-----------------------------------------------------------------------------

-- | Solution is a functor
instance Functor Solution where
    fmap f (YES x xs) = YES (f x) (fmap f xs)
    fmap f (NO x)     = NO  (f x)
    fmap _ MAYBE      = MAYBE

-- | Solution is Applicative (GHC 7.10)
instance Applicative Solution where
  pure x = YES x MAYBE {- move the definition of `return` from the `Monad` instance here -}
--  (<*>) = ap  {- defined in Control.Monad -}
  -- or alternatively:
  f1 <*> f2 = f1 >>= \v1 -> f2 >>= (pure . v1)
--  (*>) = {- move the definition of `>>` from the `Monad` instance here -}
  NO  a *> (YES x _) = NO x
  NO  a *> NO b      = NO b
--  NO  a *> MAYBE     = MAYBE

-- | Solution is Alternative (GHC 7.10)
instance Alternative Solution where
   empty = MAYBE {- move definition of `mzero` here -}
   (<|>) = splus {- move definition of `mplus` here -}

-- | Isomorphic to the Maybe monad (return == YES)
instance Monad Solution where
  return = pure {- definition moved to `Applicative(pure)` -}
  (YES x xs) >>= f = sconcat . fmap f $ (YES x xs)
  NO  a >>= f = do { x <- f a; NO x } -- f a >>= \x -> NO x
  MAYBE >>= _ = MAYBE
  (>>) = (*>) {- definition moved to `Applicative((*>))` -}

-- | The MonadPlus instance elegantly enforces the priority of NO over YES
instance MonadPlus Solution
--
{-- | Isomorphic to the Maybe monad (return == YES)
instance Monad Solution where
  return x = YES x MAYBE
  (YES x xs) >>= f = sconcat . fmap f $ (YES x xs)
  NO  a >>= f = do { x <- f a; NO x } -- f a >>= \x -> NO x
  MAYBE >>= _ = MAYBE
  NO  a >> (YES x _) = NO x
  NO  a >> NO b      = NO b
-- NO  a >> MAYBE     = MAYBE

-- | The MonadPlus instance elegantly enforces the priority of NO over YES
instance MonadPlus Solution where
  mzero = MAYBE
  mplus = splus
--}

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | head for 'Solution'
shead (YES x _) = x
shead (NO x)    = x

-- | append for 'Solution'
splus :: Solution t -> Solution t -> Solution t
(NO x)     `splus` _          = NO x
_          `splus` (NO x)     = NO x
(YES x xs) `splus` (YES y ys) = YES x (xs `splus` (YES y ys))
MAYBE      `splus` x          = x
x          `splus` MAYBE      = x

-- | concat for 'Solution'
sconcat :: Solution (Solution t) -> Solution t
sconcat (NO (NO x))         = NO x
-- sconcat (NO (YES x _))      = NO x  -- is it possible?
-- sconcat (NO MAYBE)          = MAYBE -- is it possible?
sconcat (YES (YES y ys) xs) = YES y ((sconcat xs) `splus` ys)
sconcat (YES (NO x) _)      = NO x -- NO is always prioritary
sconcat (YES MAYBE xs)      = sconcat xs -- This is an 'or'
sconcat MAYBE               = MAYBE

isNO (NO _) = True
isNO _ = False

-- | Obtain the solution, collecting the proof path in the way
evalSolF :: ProofF info Solution (Solution (Proof info t ())) -> Solution (Proof info t ())
evalSolF Refuted{..}    = NO (Impure Refuted{..})
evalSolF DontKnow{}     = mzero
evalSolF MDone          = return (Impure MDone)
evalSolF Success{..}    = return (Impure Success{..})
evalSolF (Search mk)    = join mk
evalSolF (And pi pb []) = return (Impure $ Success pi pb)
evalSolF (And pi pb ll) = case negSol of
                            (l:_) -> (Impure . Single pi pb) `liftM` l
                            _ -> (Impure . And pi pb) `liftM` sequence ll
       where negSol = filter isNO ll
evalSolF (Or  pi pb ll) = (Impure . Single pi pb) `liftM` join ll
evalSolF (MAnd  p1 p2)  = p1 >>= \s1 -> p2 >>= \s2 ->
                           return (Impure $ MAnd s1 s2)
evalSolF (Single pi pb p) = (Impure . Single pi pb) `liftM` p

-- | Apply the evaluation returning the relevant proof subtree
runProofSol :: Proof info Solution a -> Solution (Proof info m ())
runProofSol = foldFree (const mzero) evalSolF
