{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE NoMonomorphismRestriction #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Framework.TRSAux
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module gives functions and data shared between several processors
--
-----------------------------------------------------------------------------

module Framework.TRSAux (

-- * Exported data

PosSubt (..), RuleId (..), UnificationsCPBuilder (..), RuleLhsPosSubterms (..), UVar (..),

-- * Joinability Exported data

JoinabilityInfo (..), RewTerms (..),

-- * Exported functions

getRulesIds, getCRulesIds, posMuSubtermsRule, isLHRV, isCSTRSSymb, uVarsCreation, maybeConvergentCPs, cpsInfo,
criticalPairUnification, criticalPairUnificationNW, quicksortCPByNumOfVars, cleanUnifications, createCPs, checkUnify, 
getVarsRules, getVarsCRules, pairVarsCreation, posMuSubterm, isTRSLinear, cleanRepeatedCCPs, activeVars, frozenVars,
quicksortRuleIDByNumOfVars, getMRCanonRepMap, setRepMap, getBotRepMap, getTopRepMap, getRepMap,
getLRCanoneRepMaps, getDefinedSymbs, getConstructorSymbs, applySigmaLCond, activeVarsPos, frozenVarsPos,
activeVarsPosByVars, quicksortCPByNumOfNormalTerms, getLHCPPairVarYFromCPIs, criticalPairsInfoVars,
maxNumOfConds,

-- * Joinability Exported functions
createRewritingGraph, maxDepthReached, haveJoinTerm, isNormalTerm, groundNormalTerms, isGroundTerm, normalWeight,
reflexiveClosureGraph

) where

import TRS.TRSs(HasSignature (..), HasCRules (..), HasRules (..), HasOSSignature(..), isLeftLinear)
import TRS.Terms(Term(..), subTerm, mgu, applySigma, replace, isLinear, vars, isVTerm, matchSubstitution)
import TRS.Identifiers (HasName (..), IdFTRS(..), IdVTRS(..), HasFSort(..), HasSort(..), Sort(..), HasArity (..), HasMu (..))
import TRS.Rules as R (Rule(..), CRule(..), Cond (..), lhs, rhs, freshCRule, cRuleVars, conds, lcond, rcond, condVars, deterministic, 
                       orientedConds, ruleVars, rule)
import Framework.Problems(CriticalPair (..), CriticalPairInfo (..))
import Utils.Auxiliar (subsequences, doubArrayComb, shuffle)

import Text.PrettyPrint.HughesPJClass as Doc (Pretty (..), Doc, text, char, (<+>), (<>), punctuate, hcat, parens, 
  brackets, comma, hsep, nest, braces, lparen, rparen, fcat, ($$), doubleQuotes, quotes)
import Data.Set as S (Set, elems, null, findMax, map, intersection, singleton, union, unions, toList, empty, fromList, (\\), size, 
                      delete, elemAt, insert, member)
import Data.Map as M (Map, toList, fromList, lookup, alter, insert, mapWithKey, empty)
import Data.Maybe (isJust, fromJust, isNothing)
import System.IO.Unsafe (unsafePerformIO)

import Data.Graph.Inductive.Graph as G
import Data.Graph.Inductive.PatriciaTree
import Algorithm.Search (dfs)

-----------------------------------------------------------------------------
-- Data
-----------------------------------------------------------------------------


data PosSubt idF idV =
  PosSubt { pos     :: [Int]
          , subterm :: Term idF idV
          }


data RuleId idF idV =
    RuleId { trule   :: Rule (Term idF idV)
           , tidr    :: Int
           }
 |  CRuleId { crule  :: CRule (Term idF idV)
            , cidr    :: Int
            }
    deriving (Eq, Ord)


data RuleLhsPosSubterms idF idV =
  RuleLhsPosSubterms { rid    :: RuleId idF idV
                     , pSubts :: [PosSubt idF idV]
                     }


data UnificationsCPBuilder idF idV =
  UnificationsCPBuilder { l     :: Term idF idV
                        , l1    :: Term idF idV
                        , r     :: Term idF idV
                        , r1    :: Term idF idV
                        , p     :: [Int]
                        , lp    :: Term idF idV
                        , css   :: [Cond (Term idF idV)]
                        , id1   :: Int
                        , id2   :: Int
                        , sigma :: Maybe (Map idV (Term idF idV))
                        }


data UVar idV =
  UVar { uvRuleId  :: Int
       , uvLActive :: Set idV
       , uvRActive :: Set idV
       , uvLFrozen :: Set idV
       , uvRFrozen :: Set idV
       }

data PreCombinationRM idF = 
  PreCombinationRM { fSymb      :: idF
                   , canonCombs :: [[Int]]
                   }


-----------------------------------------------------------------------------
-- JOINABILITY Data
-----------------------------------------------------------------------------

data JoinabilityInfo idF idV =
  JoinabilityInfo { joinable    :: Bool 
                  , union       :: Maybe (Term idF idV)
                  , maxDReached :: Maybe (Bool)
                  , sNodesPath  :: [Term idF idV]
                  , tNodesPath  :: [Term idF idV]
                  }

data RewTerms idF idV =
  RewTerms { term      :: Term idF idV
           , rtid      :: Int
           , depth     :: Int
           , rtrid     :: Int
           , rtsigma   :: Maybe (Map idV (Term idF idV))
           , posic     :: [Int]
           , newr      :: Maybe(Term idF idV)
           }


-- JOINABILITY Pretty

instance (Pretty idF, Pretty idV) => Pretty (RewTerms idF idV) where
  pPrint (RewTerms t 0 _ _ _ _ _)
    = text "ID: 0 =>" Doc.<+> parens (quotes (pPrint t)
      Doc.<> comma Doc.<+> text "D" Doc.<> text "0")
  pPrint (RewTerms t id d r s p nr)
    = text "ID: " Doc.<> pPrint id Doc.<+> text "=>" Doc.<+> parens (quotes (pPrint t)
        Doc.<> comma Doc.<+> text "D" Doc.<> pPrint d
        Doc.<> comma Doc.<+> text "R" Doc.<> pPrint r
        Doc.<> comma Doc.<+> text "P" Doc.<> (brackets . hsep . punctuate comma . Prelude.map pPrint $ p)
        Doc.<> comma Doc.<+> text "S" Doc.<> (braces . hsep . punctuate comma . Prelude.map (\(k,v) -> pPrint k Doc.<> text " -> " Doc.<> pPrint v)
          $ (M.toList $ (if isNothing s then M.empty else fromJust s))))
        Doc.<> comma Doc.<+> text "NR: " Doc.<> (if isNothing nr then text "no newR" else quotes (pPrint $ fromJust nr))



-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

getRulesIds :: [Rule (Term idF idV)] -> [RuleId idF idV]
getRulesIds rs = [RuleId (rs!!(i - 1)) i | i <- [1..(length rs)]]

getCRulesIds :: [CRule (Term idF idV)] -> [RuleId idF idV]
getCRulesIds rs = [CRuleId (rs!!(i - 1)) i | i <- [1..(length rs)]]


getVarsRules :: Ord idV => [Rule (Term idF idV)] -> [idV]
getVarsRules [] = []
getVarsRules (r:rs) = (getVarsRule r) ++ (getVarsRules rs)

getVarsRule :: Ord idV => Rule (Term idF idV) -> [idV]
getVarsRule r = elems $ ruleVars r

getVarsCRules :: Ord idV => [CRule (Term idF idV)] -> [idV]
getVarsCRules [] = []
getVarsCRules (r:rs) = (getVarsCRule r) ++ (getVarsCRules rs)

getVarsCRule :: Ord idV => CRule (Term idF idV) -> [idV]
getVarsCRule r = elems $ cRuleVars r


maxNumOfConds :: Ord idV => [CRule (Term idF idV)] -> Int -> Int
maxNumOfConds [] m = m
maxNumOfConds (cr:crs) m
  | n > m = maxNumOfConds crs n
  | otherwise = maxNumOfConds crs m
  where n = length $ conds cr


checkUnify :: (Ord idV, Eq idF, Eq idV, HasName idV, Pretty idV) => [RuleLhsPosSubterms idF idV] -> [RuleId idF idV] -> Int -> [UnificationsCPBuilder idF idV]
checkUnify [] _ _ = []
checkUnify (rp:rps) rules c = (checkUnifyRulesSubterms (pSubts rp) rp rules c) ++ (checkUnify rps rules c)


checkUnifyRulesSubterms :: (Ord idV, Eq idF, Eq idV, HasName idV, Pretty idV) 
                        => [PosSubt idF idV] -> RuleLhsPosSubterms idF idV -> [RuleId idF idV] -> Int -> [UnificationsCPBuilder idF idV]
checkUnifyRulesSubterms _ _ [] _ = []
checkUnifyRulesSubterms [] _ _ _ = []
checkUnifyRulesSubterms (t:ts) rp rl c = (checkUnifyRulesSubterm t rp rl c) ++ (checkUnifyRulesSubterms ts rp rl c)


checkUnifyRulesSubterm :: (Ord idV, Eq idF, Eq idV, HasName idV, Pretty idV) 
                       => PosSubt idF idV -> RuleLhsPosSubterms idF idV -> [RuleId idF idV] -> Int -> [UnificationsCPBuilder idF idV]
checkUnifyRulesSubterm t rp [] _ = [] 
checkUnifyRulesSubterm t rp (rl:rls) c = (checkUnifyRuleSubterm t rp rl c) ++ (checkUnifyRulesSubterm t rp rls c)


checkUnifyRuleSubterm :: (Ord idV, Eq idF, Eq idV, HasName idV, Pretty idV) 
                      => PosSubt idF idV -> RuleLhsPosSubterms idF idV -> RuleId idF idV -> Int -> [UnificationsCPBuilder idF idV]
checkUnifyRuleSubterm t rp (CRuleId cr2 id2) c
                    | (id1 == id2) && (p == [])            = []
                    | isJust sigma                         = [UnificationsCPBuilder l l1 r r1 p lp cds id1 id2 sigma]
                    | otherwise                            = []
  where cr1   = crule $ rid rp
        ncr2 = prettyCRuleRenaming (S.fromList $ getVarsCRule cr1) cr2 c
        rl1   = R.rule cr1
        rl2   = R.rule ncr2
        l     = lhs rl1
        r     = rhs rl1
        r1    = rhs rl2
        l1    = lhs rl2
        id1   = cidr $ rid rp
        p     = pos t
        lp    = subterm t
        sigma = mgu lp l1
        cds   = (conds cr1)++(conds ncr2)
checkUnifyRuleSubterm t rp (RuleId rl2 id2) c
                    | (id1 == id2) && (p == [])            = []
                    | isJust sigma                         = [UnificationsCPBuilder l l1 r r1 p lp [] id1 id2 sigma]
                    | otherwise                            = []
  where rl1 = trule $ rid rp
        nrl2 = prettyRuleRenaming (S.fromList $ getVarsRule rl1) rl2 c
        l     = lhs rl1
        r     = rhs rl1
        r1    = rhs nrl2
        l1    = lhs nrl2
        id1   = tidr $ rid rp
        p     = pos t
        lp    = subterm t
        sigma = mgu lp l1


prettyCRuleRenaming :: (HasName idV, Ord idV, Pretty idV) => Set idV -> CRule (Term idF idV) -> Int -> CRule (Term idF idV)
prettyCRuleRenaming vs cr2 c = fmap (renVariables pvs) cr2
  where pvs = [(getPrettyNonRepeatedNameVar vs Nothing c v2, [v2]) | v2 <- getVarsCRule cr2]

prettyRuleRenaming :: (HasName idV, Ord idV, Pretty idV) => Set idV -> Rule (Term idF idV) -> Int -> Rule (Term idF idV)
prettyRuleRenaming vs r2 c = fmap (renVariables pvs) r2
  where pvs = [(getPrettyNonRepeatedNameVar vs Nothing c v2, [v2]) | v2 <- getVarsRule r2]


getPrettyNonRepeatedNameVar :: (HasName idV) => Set idV -> Maybe Int -> Int -> idV -> idV
getPrettyNonRepeatedNameVar vs Nothing c v
    | member name names = getPrettyNonRepeatedNameVar vs (Just 0) c v
    | otherwise = (setName (Just name) . setId (id + c)) v
    where names = S.map getName vs
          name = getName v
          id = getId v
getPrettyNonRepeatedNameVar vs (Just 0) c v
    | member newN names = getPrettyNonRepeatedNameVar vs (Just 1) c v
    | otherwise = (setName (Just newN) . setId (id + c)) v
    where names = S.map getName vs
          newN = (getName v) ++ "'"
          id = getId v
getPrettyNonRepeatedNameVar vs (Just n) c v
    | member newN names = getPrettyNonRepeatedNameVar vs (Just (n + 1)) c v
    | otherwise = (setName (Just newN) . setId (id + c)) v
    where names = S.map getName vs
          newN = (getName v) ++ ((show n) ++  "'")
          id = getId v


cleanUnifications :: [UnificationsCPBuilder idF idV] -> [UnificationsCPBuilder idF idV]
cleanUnifications [] = []
cleanUnifications (u:us) = cleanUnification u us ++ cleanUnifications us


cleanUnification :: UnificationsCPBuilder idF idV -> [UnificationsCPBuilder idF idV] -> [UnificationsCPBuilder idF idV]
cleanUnification uc [] = [uc]
cleanUnification uc (u:us)
               | isNotUniqueUnification uc u = cleanUnification uc us
               | otherwise                   = []


isNotUniqueUnification :: UnificationsCPBuilder idF idV -> UnificationsCPBuilder idF idV -> Bool
isNotUniqueUnification uc uq = not ((id11 == id22 && id12 == id21) && (p uc) == [])
  where id11 = id1 uc
        id12 = id2 uc
        id21 = id1 uq
        id22 = id2 uq


criticalPairUnification :: (Ord idV, Eq idF, HasMu idF) => UnificationsCPBuilder idF idV -> CriticalPair idF idV
criticalPairUnification unif = CriticalPair s t cpPos ccpcds 0
  where sig    = fromJust (sigma unif)
        s      = replace (applySigma sig (r1 unif)) (p unif) (applySigma sig (l unif))
        t      = applySigma sig (r unif)
        cpPos  = p unif
        ccpcds = cleanCDs $ (Prelude.map . fmap) (applySigma sig) (css unif)

criticalPairUnificationNW :: (Ord idV, Eq idF, HasMu idF) => UnificationsCPBuilder idF idV -> [RuleId idF idV] -> CriticalPair idF idV
criticalPairUnificationNW unif rIds = CriticalPair s t cpPos ccpcds nw
  where sig    = fromJust (sigma unif)
        s      = replace (applySigma sig (r1 unif)) (p unif) (applySigma sig (l unif))
        t      = applySigma sig (r unif)
        cpPos  = p unif
        ccpcds = cleanCDs $ (Prelude.map . fmap) (applySigma sig) (css unif)
        nw = ((normalWeight s rIds) + (normalWeight t rIds))

cleanCDs :: (Ord idV, Eq idF, Eq idV) => [Cond (Term idF idV)] -> [Cond (Term idF idV)]
cleanCDs [] = []
cleanCDs (cd:cds)
  | rep = cleanCDs cds
  | otherwise = cd:(cleanCDs cds)
  where rep = isCondRepeated cd cds


isCondRepeated :: (Ord idV, Eq idF, Eq idV) => Cond (Term idF idV) -> [Cond (Term idF idV)] -> Bool
isCondRepeated c1 [] = False
isCondRepeated c1 (c2:cds)
  | ((lcond c1) == (lcond c2)) && ((rcond c1) == (rcond c2)) = True
  | otherwise = isCondRepeated c1 cds



isRightLinear :: (HasRules trs (Term idF idV), Ord idV) => trs -> Bool
isRightLinear trs = and (Prelude.map isRuleRightLinear (elems . getTRSRules $ trs))


isRuleRightLinear :: Ord idV => Rule (Term idF idV) -> Bool
isRuleRightLinear (l :-> r) = isLinear r


isTRSLinear :: (HasRules trs (Term idF idV), Ord idV) => trs -> Bool
isTRSLinear trs = (isLeftLinear trs) && (isRightLinear trs)


isLHRV :: (HasMu idF, HasArity idF, Ord idV) => Rule (Term idF idV) -> Bool
isLHRV (l :-> r) = S.null inter
  where lActVars = activeVars l
        lFrozVars = frozenVars l
        rFrozVars = frozenVars r
        rlFrozVars = lFrozVars `S.union` rFrozVars
        inter = lActVars `intersection` rlFrozVars


activeVars :: (HasMu idF, Ord idV) => Term idF idV -> Set idV
activeVars t@(F f ts) = unions [activeVars (ts!!(i-1)) | i <- mu]
  where mu = getFMu f
activeVars t@(V v) = S.singleton v

frozenVars :: (HasMu idF, HasArity idF, Ord idV) => Term idF idV -> Set idV
frozenVars t@(F f ts)
  | Prelude.null diff = unions [frozenVars (ts!!(a-1)) | a <- [1..getFArity f]]
  | otherwise = (unions [frozenVars (ts!!(m-1)) | m <- (getFMu f)]) `S.union` (unions [vars (ts!!(d-1)) | d <- diff])
  where ar = S.fromList [1..getFArity f]
        mu = S.fromList $ getFMu f
        diff = S.toList (ar \\ mu)
frozenVars _ = S.empty


activeVarsPos :: (HasMu idF, Ord idV) => Term idF idV -> Set (idV, [Int])
activeVarsPos t = activeVarsPos' t []

frozenVarsPos :: (HasMu idF, HasArity idF, Ord idV) => Term idF idV-> Set (idV, [Int])
frozenVarsPos t = frozenVarsPos' t []

activeVarsPos' :: (HasMu idF, Ord idV) => Term idF idV -> [Int] -> Set (idV, [Int])
activeVarsPos' t@(F f ts) p = unions [activeVarsPos' (ts!!(i-1)) (p++[i]) | i <- mu]
  where mu = getFMu f
activeVarsPos' t@(V v) p = S.singleton (v,p)

frozenVarsPos' :: (HasMu idF, HasArity idF, Ord idV) => Term idF idV -> [Int] -> Set (idV, [Int])
frozenVarsPos' t@(F f ts) p
  | Prelude.null diff = unions [frozenVarsPos' (ts!!(a-1)) (p++[a]) | a <- [1..getFArity f]]
  | otherwise = (unions [frozenVarsPos' (ts!!(m-1)) (p++[m]) | m <- (getFMu f)]) `S.union` (unions [termVarsPos (ts!!(d-1)) (p++[d]) | d <- diff])
  where ar = S.fromList [1..getFArity f]
        mu = S.fromList $ getFMu f
        diff = S.toList (ar \\ mu)
frozenVarsPos' _ _ = S.empty

termVarsPos :: (HasArity idF, Ord idV) => Term idF idV -> [Int] -> Set (idV, [Int])
termVarsPos t@(F f ts) p = unions [termVarsPos (ts!!(a-1)) (p++[a]) | a <- [1..getFArity f]]
termVarsPos t@(V v) p = S.singleton (v,p)

activeVarsPosByVars :: (HasMu idF, Ord idV) => Term idF idV -> Set idV -> Set (idV, [Int])
activeVarsPosByVars t vs
  | S.null vs = S.empty
  | otherwise = varsPosByVars actVsPos (elems vs)
  where actVsPos = activeVarsPos t

varsPosByVars :: (Ord idV) => Set (idV, [Int]) -> [idV] -> Set (idV, [Int])
varsPosByVars avp [] = S.empty
varsPosByVars avp (v:vs) = S.union (varsPosVOccurrences avp v) (varsPosByVars avp vs)

varsPosVOccurrences :: (Ord idV) => Set (idV, [Int]) -> idV -> Set (idV, [Int])
varsPosVOccurrences vps v1 = S.unions [S.singleton (v2,p) | (v2,p) <- (elems vps), v1 == v2]


posMuSubtermsRule :: HasMu idF => RuleId idF idV -> RuleLhsPosSubterms idF idV
posMuSubtermsRule r@(CRuleId rl id) = RuleLhsPosSubterms {rid = r, pSubts = posMuSubterm [] (lhs $ rule rl)}
posMuSubtermsRule r@(RuleId rl id) = RuleLhsPosSubterms {rid = r, pSubts = posMuSubterm [] (lhs rl)}

posMuSubterm :: HasMu idF => [Int] -> Term idF idV -> [PosSubt idF idV]
posMuSubterm p t@(V v) = []
posMuSubterm p t@(F f ts) = insPosMuSubterm p s
  where s = subTerm t p

insPosMuSubterm :: HasMu idF => [Int] -> Term idF idV -> [PosSubt idF idV]
insPosMuSubterm p t@(V v) = []
insPosMuSubterm p t@(F f ts) = (PosSubt p t):(concat [insPosMuSubterm (p++[p']) (ts!!(p' - 1)) | p' <- mu])
  where mu = getFMu f


isCSTRSSymb :: (HasMu idF, HasArity idF) => idF -> Bool
isCSTRSSymb f = ar > (length mu)
  where mu = getFMu f
        ar = getFArity f


createCPs :: (HasName idV, Ord idV, Pretty idV, Eq idV) => [CriticalPair idF idV] -> [CriticalPair idF idV]
createCPs [] = []
createCPs (cp:cps) = (CriticalPair (s cp) (t cp) (cpPos cp) (ccpConds cp) (normWeight cp)):(createCPs cps)


criticalPairsInfoVars :: (HasName idV, Ord idV, Pretty idV, Eq idV) => [CriticalPairInfo idF idV] -> Set idV
criticalPairsInfoVars [] = S.empty
criticalPairsInfoVars (cpi:cpis) = ((svs `S.union` tvs) `S.union` cdsvs) `S.union` (criticalPairsInfoVars cpis)
  where cp = critPair cpi
        svs = vars $ s cp
        tvs = vars $ t cp
        cdsvs = S.unions $ Prelude.map condVars (ccpConds cp)


getLHCPPairVarYFromCPIs :: [CriticalPairInfo idF idV] -> Maybe (idV,[idV])
getLHCPPairVarYFromCPIs [] = Nothing
getLHCPPairVarYFromCPIs (cpi:_)
  | Prelude.null cds = Nothing
  | otherwise = Just (v,[v])
  where cds = ccpConds $ critPair cpi
        V v = rcond $ head cds


cleanRepeatedCCPs :: (HasName idV, Ord idV, Pretty idV, Eq idF, Eq idV) => [CriticalPair idF idV] -> [CriticalPair idF idV]
cleanRepeatedCCPs [] = []
cleanRepeatedCCPs (cp:cps)
  | (or $ Prelude.map (==(s cp)) ss) && (or $ Prelude.map (==(t cp)) ts) && (or $ Prelude.map (==(ccpConds cp)) cs) = cleanRepeatedCCPs cps
  | otherwise = cp:(cleanRepeatedCCPs cps)
    where ss = Prelude.map s cps
          ts = Prelude.map t cps
          cs = Prelude.map ccpConds cps


-- | We apply the substitution to the existing variables
renVariables :: (HasName idV, Ord idV, Pretty idV) => [(idV, [idV])] -> Term idF idV -> Term idF idV
renVariables sigma (F f tt) = F f (Prelude.map (renVariables sigma) tt)
renVariables sigma (V v)    = case maybePairedValue v sigma of
                                 Just v' -> V v'
                                 Nothing -> error ("Error (Terms): Variable " ++ (show . pPrint $ v) ++ " cannot be found.")


maybePairedValue :: (HasName idV, Eq idV) => idV -> [(idV, [idV])] -> Maybe idV
maybePairedValue v [] = Nothing
maybePairedValue v ((k,vs):pvs) = if isVarInPairedEntry v (k,vs) then Just k else maybePairedValue v pvs


isVarInPairedEntry :: (HasName idV, Eq idV) => idV -> (idV, [idV]) -> Bool
isVarInPairedEntry _ (_,[]) = False
isVarInPairedEntry a (k,v:vs) = if (getName a == getName v || getName a == getName k) then True else isVarInPairedEntry a (k,vs) 


quicksortCPByNumOfNormalTerms :: (HasName idV, Ord idV, Eq idV) => [CriticalPair idF idV] -> [CriticalPair idF idV]
quicksortCPByNumOfNormalTerms [] = []
quicksortCPByNumOfNormalTerms xs = quicksortCPByNumOfNormalTerms' $ shuffle xs

quicksortCPByNumOfNormalTerms' :: (HasName idV, Ord idV, Eq idV) => [CriticalPair idF idV] -> [CriticalPair idF idV]
quicksortCPByNumOfNormalTerms' [] = []
quicksortCPByNumOfNormalTerms' (x:xs) = 
     quicksortCPByNumOfNormalTerms' [y | y <- xs, (normWeight y) > normwsx] 
  ++ [x] 
  ++ quicksortCPByNumOfNormalTerms' [y | y <- xs, (normWeight y) == normwsx] 
  ++ quicksortCPByNumOfNormalTerms' [y | y <- xs, (normWeight y) < normwsx]
  where normwsx = normWeight x

quicksortCPByNumOfVars :: (HasName idV, Ord idV, Eq idV) => [CriticalPair idF idV] -> [CriticalPair idF idV]
quicksortCPByNumOfVars [] = []
quicksortCPByNumOfVars xs = quicksortCPByNumOfVars' $ shuffle xs

quicksortCPByNumOfVars' :: (HasName idV, Ord idV, Eq idV) => [CriticalPair idF idV] -> [CriticalPair idF idV]
quicksortCPByNumOfVars' [] = []
quicksortCPByNumOfVars' (x:xs) = 
     quicksortCPByNumOfVars' [y | y <- xs, (length $ vars (s y)) + (length $ vars (t y)) < nxvars] 
  ++ [x] 
  ++ quicksortCPByNumOfVars' [y | y <- xs, (length $ vars (s y)) + (length $ vars (t y)) == nxvars] 
  ++ quicksortCPByNumOfVars' [y | y <- xs, (length $ vars (s y)) + (length $ vars (t y)) > nxvars]
  where nxvars = (length $ vars (s x)) + (length $ vars (t x))

quicksortRuleIDByNumOfVars :: (Ord idV, Eq idV) => [RuleId idF idV] -> [RuleId idF idV]
quicksortRuleIDByNumOfVars [] = []
quicksortRuleIDByNumOfVars xs = quicksortRuleIDByNumOfVars' $ shuffle xs

quicksortRuleIDByNumOfVars' :: (Ord idV, Eq idV) => [RuleId idF idV] -> [RuleId idF idV]
quicksortRuleIDByNumOfVars' [] = []
quicksortRuleIDByNumOfVars' (x:xs) = 
     quicksortRuleIDByNumOfVars' [y | y <- xs, (S.size . ruleVars $ trule y) < nxvars]
  ++ [x] 
  ++ quicksortRuleIDByNumOfVars' [y | y <- xs, (S.size . ruleVars $ trule y) == nxvars] 
  ++ quicksortRuleIDByNumOfVars' [y | y <- xs, (S.size . ruleVars $ trule y) > nxvars]
  where nxvars = S.size . ruleVars $ trule x

quicksortRepMapByMuLenghtAsc :: (Ord idF, Eq idF) => [(idF, [Int])] -> [(idF, [Int])]
quicksortRepMapByMuLenghtAsc [] = []
quicksortRepMapByMuLenghtAsc xs = quicksortRepMapByMuLenghtAsc' $ shuffle xs

quicksortRepMapByMuLenghtAsc' :: (Ord idF, Eq idF) => [(idF, [Int])] -> [(idF, [Int])]
quicksortRepMapByMuLenghtAsc' [] = []
quicksortRepMapByMuLenghtAsc' (x:xs) = 
     quicksortRepMapByMuLenghtAsc' [y | y <- xs, (length $ snd y) < lMu]
  ++ [x]
  ++ quicksortRepMapByMuLenghtAsc' [y | y <- xs, (length $ snd y) == lMu] 
  ++ quicksortRepMapByMuLenghtAsc' [y | y <- xs, (length $ snd y) > lMu]
  where lMu = length $ snd x

quicksortRepMapByMuLenghtDesc :: (Ord idF, Eq idF) => [(idF, [Int])] -> [(idF, [Int])]
quicksortRepMapByMuLenghtDesc [] = []
quicksortRepMapByMuLenghtDesc xs = quicksortRepMapByMuLenghtDesc' $ shuffle xs

quicksortRepMapByMuLenghtDesc' :: (Ord idF, Eq idF) => [(idF, [Int])] -> [(idF, [Int])]
quicksortRepMapByMuLenghtDesc' [] = []
quicksortRepMapByMuLenghtDesc' (x:xs) = 
  quicksortRepMapByMuLenghtDesc' [y | y <- xs, (length $ snd y) > lMu]
  ++ [x]
  ++ quicksortRepMapByMuLenghtDesc' [y | y <- xs, (length $ snd y) == lMu] 
  ++ quicksortRepMapByMuLenghtDesc' [y | y <- xs, (length $ snd y) < lMu]
  where lMu = length $ snd x

quicksortLength :: (Ord a, Eq a) => [[a]] -> [[a]]
quicksortLength [] = []
quicksortLength xs = quicksortLength' $ shuffle xs

quicksortLength' :: (Ord a, Eq a) => [[a]] -> [[a]]
quicksortLength' [] = []
quicksortLength' (x:xs) = 
     quicksortLength' [y | y <- xs, length y < l]
  ++ [x]
  ++ quicksortLength' [y | y <- xs, length y == l] 
  ++ quicksortLength' [y | y <- xs, length y > l]
  where l = length x


cpsInfo :: (Ord idV, Eq idF) => [CriticalPair idF idV] -> Int -> [CriticalPairInfo idF idV]
cpsInfo [] i = []
cpsInfo (cp:cps) i = cpi:(cpsInfo cps (i+1))
  where cpi = CriticalPairInfo (isTrivial cp) (isOverlay cp) cp i Nothing

isTrivial :: (Ord idV, Eq idF) => CriticalPair idF idV -> Bool
isTrivial cp = (s cp) == (t cp)

isOverlay :: (Ord idV, Eq idF) => CriticalPair idF idV -> Bool
isOverlay cp = (cpPos cp) == []


maybeConvergentCPs :: (Ord idV, Eq idF) => [CriticalPairInfo idF idV] -> [CriticalPairInfo idF idV]
maybeConvergentCPs [] = []
maybeConvergentCPs (c:cpis) = (maybeConvergentCP c) ++ maybeConvergentCPs cpis


maybeConvergentCP :: (Ord idV, Eq idF) => CriticalPairInfo idF idV -> [CriticalPairInfo idF idV]
maybeConvergentCP cpi
                | trivial cpi = []
                | otherwise    = [cpi]


pairVarsCreation :: Ord idV => [idV] -> [idV] -> [(idV, [idV])]
pairVarsCreation [] _ = []
pairVarsCreation (v:vs) (f:fvs) = pairVars [(v,[f])] vs fvs

pairVars :: Ord idV => [(idV, [idV])] -> [idV] -> [idV] -> [(idV, [idV])]
pairVars p [] _ = p
pairVars p (v:vs) (f:fvs) = pairVars (pairVar p v f) vs fvs

pairVar :: Ord idV => [(idV, [idV])] -> idV -> idV -> [(idV, [idV])]
pairVar p v f = if isNothing $ Prelude.lookup v p then (p ++ [(v,[f])]) else insertValue p v f


uVarsCreation :: (HasMu idF, HasArity idF, Ord idV) => RuleId idF idV -> UVar idV
uVarsCreation (RuleId rl i) = UVar i (activeVars l) (activeVars r) (frozenVars l) (frozenVars r)
  where (l :-> r) = rl


insertValue :: Eq a => [(a, [a])] -> a -> a -> [(a, [a])]
insertValue [] _ _ = []
insertValue ((dk,dv):ds) k v = if dk == k then [(dk,dv++[v])] ++ ds else [(dk,dv)] ++ (insertValue ds k v)


-- Get Most Restrictive Replacement Map
getMRCanonRepMap :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasArity idF, HasMu idF, Ord idF, Ord idV) => trs -> Maybe (Map idF [Int])
getMRCanonRepMap trs
  | repMap == initRepMap = Nothing
  | otherwise = Just repMap
  where symbs = elems $ getTRSSymbols trs
        initRepMap = getTopRepMap trs
        initRepMapList = quicksortRepMapByMuLenghtDesc $ M.toList initRepMap
        lts = S.map lhs $ getTRSRules trs
        repMap = M.fromList $ removeFMuUpToMostRestrCanon lts initRepMap initRepMapList


removeFMuUpToMostRestrCanon :: (HasArity idF, Ord idF, Ord idV) => Set (Term idF idV) -> Map idF [Int] -> [(idF, [Int])] -> [(idF, [Int])]
removeFMuUpToMostRestrCanon lts rm [] = M.toList rm
removeFMuUpToMostRestrCanon lts rm (u:rml)
  | Prelude.null mu = M.toList rm
  | notInL = removeFMuUpToMostRestrCanon lts rm rml
  | mu == m = removeFMuUpToMostRestrCanon lts rm rml
  | otherwise = removeFMuUpToMostRestrCanon lts canonRM rml
  where (s,mu) = u
        notInL = and $ Prelude.map (\t-> not $ inTerm t s) (S.toList lts)
        subs = quicksortLength $ subsequences mu
        m = getMRCanonicalMuFromSubseq lts s subs
        f _ = Just m
        canonRM = M.alter f s rm


-- Get Less Restrictive Replacement Maps
getLRCanoneRepMaps :: (HasSignature trs idF idV, HasRules trs (Term idF idV), HasArity idF, HasMu idF, Ord idF, Ord idV) => trs -> Set (Map idF [Int])
getLRCanoneRepMaps trs
  | isJust mostRestrCanonRM = delete (fromJust mostRestrCanonRM) (delete (getTopRepMap trs) repMaps)
  | otherwise = S.empty
  where mostRestrCanonRM = getMRCanonRepMap trs
        mrcrmList = M.toList $ fromJust mostRestrCanonRM
        canonPreCombs = Prelude.map getCanonPreCombination mrcrmList
        symbs = S.toList . S.fromList $ Prelude.map fSymb canonPreCombs
        combs = doubArrayComb $ Prelude.map canonCombs canonPreCombs
        repMaps = S.fromList $ Prelude.map (\cs -> M.fromList [(symbs!!(i-1),cs!!(i-1)) | i <- [1..(length symbs)]]) $ combs


getCanonPreCombination :: (HasArity idF, Ord idF) => (idF, [Int]) -> PreCombinationRM idF
getCanonPreCombination (s, mu) = PreCombinationRM s ccombs
  where ar = getFArity s
        diff = (S.fromList [1..ar]) \\ (S.fromList mu)
        ccombs = Prelude.map (++ mu) . subsequences $ S.toList diff


inTerm :: (Ord idF) => Term idF idV -> idF -> Bool
inTerm t@(F f ts) f'
  | f == f' = True
  | otherwise = or [inTerm (ts!!(i-1)) f' | i <- [1..length ts]]
inTerm t@(V v) f' = False


getMRCanonicalMuFromSubseq :: (HasArity idF, Ord idF, Ord idV) => Set (Term idF idV) -> idF -> [[Int]] -> [Int]
getMRCanonicalMuFromSubseq _ f  [] = [getFArity f]
getMRCanonicalMuFromSubseq lts f (mu:mus)
  | canon = mu
  | otherwise = getMRCanonicalMuFromSubseq lts f mus
  where g = nonVarSubtsAreActive f mu
        canon = and . S.toList $ S.map g lts


nonVarSubtsAreActive :: (HasArity idF, Ord idF, Ord idV) => idF -> [Int] -> Term idF idV -> Bool
nonVarSubtsAreActive f mu t@(F f' ts)
  | f == f' = and [isVTerm (ts!!(d-1)) | d <- diff]
  | otherwise = and [nonVarSubtsAreActive f mu (ts!!(a-1)) | a <- [1..getFArity f']]
  where ar = S.fromList [1..getFArity f']
        diff = S.toList (ar \\ (S.fromList mu))
nonVarSubtsAreActive _ _ _ = True


getBotRepMap :: (HasSignature trs idF idV, HasArity idF, HasMu idF, Ord idF, Ord idV) => trs -> Map idF [Int]
getBotRepMap trs = repMap
  where symbs = elems $ getTRSSymbols trs
        repMap = M.fromList $ Prelude.map (\s -> (s, [])) symbs


getTopRepMap :: (HasSignature trs idF idV, HasArity idF, HasMu idF, Ord idF, Ord idV) => trs -> Map idF [Int]
getTopRepMap trs = repMap
  where symbs = elems $ getTRSSymbols trs
        repMap = M.fromList $ Prelude.map (\s -> (s, [1..getFArity s])) symbs


setRepMap :: (HasSignature trs idF idV, HasArity idF, HasMu idF, Ord idF, Ord idV) =>  Map idF [Int] -> trs -> trs
setRepMap rm trs = setTRSDSymbols newDSymbs $ setTRSCSymbols newCSymbs trs
  where dSymbs = getTRSDSymbols trs
        cSymbs = getTRSCSymbols trs
        repMu = replaceFMu rm
        newDSymbs = S.map repMu dSymbs
        newCSymbs = S.map repMu cSymbs


getRepMap :: (HasSignature trs idF idV, HasMu idF, Ord idF, Ord idV) => trs -> Map idF [Int]
getRepMap trs = M.fromList (repl)
  where symbs = elems $ getTRSSymbols trs
        repl = Prelude.map (\k -> (k,getFMu k)) symbs


replaceFMu :: (HasArity idF, HasMu idF, Ord idF) => Map idF [Int] -> idF -> idF
replaceFMu rm s
  | isNothing mu = s
  | otherwise = setFMu (fromJust mu) s
  where mu = M.lookup s rm


getDefinedSymbs :: (HasArity idF, Ord idF, Ord idV) => Set (Rule (Term idF idV)) -> Set idF
getDefinedSymbs rls
  | S.null rls = S.empty
  | isNothing f = getDefinedSymbs rlss
  | otherwise = S.insert (fromJust f) (getDefinedSymbs rlss)
  where (l :-> _) = elemAt 0 rls
        f = getF l
        rlss = S.fromList . tail . S.elems $ rls


getConstructorSymbs :: (HasArity idF, Ord idF, Ord idV) => Set (Rule (Term idF idV)) -> Set idF
getConstructorSymbs rls
  | S.null rls = S.empty
  | isNothing f = getConstructorSymbs rlss
  | otherwise = S.insert (fromJust f) (getConstructorSymbs rlss)
  where (_ :-> r) = elemAt 0 rls
        f = getF r
        rlss = S.fromList . tail . S.elems $ rls


getF :: (HasArity idF) => Term idF idV -> Maybe idF
getF (F f a) = Just (setFArity (length a) f)
getF _ = Nothing


applySigmaLCond :: (Ord idV) => (Map idV (Term idF idV)) -> Cond (Term idF idV) -> Cond (Term idF idV)
applySigmaLCond s (l :->* r) = (applySigma s l) :->* r
applySigmaLCond s (l :~> r) = (applySigma s l) :~> r
applySigmaLCond s (l :->+ r) = (applySigma s l) :->+ r
applySigmaLCond s (l :->*<- r) = (applySigma s l) :->*<- r
applySigmaLCond s (l :<--> r) = (applySigma s l) :<--> r
applySigmaLCond s (l :<-->* r) = (applySigma s l) :<-->* r
applySigmaLCond s (l :\->* r) = (applySigma s l) :\->* r
applySigmaLCond s (l :\-> r) = (applySigma s l) :\-> r
applySigmaLCond s (l :\->+ r) = (applySigma s l) :\->+ r
applySigmaLCond s (l :\->*<-/ r) = (applySigma s l) :\->*<-/ r
applySigmaLCond s (l :<-/\-> r) = (applySigma s l) :<-/\-> r
applySigmaLCond s (l :<-/\->* r) = (applySigma s l) :<-/\->* r
applySigmaLCond s (l :|>= r) = (applySigma s l) :|>= r
applySigmaLCond s (l :->= r) = (applySigma s l) :->= r
applySigmaLCond s (l :\->= r) = (applySigma s l) :\->= r



-----------------------------------------------------------------------------
-- JOINABILITY Functions
-----------------------------------------------------------------------------




createRewritingGraph :: (HasMu idF, Eq idF, Ord idV) => Term idF idV -> [RuleId idF idV] -> Maybe(Int) -> Gr (RewTerms idF idV) ()
createRewritingGraph x [] _ = G.empty
createRewritingGraph x rs md = iterateRewNode root rs (0, md)
  where rootRT = RewTerms x 0 0 0 Nothing [] Nothing
        root = mkGraph [(0, rootRT)] []


iterateRewNode :: (HasMu idF, Eq idF, Ord idV) => Gr (RewTerms idF idV) () -> [RuleId idF idV] -> (Int, Maybe(Int)) -> Gr (RewTerms idF idV) ()
iterateRewNode gr rs (d, (Just md))
  | d >= md = gr
  | otherwise = fgr
    where ndepth = d + 1
          nilnds = nonInspectedLNodes (labNodes gr) gr
          wgr = iterateRewNodeRules nilnds rs ndepth gr
          fgr = iterateRewNode wgr rs (ndepth, Just md)


nonInspectedLNodes :: [LNode (RewTerms idF idV)] -> Gr (RewTerms idF idV) () -> [LNode (RewTerms idF idV)]
nonInspectedLNodes [] _ = []
nonInspectedLNodes (n:nds) gr
  | (length successors) == 0 = n:(nonInspectedLNodes nds gr)
  | otherwise = nonInspectedLNodes nds gr
    where successors = lsuc gr (fst n)


iterateRewNodeRules :: (HasMu idF, Eq idF, Ord idV) => [LNode (RewTerms idF idV)] -> [RuleId idF idV] -> Int -> Gr (RewTerms idF idV) () -> Gr (RewTerms idF idV) ()
iterateRewNodeRules [] _ _ gr = gr
iterateRewNodeRules (n:ns) rs d gr = gro
  where gri = iterateRewNodeRule n rs d gr
        gro = iterateRewNodeRules ns rs d gri


iterateRewNodeRule :: (HasMu idF, Eq idF, Ord idV) => LNode (RewTerms idF idV) -> [RuleId idF idV] -> Int -> Gr (RewTerms idF idV) () -> Gr (RewTerms idF idV) ()
iterateRewNodeRule _ [] _ gr = gr
iterateRewNodeRule n (r:rs) d gr = gro
  where gri = iterateRewNodeFromPSubts (fPosSubterm [] (term $ snd n)) n r d gr
        gro = iterateRewNodeRule n rs d gri


iterateRewNodeFromPSubts :: (Eq idF, Ord idV) => [PosSubt idF idV] -> LNode (RewTerms idF idV) -> RuleId idF idV -> Int -> Gr (RewTerms idF idV) () -> Gr (RewTerms idF idV) ()
iterateRewNodeFromPSubts [] _ _ _ gr = gr
iterateRewNodeFromPSubts (p:ps) n r d gr = gro
  where gri = iterateRewNodeFromPSubt p n r d gr
        gro = iterateRewNodeFromPSubts ps n r d gri


iterateRewNodeFromPSubt :: (Eq idF, Ord idV) => PosSubt idF idV -> LNode (RewTerms idF idV) -> RuleId idF idV -> Int -> Gr (RewTerms idF idV) () -> Gr (RewTerms idF idV) ()
iterateRewNodeFromPSubt ps n (CRuleId crl ri) d gr = iterateRewNodeFromPSubt ps n (RuleId (rule crl) ri) d gr
iterateRewNodeFromPSubt ps n (RuleId rl ri) d gr
  | isNothing sig = gr
  | lid /= id = insEdge (fst n, lid, ()) gr
  | otherwise = (insEdge (fst n, id, ()) . insNode (id, rt)) gr
    where (l :-> r) = rl
          subt = subterm ps
          sig = matchSubstitution l subt
          p = pos ps
          newR = Just (applySigma (fromJust sig) r)
          w = term $ snd n
          id = noNodes gr
          lid = getLoopId newTerm gr
          newTerm = replace (fromJust newR) p w
          rt = RewTerms newTerm id d ri sig p newR
          terms = Prelude.map term . Prelude.map snd $ labNodes gr


getLoopId :: (Eq idF, Ord idV) => Term idF idV -> Gr (RewTerms idF idV) () -> Int
getLoopId t gr = minimum $ Prelude.map (\(i,x) -> if (term x) == t then i else id) (labNodes gr)
  where id = noNodes gr


fPosSubterm :: HasMu idF => [Int] -> Term idF idV -> [PosSubt idF idV]
fPosSubterm p t@(V v) = []
fPosSubterm p t@(F f ts) = (PosSubt p t):(concat [fPosSubterm (p++[p']) (ts!!(p' - 1)) | p' <- mu])
  where mu = getFMu f


maxDepthReached :: Maybe(Int) -> Gr (RewTerms idF idV) () -> Bool
maxDepthReached Nothing _ = False
maxDepthReached md gr
    | reachMaxDepth md nilnds = True
    | otherwise = False
  where nilnds = nonInspectedLNodes (labNodes gr) gr


reachMaxDepth :: Maybe(Int) -> [LNode (RewTerms idF idV)] -> Bool
reachMaxDepth (Just md) [] = False
reachMaxDepth (Just md) (n:ns)
  | depth (snd n) == md = True
  | otherwise = reachMaxDepth (Just md) ns


haveJoinTerm :: (Eq idF, Ord idV) => Gr (RewTerms idF idV) () -> Gr (RewTerms idF idV) () -> (Bool, Maybe(Term idF idV), [Term idF idV], [Term idF idV])
haveJoinTerm gr1 gr2
  | isEmpty gr1 = (False, Nothing, [], [])
  | isEmpty gr2 = (False, Nothing, [], [])
  | otherwise = (found,u,sp,tp)
  where lnodes1 = labNodes gr1
        lnodes2 = labNodes gr2
        (found,u,sn,tn) = findJoinTerm lnodes1 lnodes2
        sp = getDFSTermPath gr1 sn
        tp = getDFSTermPath gr2 tn

findJoinTerm :: (Eq idF, Ord idV) => [LNode (RewTerms idF idV)] -> [LNode (RewTerms idF idV)] -> (Bool, Maybe(Term idF idV), Node, Node)
findJoinTerm _ [] = (False, Nothing, 0, 0)
findJoinTerm [] _ = (False, Nothing, 0, 0)
findJoinTerm (a:as) bs = if found then (found,u,sn,tn) else findJoinTerm as bs
  where (found,u,sn,tn) = findJoinTerm' a bs

findJoinTerm' :: (Eq idF, Ord idV) => LNode (RewTerms idF idV) -> [LNode (RewTerms idF idV)] -> (Bool, Maybe(Term idF idV), Node, Node)
findJoinTerm' _ [] = (False, Nothing, 0, 0)
findJoinTerm' (n1,rt1) ((n2,rt2):bs)
  | term rt1 == term rt2 = (True, Just $ term rt1, n1, n2)
  | otherwise = findJoinTerm' (n1,rt1) bs

getDFSTermPath :: Gr (RewTerms idF idV) () -> Node -> [Term idF idV]
getDFSTermPath gr u
  | isNothing nodePath = []
  | otherwise = (term rt1):(Prelude.map getTermFromNode (fromJust nodePath))
  where (n1, rt1) = (labNodes gr) !! 0
        nodePath = dfs (suc gr) (== u) 0
        getTermFromNode n = term . fromJust $ lab gr n


reflexiveClosureGraph :: Gr (RewTerms idF idV) () -> Gr (RewTerms idF idV) ()
reflexiveClosureGraph gr = labnfilter refClosure gr
  where refClosure (n, rt) = depth rt <= 1


isNormalTerm :: (HasMu idF, Eq idF, Ord idV) => Term idF idV -> [RuleId idF idV] -> Bool
isNormalTerm t rs = (ngr == 1)
  where md = Just 1
        gr = createRewritingGraph t rs md
        ngr = noNodes gr


isGroundTerm :: Ord idV => Term idF idV -> Bool
isGroundTerm t = (length $ vars t) == 0


groundNormalTerms :: (HasMu idF, Eq idF, Ord idV) => (Term idF idV, Term idF idV) -> [RuleId idF idV] -> (Bool, Term idF idV, Term idF idV)
groundNormalTerms (t1, t2) rs
  | (isNormalTerm t1 rs) && (isGroundTerm t2) = (True,t2,t1)
  | (isNormalTerm t2 rs) && (isGroundTerm t1) = (True,t1,t2)
  | otherwise = (False,t1,t2)


normalWeight :: (Ord idV, Eq idF, HasMu idF) => Term idF idV -> [RuleId idF idV] -> Int
normalWeight t rids
  | isNormalTerm t rids = 1
  | otherwise = 0

{-
-- Examples
ff = IdFTRS { fid = 1, fname = "f", arity = 2}
cc = IdFTRS { fid = 2, fname = "c", arity = 0}
v1 = IdVTRS { vid = 1, vname = Just "x"}
v2 = IdVTRS { vid = 2, vname = Just "y"}
replMapEx = M.fromList [(ff, [1,2]), (cc, [])]

t1 :: Term IdFTRS IdVTRS
t1 = F ff [V v1, F ff [F ff [V v2, V v1], F cc []]]

t2 :: Term IdFTRS IdVTRS
t2 = F ff [V v2, V v1]

t3 :: Term IdFTRS IdVTRS
t3 = F ff [V v1, F ff [F ff [V v2, V v1], F cc []]]

t4 :: Term IdFTRS IdVTRS
t4 = F ff [V v2, V v1]

rule1 :: Rule (Term IdFTRS IdVTRS)
rule1 = t1 :-> t2

rule2 :: Rule (Term IdFTRS IdVTRS)
rule2 = t3 :-> t4

rule3 :: Rule (Term IdFTRS IdVTRS)
rule3 = t3 :-> t3
-}
