{-# LANGUAGE FlexibleContexts #-}
-----------------------------------------------------------------------------
-- |
-- Module      :  Utils.Auxiliar
-- Copyright   :  (c) muterm development team
-- License     :  see LICENSE
-- 
-- Maintainer  :  rgutierrez@dsic.upv.es
-- Stability   :  unstable
-- Portability :  non-portable 
--
-- This module manage auxiliary functions.
--
-----------------------------------------------------------------------------

module Utils.Auxiliar (

-- * Exported functions

nub2, concat2, getInt, bconcat, frmStr, isTrue, boolGroups, subsequences,
doubArrayComb, replaceString, shuffle, replaceNElem, noRiskyChars, 
translateFromRiskyChars

) where

import Data.List (nub, concat, lookup)
import Control.Monad.State (MonadState, get, put)
import Control.Monad (replicateM)
import System.Random (randomRIO)
import System.IO.Unsafe (unsafePerformIO)

import qualified Data.ByteString.Lazy.Char8 as B (pack)
import qualified Data.Binary.Builder as C (Builder, fromLazyByteString, empty, append)
import Data.Text(pack, unpack, replace)
import Data.Map as M(Map, fromList, fold)

-----------------------------------------------------------------------------
-- Functions
-----------------------------------------------------------------------------

-- | Tuple nub
nub2 :: (Eq a, Eq b) => ([a],[b]) -> ([a],[b])
nub2 (x,y) = (nub x, nub y)


-- | Tuple concat
concat2 :: ([[a]],[[b]]) -> ([a],[b])
concat2 (x,y) = (concat x, concat y)

-- | To obtain a unique numeric identifier for the function symbols
-- and variable symbols, we store this counter in a 'State' monad, and
-- each time we 'getInt' we obtain a new number.
getInt :: MonadState Int m => m Int
getInt = do n <- get
            put (n+1)
            return n

-- | append for ByteString
bs1 +|+ bs2 = C.append bs1 bs2

-- | frmBS = C.fromLazyByteString
frmStr :: String -> C.Builder
frmStr str = C.fromLazyByteString $ B.pack str

-- | concat a list of ByteString
bconcat [] = C.empty
bconcat (x:xs) = x +|+ (bconcat xs)


isTrue :: Maybe(Bool) -> Bool
isTrue Nothing = False
isTrue (Just t) = t


boolGroups :: Int -> [[Bool]]
boolGroups n
  | n > 0 = 
      let perms = replicateM n [True,False]
          np = 2 ^ (n - 1)
      in snd $ splitAt 1 (fst $ splitAt np perms)
  | otherwise = 
      let perms = replicateM n [True,False]
          np = 0
      in snd $ splitAt 1 (fst $ splitAt np perms)


-- | The 'subsequences' function returns the list of all subsequences of the argument.
--
-- >>> subsequences "abc"
-- ["","a","b","ab","c","ac","bc","abc"]
subsequences            :: [a] -> [[a]]
subsequences xs         =  [] : nonEmptySubsequences xs

-- | The 'nonEmptySubsequences' function returns the list of all subsequences of the argument,
--   except for the empty list.
--
-- >>> nonEmptySubsequences "abc"
-- ["a","b","ab","c","ac","bc","abc"]
nonEmptySubsequences         :: [a] -> [[a]]
nonEmptySubsequences []      =  []
nonEmptySubsequences (x:xs)  =  [x] : foldr f [] (nonEmptySubsequences xs)
  where f ys r = ys : (x : ys) : r


doubArrayComb :: [[a]] -> [[a]]
doubArrayComb [] = []
doubArrayComb (x:[]) = map (:[]) x
doubArrayComb (x:ys) = [[xi] ++ y | xi <- x, y <- doubArrayComb ys]

-- | Replace String
replaceString :: String -> String -> String -> String
replaceString src dst = unpack . replace (pack src) (pack dst) . pack

riskyCharsKVList :: [(String,String)]
riskyCharsKVList = 
  [
    --("'", "rc39"),
    ("*", "rc42"),
    ("_", "rc95")
  ]

riskyExcepCharsKVList :: [(String,String)]
riskyExcepCharsKVList = 
  [
    ("->rc42", "->*")
  ]

noRiskyChars :: String -> String
noRiskyChars st = foldl f (foldl f st riskyCharsKVList) riskyExcepCharsKVList
  where f s (k,v) = replaceString k v s

translateFromRiskyChars :: String -> String
translateFromRiskyChars st = foldl f st riskyCharsKVList
  where f s (k,v) = replaceString v k s


shuffle :: [a] -> [a]
shuffle [] = []
shuffle xs = (xs!!i):r
  where i = unsafePerformIO $ randomRIO (0, ((length xs) - 1) :: Int)
        r = shuffle (take i xs ++ drop (i+1) xs)


replaceNElem :: Int -> a -> [a] -> [a]
replaceNElem _ _ [] = []
replaceNElem i e (x:xs)
  | i < 0 = (x:xs)
  | i > 0 = x:(replaceNElem (i-1) e xs)
  | otherwise = e:xs